@@ -377,7 +377,7 @@ setMethod("determinant", signature(x = "Matrix", logarithm = "logical"),
 	  determinant(as(x,"dMatrix"), logarithm=logarithm, ...))
 
 setMethod("diag", signature(x = "Matrix"),
-	  function(x, nrow, ncol) .bail.out.1("diag", class(x)))
+	  function(x, nrow, ncol, names=TRUE) .bail.out.1("diag", class(x)))
 if(FALSE)## TODO: activate later
 setMethod("diag<-", signature(x = "Matrix"),
 	  function(x, value) .bail.out.1("diag", class(x)))
