@@ -1,7 +1,7 @@
 #  File src/library/base/R/datetime.R
 #  Part of the R package, https://www.R-project.org
 #
-#  Copyright (C) 1995-2019 The R Core Team
+#  Copyright (C) 1995-2022 The R Core Team
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -41,7 +41,11 @@ Sys.timezone <- function(location = TRUE)
                          inherits = FALSE, ifnotfound = NA_character_)))
         return(tz)
 
-    cacheIt <- function(tz) assign(".sys.timezone", tz, baseenv())
+    cacheIt <- function(tz) {
+        unlockBinding(".sys.timezone", baseenv())
+        assign(".sys.timezone", tz, baseenv())
+        lockBinding(".sys.timezone", baseenv())
+    }
 
     ## Many Unix set TZ, e.g. Solaris and AIX.
     ## For Solaris the system setting is a line in /etc/TIMEZONE
@@ -64,6 +68,9 @@ Sys.timezone <- function(location = TRUE)
                 Sys.setenv(TZDIR = "macOS")
         }
     }
+    if(Sys.getenv("TZDIR") == "macOS" && grepl("darwin", R.Version()$os))
+        Sys.setenv(TZDIR = "/var/db/timezone/zoneinfo")
+
 
     ## At least tzcode and glibc respect TZDIR.
     ## glibc uses $(datadir)/zoneinfo
@@ -354,7 +361,7 @@ as.POSIXct.default <- function(x, tz = "", ...)
     if(is.character(x) || is.factor(x))
 	return(as.POSIXct(as.POSIXlt(x, tz, ...), tz, ...))
     if(is.logical(x) && all(is.na(x)))
-        return(.POSIXct(as.numeric(x)))
+        return(.POSIXct(as.numeric(x), tz))
     stop(gettextf("do not know how to convert '%s' to class %s",
                   deparse1(substitute(x)),
                   dQuote("POSIXct")),
@@ -373,15 +380,15 @@ length.POSIXlt <- function(x) length(unclass(x)[[1L]])
     .POSIXlt(lapply(unclass(x), `length<-`, value),
              attr(x, "tzone"), oldClass(x))
 
-format.POSIXlt <- function(x, format = "", usetz = FALSE, ...)
+format.POSIXlt <- function(x, format = "", usetz = FALSE,
+                           digits = getOption("digits.secs"), ...)
 {
     if(!inherits(x, "POSIXlt")) stop("wrong class")
     if(any(f0 <- format == "")) {
         ## need list [ method here.
 	times <- unlist(unclass(x)[1L:3L])[f0]
-	secs <- x$sec[f0]; secs <- secs[!is.na(secs)]
-        np <- getOption("digits.secs")
-        np <- if(is.null(np)) 0L else min(6L, np)
+	secs <- x$sec[f0]; secs <- secs[is.finite(secs)]
+        np <- if(is.null(digits)) 0L else min(6L, digits)
         if(np >= 1L)
             for (i in seq_len(np)- 1L)
                 if(all( abs(secs - round(secs, i)) < 1e-6 )) {
@@ -389,7 +396,7 @@ format.POSIXlt <- function(x, format = "", usetz = FALSE, ...)
                     break
                 }
 	format[f0] <-
-	    if(all(times[!is.na(times)] == 0)) "%Y-%m-%d"
+	    if(all(times[is.finite(times)] == 0)) "%Y-%m-%d"
 	    else if(np == 0L) "%Y-%m-%d %H:%M:%S"
 	    else paste0("%Y-%m-%d %H:%M:%OS", np)
     }
@@ -541,7 +548,7 @@ Summary.POSIXct <- function (..., na.rm)
         stop(gettextf("'%s' not defined for \"POSIXt\" objects", .Generic),
              domain = NA)
     args <- list(...)
-    tz <- do.call("check_tzones", args)
+    tz <- do.call(check_tzones, args)
     .POSIXct(NextMethod(.Generic), tz = tz, cl = oldClass(args[[1L]]))
 }
 
@@ -552,7 +559,7 @@ Summary.POSIXlt <- function (..., na.rm)
         stop(gettextf("'%s' not defined for \"POSIXt\" objects", .Generic),
              domain = NA)
     args <- list(...)
-    tz <- do.call("check_tzones", args)
+    tz <- do.call(check_tzones, args)
     args <- lapply(args, as.POSIXct)
     val <- do.call(.Generic, c(args, na.rm = na.rm))
     as.POSIXlt(.POSIXct(val, tz))
@@ -592,14 +599,19 @@ anyNA.POSIXlt <- function(x, recursive = FALSE)
     anyNA(as.POSIXct(x))
 
 ## <FIXME> check the argument validity
-## This is documented to remove the timezone
-c.POSIXct <- function(..., recursive = FALSE)
-    .POSIXct(c(unlist(lapply(list(...),
-                             function(e) unclass(as.POSIXct(e))))))
+## This is documented to remove the timezone (unless all are marked with
+## the same).
+c.POSIXct <- function(..., recursive = FALSE) {
+    x <- lapply(list(...), function(e) unclass(as.POSIXct(e)))
+    tzones <- lapply(x, attr, "tzone")
+    tz <- if(length(unique(tzones)) == 1L) tzones[[1L]] else NULL
+    .POSIXct(c(unlist(x)), tz)
+}
 
 ## we need conversion to POSIXct as POSIXlt objects can be in different tz.
-c.POSIXlt <- function(..., recursive = FALSE)
-    as.POSIXlt(do.call("c", lapply(list(...), as.POSIXct)))
+c.POSIXlt <- function(..., recursive = FALSE) {
+    as.POSIXlt(do.call(c, lapply(list(...), as.POSIXct)))
+}
 
 
 ISOdatetime <- function(year, month, day, hour, min, sec, tz = "")
@@ -1008,22 +1020,30 @@ cut.POSIXt <-
         if(valid == 8L) incr <- 25*3600 # DSTdays
         if(valid == 6L) {               # months
             start$mday <- 1L
-            end <- as.POSIXlt(max(x, na.rm = TRUE))
+            maxx <- max(x, na.rm = TRUE)
+            end <- as.POSIXlt(maxx)
             step <- if(length(by2) == 2L) as.integer(by2[1L]) else 1L
             end <- as.POSIXlt(end + (31 * step * 86400))
             end$mday <- 1L
             end$isdst <- -1L
             breaks <- seq(start, end, breaks)
+            ## 31 days ahead could give an empty level, so
+	    lb <- length(breaks)
+	    if(maxx < breaks[lb-1]) breaks <- breaks[-lb]
         } else if(valid == 7L) {        # years
             start$mon <- 0L
             start$mday <- 1L
-            end <- as.POSIXlt(max(x, na.rm = TRUE))
+            maxx <- max(x, na.rm = TRUE)
+            end <- as.POSIXlt(maxx)
             step <- if(length(by2) == 2L) as.integer(by2[1L]) else 1L
             end <- as.POSIXlt(end + (366 * step* 86400))
             end$mon <- 0L
             end$mday <- 1L
             end$isdst <- -1L
             breaks <- seq(start, end, breaks)
+            ## 366 days ahead could give an empty level, so
+	    lb <- length(breaks)
+	    if(maxx < breaks[lb-1]) breaks <- breaks[-lb]
         } else if(valid == 9L) {        # quarters
             qtr <- rep(c(0L, 3L, 6L, 9L), each = 3L)
             start$mon <- qtr[start$mon + 1L]
@@ -1209,7 +1229,7 @@ function(x, units = c("secs", "mins", "hours", "days", "months", "years"))
             i <- match(i, names(x),
                        incomparables = c("", NA_character_))
         if(mj)
-            .POSIXlt(lapply(X = unclass(x), FUN = "[", i, drop = drop),
+            .POSIXlt(lapply(X = unclass(x), FUN = `[`, i, drop = drop),
                      attr(x, "tzone"), oldClass(x))
         else
             unclass(x)[[j]][i]
@@ -1324,7 +1344,7 @@ is.numeric.difftime <- function(x) FALSE
 ## Class generators added in 2.11.0, class order changed in 2.12.0.
 
 ## FIXME:
-## At least temporarily avoide structure() for performance reasons.
+## At least temporarily avoid structure() for performance reasons.
 ## .POSIXct <- function(xx, tz = NULL)
 ##     structure(xx, class = c("POSIXct", "POSIXt"), tzone = tz)
 .POSIXct <- function(xx, tz = NULL, cl = c("POSIXct", "POSIXt")) {
@@ -1374,25 +1394,32 @@ OlsonNames <- function(tzdir = NULL)
         if(.Platform$OS.type == "windows")
             tzdir <- Sys.getenv("TZDIR", file.path(R.home("share"), "zoneinfo"))
         else {
-            ## Try known locations in turn.
-            ## The list is not exhaustive (mac OS 10.13's
-            ## /usr/share/zoneinfo is a symlink) and there is a risk that
-            ## the wrong one is found.
-            ## We assume that if the second exists that the system was
-            ## configured with --with-internal-tzcode
-            tzdirs <- c(Sys.getenv("TZDIR"), # defaults to ""
-                        file.path(R.home("share"), "zoneinfo"),
-                        "/usr/share/zoneinfo", # Linux, macOS, FreeBSD
-                        "/share/zoneinfo", # in musl's search
-                        "/usr/share/lib/zoneinfo", # Solaris, AIX
-                        "/usr/lib/zoneinfo",   # early glibc
-                        "/usr/local/etc/zoneinfo", # tzcode default
-                        "/etc/zoneinfo", "/usr/etc/zoneinfo")
-            tzdirs <- tzdirs[file.exists(tzdirs)]
-            if (!length(tzdirs)) {
-                warning("no Olson database found")
-                return(character())
-            } else tzdir <- tzdirs[1L]
+            if(Sys.getenv("TZDIR") == "internal")
+                tzdir <- file.path(R.home("share"), "zoneinfo")
+            else if (grepl("darwin", R.Version()$os) &&
+                     Sys.getenv("TZDIR") == "macOS") {
+                tzdir <- "/var/db/timezone/zoneinfo"
+            } else {
+                ## Try known locations in turn.
+                ## The list is not exhaustive (mac OS 10.13's
+                ## /usr/share/zoneinfo is a symlink) and there is a risk that
+                ## the wrong one is found.
+                ## We assume that if the second exists that the system was
+                ## configured with --with-internal-tzcode
+                tzdirs <- c(Sys.getenv("TZDIR"), # defaults to ""
+                            file.path(R.home("share"), "zoneinfo"),
+                            "/usr/share/zoneinfo", # Linux, macOS, FreeBSD
+                            "/share/zoneinfo", # in musl's search
+                            "/usr/share/lib/zoneinfo", # Solaris, AIX
+                            "/usr/lib/zoneinfo",   # early glibc
+                            "/usr/local/etc/zoneinfo", # tzcode default
+                            "/etc/zoneinfo", "/usr/etc/zoneinfo")
+                tzdirs <- tzdirs[file.exists(tzdirs)]
+                if (!length(tzdirs)) {
+                    warning("no Olson database found")
+                    return(character())
+                } else tzdir <- tzdirs[1L]
+            }
         }
     } else if(!dir.exists(tzdir))
         stop(sprintf("%s is not a directory", sQuote(tzdir)), domain = NA)
@@ -1419,7 +1446,7 @@ OlsonNames <- function(tzdir = NULL)
     if(!missing(i) && is.character(i)) {
         i <- match(i, names(x), incomparables = c("", NA_character_))
     }
-    .POSIXlt(lapply(X = unclass(x), FUN = "[[", i, drop = drop),
+    .POSIXlt(lapply(X = unclass(x), FUN = `[[`, i, drop = drop),
              attr(x, "tzone"), oldClass(x))
 }
 
@@ -1459,5 +1486,18 @@ as.list.POSIXlt <- function(x, ...)
 as.list.difftime <- function(x, ...)
     lapply(unclass(x), .difftime, attr(x, "units"), oldClass(x))
 
+## Added in 4.1.0.
 
+rep.difftime <- function(x, ...)
+    .difftime(NextMethod("rep"), attr(x, "units"), oldClass(x))
 
+`[<-.difftime` <- function(x, i, value) {
+    if(inherits(value, "difftime") && !identical(units(x), units(value)))
+        units(value) <- units(x)
+    NextMethod("[<-")
+}
+
+## Added in 4.2.0.
+
+as.vector.POSIXlt <- function(x, mode = "any")
+    as.vector(as.list(x), mode)
