@@ -1,7 +1,7 @@
 #  File src/library/base/R/sort.R
 #  Part of the R package, https://www.R-project.org
 #
-#  Copyright (C) 1995-2019 The R Core Team
+#  Copyright (C) 1995-2022 The R Core Team
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -203,6 +203,8 @@ order <- function(..., na.last = TRUE, decreasing = FALSE,
     }
 
     if(method != "radix" && !is.na(na.last)) {
+        if(length(decreasing) > 1L)
+            stop("'decreasing' of length > 1 is only for method = \"radix\"")
         return(.Internal(order(na.last, decreasing, ...)))
     }
 
@@ -234,8 +236,8 @@ sort.list <- function(x, partial = NULL, na.last = TRUE, decreasing = FALSE,
 
     method <- match.arg(method)
     if (method == "auto" &&
-        (is.numeric(x) || is.factor(x) || is.logical(x) || is.object(x)) &&
-        is.integer(length(x)))
+        (is.numeric(x) || is.factor(x) || is.logical(x) ||
+         (is.object(x) && !is.atomic(x))) && is.integer(length(x)))
         method <- "radix"
     if(!is.null(partial))
         .NotYetUsed("partial != NULL")
@@ -263,9 +265,16 @@ Have you called 'sort' on a list?")
 
 ## xtfrm is now primitive
 ## xtfrm <- function(x) UseMethod("xtfrm")
-xtfrm.default <- function(x)
-    if(is.numeric(x)) unclass(x) else as.vector(rank(x, ties.method = "min",
-                                                     na.last = "keep"))
+xtfrm.default <- function(x) {
+    y <- if(is.numeric(x))
+             unclass(x)
+         else as.vector(rank(x, ties.method = "min", na.last = "keep"))
+    if(!is.numeric(y) ||
+       ((length(y) != length(x)) && !inherits(x, "data.frame")))
+        stop("cannot xtfrm 'x'")
+    y
+}
+
 xtfrm.factor <- function(x) as.integer(x) # primitive, so needs a wrapper
 
 ## ## Moved to package survival
