@@ -1,7 +1,7 @@
 #  File src/library/splines/R/splineClasses.R
-#  Part of the R package, http://www.R-project.org
+#  Part of the R package, https://www.R-project.org
+#  Copyright (C) 2000-2018 The R Core Team
 #  Copyright (C) 1998 Douglas M. Bates and William N. Venables.
-#  Copyright (C) 2000-2015 The R Core Team
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -14,7 +14,7 @@
 #  GNU General Public License for more details.
 #
 #  A copy of the GNU General Public License is available at
-#  http://www.r-project.org/Licenses/
+#  https://www.R-project.org/Licenses/
 
 #### Classes and methods for determining and manipulating interpolation
 #### splines.
@@ -37,15 +37,18 @@
 
 splineDesign <-
     ## Creates the "design matrix" for a collection of B-splines.
-    function(knots, x, ord = 4, derivs = integer(nx), outer.ok = FALSE,
+    function(knots, x, ord = 4L, derivs = 0L, outer.ok = FALSE,
              sparse = FALSE)
 {
-    knots <- sort(as.numeric(knots))
-    if((nk <- length(knots)) <= 0) stop("must have at least 'ord' knots")
+    if((nk <- length(knots <- as.numeric(knots))) <= 0)
+        stop("must have at least 'ord' knots")
+    if(is.unsorted(knots)) knots <- sort.int(knots)
     x <- as.numeric(x)
     nx <- length(x)
-    if(length(derivs) != nx)
-	stop("length of 'derivs' must match length of 'x'")
+    ## derivs is re-cycled to length(x) in C
+    if(length(derivs) > nx)
+	stop("length of 'derivs' is larger than length of 'x'")
+    if(length(derivs) < 1L) stop("empty 'derivs'")
     ord <- as.integer(ord)
     if(ord > nk || ord < 1)
 	stop("'ord' must be positive integer, at most the number of knots")
@@ -56,24 +59,27 @@ splineDesign <-
                       "2*ord -1", 2*ord -1),
              domain = NA)
 
-    o1 <- ord - 1L
+    degree <- ord - 1L
 ### FIXME: the 'outer.ok && need.outer' handling would more efficiently happen
 ###        in the underlying C code - with some programming effort though..
-    if(need.outer <- any(x < knots[ord] | knots[nk - o1] < x)) {
+    if(need.outer <- any(x < knots[ord] | knots[nk - degree] < x)) {
         if(outer.ok) { ## x[] is allowed to be 'anywhere'
-            ## extend knots set "temporarily"
-            in.x <- knots[1L] < x & x < knots[nk]
-	    knots <- knots[c(rep.int(1L, o1), 1L:nk, rep.int(nk, o1))]
-            if((x.out <- !all(in.x))) {
-                x <- x[in.x]
-                nnx <- length(x)
-            }
+	    in.x <- knots[1L] <= x & x <= knots[nk]
+	    if((x.out <- !all(in.x))) {
+		x <- x[in.x]
+		nnx <- length(x)
+	    }
+	    ## extend knots set "temporarily": the boundary knots must be repeated >= 'ord' times.
+            ## NB: If these are already repeated originally, then, on the *right* only, we need
+            ##    to make sure not to add more than needed
+            dkn <- diff(knots)[(nk-1L):1] # >= 0, since they are sorted
+	    knots <- knots[c(rep.int(1L, degree),
+                             seq_len(nk),
+                             rep.int(nk, max(0L, ord - match(TRUE, dkn > 0))))]
 	} else
-	stop(gettextf("the 'x' data must be in the range %g to %g unless you set '%s'",
-		      knots[ord],
-                      knots[nk - o1],
-                      "outer.ok = TRUE"),
-             domain = NA)
+	    stop(gettextf("the 'x' data must be in the range %g to %g unless you set '%s'",
+			  knots[ord], knots[nk - degree], "outer.ok = TRUE"),
+		 domain = NA)
     }
     temp <- .Call(C_spline_basis, knots, ord, x, derivs)
     ncoef <- nk - ord
@@ -81,7 +87,7 @@ splineDesign <-
     ii <- if(need.outer && x.out) { # only assign non-zero for x[]'s "inside" knots
         rep.int((1L:nx)[in.x], rep.int(ord, nnx))
     } else rep.int(1L:nx, rep.int(ord, nx))
-    jj <- c(outer(1L:ord, attr(temp, "Offsets"), "+"))
+    jj <- c(outer(1L:ord, attr(temp, "Offsets"), `+`))
     ## stopifnot(length(ii) == length(jj))
 
     if(sparse) {
@@ -89,27 +95,21 @@ splineDesign <-
 	    stop(gettextf("%s needs package 'Matrix' correctly installed",
                           "splineDesign(*, sparse=TRUE)"),
                  domain = NA)
-
 	if(need.outer) { ## shift column numbers and drop those "outside"
-	    jj <- jj - o1 - 1L
+	    jj <- jj - degree - 1L
 	    ok <- 0 <= jj & jj < ncoef
-	    as(new("dgTMatrix",
-		   i = ii[ok] - 1L,
-		   j = jj[ok],
-		   x = as.double(temp[ok]),
-		   Dim = c(nx, ncoef)), "CsparseMatrix")
+	    methods::as(methods::new("dgTMatrix", i = ii[ok] - 1L, j = jj[ok],
+				     x = as.double(temp[ok]), # vector, not matrix
+				     Dim = c(nx, ncoef)), "CsparseMatrix")
 	}
 	else
-	    as(new("dgTMatrix",
-		   i = ii - 1L,
-		   j = jj - 1L,
-		   x = as.double(temp),
-		   Dim = c(nx, ncoef)), "CsparseMatrix")
-
+	    methods::as(methods::new("dgTMatrix", i = ii - 1L, j = jj - 1L,
+				     x = as.double(temp), # vector
+				     Dim = c(nx, ncoef)), "CsparseMatrix")
     } else { ## traditional (dense) matrix
 	design <- matrix(double(nx * ncoef), nx, ncoef)
 	if(need.outer) { ## shift column numbers and drop those "outside"
-	    jj <- jj - o1
+	    jj <- jj - degree
 	    ok <- 1 <= jj & jj <= ncoef
 	    design[cbind(ii, jj)[ok , , drop=FALSE]] <- temp[ok]
 	}
@@ -121,26 +121,39 @@ splineDesign <-
 
 interpSpline <-
     ## Determine the natural interpolation spline.
-    function(obj1, obj2, bSpline = FALSE, period = NULL, na.action = na.fail)
+    function(obj1, obj2, bSpline = FALSE, period = NULL, ord = 4L,
+             na.action = na.fail, sparse = FALSE)
     UseMethod("interpSpline")
 
+##>>> FIXME: any  ord != 4 needs adaption in splineDesign(), i.e.,
+##>>> =====       --------  probably in spline_basis() in ../src/splines.c
 interpSpline.default <-
-    function(obj1, obj2, bSpline = FALSE, period = NULL, na.action = na.fail)
+    function(obj1, obj2, bSpline = FALSE, period = NULL,
+             ord = 4L, na.action = na.fail, sparse = FALSE)
 {
-    ord <- 4 # spline order -- future: want to use other 'order'/'degree'!
-    deg <- ord - 1
-
+    ## spline order 'ord' == 'degree' + 1
+    stopifnot(exprs = {
+        (degree <- ord - 1L) >= 0
+        length(degree) == 1L
+        degree == as.integer(degree)
+    })
     frm <- na.action(data.frame(x = as.numeric(obj1), y = as.numeric(obj2)))
     frm <- frm[order(frm$x), ]
     ndat <- nrow(frm)
     x <- frm$x
     if(anyDuplicated(x))
 	stop("values of 'x' must be distinct")
-    ## 'deg' extra knots (shifted) out on each side :
-    knots <- c(x[1L:deg] + x[1L] - x[ord], x,
-	       x[ndat + (1L:deg) - deg] + x[ndat] - x[ndat - deg])
-    derivs <- c(2, integer(ndat), 2) # 2nd derivs coerced to 0 in solve() below
-    x	   <- c(x[1L], x, x[ndat])
+    if(length(x) < ord)
+        stop(gettextf("must have at least 'ord'=%d points", ord), domain=NA)
+    ## 'degree' extra knots (shifted) out on each side :
+    iDeg <- seq_len(degree)
+    knots <- c(x[iDeg] + x[1L] - x[ord], x,
+	       x[ndat + iDeg - degree] + x[ndat] - x[ndat - degree])
+    if(even <- (ord %% 2 == 0)) { ## natural boundary conditions:
+	nu <- ord %/% 2L          ## nu-th derivs coerced to 0  [in solve() below]
+	derivs <- c(nu, integer(ndat),  nu)
+	x      <- c(x[1L],   x,    x[ndat])
+    }
 ## Solving the system of equations for the spline coefficients can be
 ## simplified by using banded matrices but the required LINPACK routines
 ## are not loaded as part of S.
@@ -175,24 +188,27 @@ interpSpline.default <-
 ##	ipvt = z$ipvt,
 ##	b = c(0, y, 0),
 ##	job = 1L)
-    des <- splineDesign(knots, x, ord, derivs)
-    coeff <- solve(as.matrix(des), c(0, frm$y, 0))
-    value <- list(knots = knots, coefficients = coeff, order = ord)
-    attr(value, "formula") <-
-	do.call("~", list(substitute(obj2), substitute(obj1)))
-    class(value) <- c("nbSpline", "bSpline", "spline")
+    des <- splineDesign(knots, x, ord, derivs, sparse=sparse)
+    y <- c(0, frm$y, 0)
+    coeff <- if(sparse) Matrix::solve(des, Matrix::..2dge(y), sparse=TRUE)
+             else solve(des, y)
+    value <- structure(
+        list(knots = knots, coefficients = coeff, order = ord),
+             formula = do.call("~", list(substitute(obj2), substitute(obj1))),
+        class = c("nbSpline", "bSpline", "spline"))
     if (bSpline) return(value)
     ## else convert from B- to poly-Spline:
     value <- polySpline(value)
     coeff <- coef(value)
-    coeff[ , 1] <- frm$y
-    coeff[1, deg] <- coeff[nrow(coeff), deg] <- 0
+    coeff[ , 1L] <- frm$y
+    coeff[1L, degree] <- coeff[nrow(coeff), degree] <- 0
     value$coefficients <- coeff
     value
 }
 
 interpSpline.formula <-
-    function(obj1, obj2, bSpline = FALSE, period = NULL, na.action = na.fail)
+    function(obj1, obj2, bSpline = FALSE, period = NULL,
+             ord = 4L, na.action = na.fail, sparse = FALSE)
 {
     form <- as.formula(obj1)
     if (length(form) != 3)
@@ -200,19 +216,19 @@ interpSpline.formula <-
     local <- if (missing(obj2)) sys.parent(1) else as.data.frame(obj2)
     value <- interpSpline(as.numeric(eval(form[[3L]], local)),
 			  as.numeric(eval(form[[2L]], local)),
-			  bSpline = bSpline, period = period,
-			  na.action = na.action)
+			  bSpline=bSpline, period=period, ord=ord,
+			  na.action=na.action, sparse=sparse)
     attr(value, "formula") <- form
     value
 }
 
 periodicSpline <-
     ## Determine the periodic interpolation spline.
-    function(obj1, obj2, knots, period = 2 * pi, ord = 4)
+    function(obj1, obj2, knots, period = 2 * pi, ord = 4L)
     UseMethod("periodicSpline")
 
 periodicSpline.default <-
-    function(obj1, obj2, knots, period = 2 * pi, ord = 4)
+    function(obj1, obj2, knots, period = 2 * pi, ord = 4L)
 {
     x <- as.numeric(obj1)
     y <- as.numeric(obj2)
@@ -226,9 +242,9 @@ periodicSpline.default <-
     if(any((x[-1L] - x[ - lenx]) <= 0))
 	stop("values of 'x' must be strictly increasing")
     if(ord < 2) stop("'ord' must be >= 2")
-    o1 <- ord - 1
+    degree <- ord - 1L
     if(!missing(knots)) {
-	period <- knots[length(knots) - o1] - knots[1L]
+	period <- knots[length(knots) - degree] - knots[1L]
     }
     else {
 	knots <- c(x[(lenx - (ord - 2)):lenx] - period, x, x[1L:ord] + period)
@@ -237,30 +253,28 @@ periodicSpline.default <-
 	stop("the range of 'x' values exceeds one period")
     y <- y[ind]
     coeff.mat <- splineDesign(knots, x, ord)
+    i1 <- seq_len(degree)
     sys.mat <- coeff.mat[, (1L:lenx)]
-    sys.mat[, seq_len(o1)] <- sys.mat[, seq_len(o1)] +
-	coeff.mat[, lenx + seq_len(o1)]
+    sys.mat[, i1] <- sys.mat[, i1] + coeff.mat[, lenx + i1]
     coeff <- qr.coef(qr(sys.mat), y)
-    coeff <- c(coeff, coeff[seq_len(o1)])
-    value <- list(knots = knots, coefficients = coeff, order = ord,
-		  period = period)
-    attr(value, "formula") <- do.call("~", as.list(sys.call())[3:2])
-    class(value) <- c("pbSpline", "bSpline", "spline")
-    value
+    coeff <- c(coeff, coeff[i1])
+    structure(list(knots = knots, coefficients = coeff,
+		   order = ord, period = period),
+	      formula = do.call("~", as.list(sys.call())[3:2]),
+	      class = c("pbSpline", "bSpline", "spline"))
 }
 
-periodicSpline.formula <- function(obj1, obj2, knots, period = 2 * pi, ord = 4)
+periodicSpline.formula <- function(obj1, obj2, knots, period = 2 * pi, ord = 4L)
 {
     form <- as.formula(obj1)
     if (length(form) != 3)
 	stop("'formula' must be of the form \"y ~ x\"")
     local <- if (missing(obj2)) sys.parent(1) else as.data.frame(obj2)
     ## 'missing(knots)' is transfered :
-    value <-  periodicSpline(as.numeric(eval(form[[3L]], local)),
-			     as.numeric(eval(form[[2L]], local)),
-			     knots = knots, period = period, ord = ord)
-    attr(value, "formula") <- form
-    value
+    structure(periodicSpline.default(as.numeric(eval(form[[3L]], local)),
+				     as.numeric(eval(form[[2L]], local)),
+				     knots = knots, period = period, ord = ord),
+	      formula = form)
 }
 
 polySpline <-
@@ -284,21 +298,19 @@ polySpline.bSpline <- function(object, ...)
     coeff[, 1] <- asVector(predict(object, knots))
     if(ord > 1) {
 	for(i in 2:ord) {
-	    coeff[, i] <- asVector(predict(object, knots, deriv = i - 1))/
-		prod(1L:(i - 1))
+	    coeff[, i] <- asVector(predict(object, knots, deriv = i - 1L))/
+		prod(seq_len(i - 1L))
 	}
     }
-    value <- list(knots = knots, coefficients = coeff)
-    attr(value, "formula") <- attr(object, "formula")
-    class(value) <- c("polySpline", "spline")
-    value
+    structure(list(knots = knots, coefficients = coeff),
+	      formula = attr(object, "formula"),
+	      class = c("polySpline", "spline"))
 }
 
 polySpline.nbSpline <- function(object, ...)
 {
-    value <- NextMethod("polySpline")
-    class(value) <- c("npolySpline", "polySpline", "spline")
-    value
+    structure(NextMethod("polySpline"),
+              class = c("npolySpline", "polySpline", "spline"))
 }
 
 polySpline.pbSpline <- function(object, ...)
@@ -333,9 +345,8 @@ xyVector <- ## Constructor for the xyVector class
 {
     x <- as.vector(x)
     y <- as.vector(y)
-    if(length(x) != length(y)) {
+    if(length(x) != length(y))
 	stop("lengths of 'x' and 'y' must be the same")
-    }
     structure(list(x = x, y = y), class = "xyVector")
 }
 
@@ -368,13 +379,13 @@ predict.polySpline <- function(object, x, nseg = 50, deriv = 0, ...)
 	stop(gettextf("'deriv' must be between 0 and %d", ord - 1),
              domain = NA)
     while(deriv > 0) {
-	ord <- ord - 1
-	coeff <- t(t(coeff[, -1]) * (1L:ord))
-	deriv <- deriv - 1
+	ord <- ord - 1L
+	coeff <- t(t(coeff[, -1]) * seq_len(ord))
+	deriv <- deriv - 1L
     }
     y <- coeff[i, ord]
     if(ord > 1) {
-	for(j in (ord - 1):1)
+	for(j in (ord - 1L):1)
 	    y <- y * delx + coeff[i, j]
     }
     xyVector(x = x, y = y)
@@ -395,14 +406,14 @@ predict.bSpline <- function(object, x, nseg = 50, deriv = 0, ...)
 	accept <- TRUE
     } else accept <- knots[ord] <= x & x <= knots[ncoeff + 1]
     y <- x
-    y[!accept] <- NA
+    y[!accept] <- NA # (C_spline_value's FIXME)
     y[accept] <- .Call(C_spline_value, knots, coeff, ord, x[accept], deriv)
     xyVector(x = x, y = y)
 }
 
 predict.nbSpline <- function(object, x, nseg = 50, deriv = 0, ...)
 {
-    value <- NextMethod("predict")
+    value <- NextMethod("predict") # predict.bSpline() -> NaN outside knots
     if(!any(is.na(value$y))) # when x were inside knots
 	return(value)
 
@@ -442,6 +453,7 @@ predict.pbSpline <- function(object, x, nseg = 50, deriv = 0, ...)
     ncoeff <- length(coef(object))
     if(missing(x))
 	x <- seq.int(knots[ord], knots[ord] + period, length.out = nseg + 1)
+    ## Because of C_spline_value's FIXME, we move the outside-knots x[] values:
     x.original <- x
     if(any(ind <- x < knots[ord]))
 	x[ind] <- x[ind] + period * (1 + (knots[ord] - x[ind]) %/% period)
@@ -453,7 +465,7 @@ predict.pbSpline <- function(object, x, nseg = 50, deriv = 0, ...)
 
 predict.npolySpline <- function(object, x, nseg = 50, deriv = 0, ...)
 {
-    value <- NextMethod()
+    value <- NextMethod() # typically predict.polySpline()
     if(!any(is.na(value$y))) # when x were inside knots
 	return(value)
 
@@ -566,6 +578,8 @@ backSpline.npolySpline <- function(object)
     if(any(kdiff <= 0))
 	stop("knot positions must be strictly increasing")
     coeff <- coef(object)
+    if((ord <- ncol(coeff)) != 4)
+	stop("currently implemented only for cubic splines")
     bknots <- coeff[, 1]
     adiff <- diff(bknots)
     if(all(adiff < 0))
