@@ -1,7 +1,7 @@
 #  File src/library/stats/R/contr.poly.R
 #  Part of the R package, https://www.R-project.org
 #
-#  Copyright (C) 1995-2017 The R Core Team
+#  Copyright (C) 1995-2019 The R Core Team
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -24,12 +24,12 @@ contr.poly <- function (n, scores = 1:n, contrasts = TRUE, sparse = FALSE)
     make.poly <- function(n, scores)
     {
 	y <- scores - mean(scores)
-	X <- outer(y, seq_len(n) - 1, "^")
+	X <- outer(y, seq_len(n) - 1, `^`)
 	QR <- qr(X)
 	z <- QR$qr
 	z <- z *(row(z) == col(z))
 	raw <- qr.qy(QR, z)
-	Z <- sweep(raw, 2L, apply(raw, 2L, function(x) sqrt(sum(x^2))), "/",
+	Z <- sweep(raw, 2L, apply(raw, 2L, function(x) sqrt(sum(x^2))), `/`,
 		   check.margin=FALSE)
 	colnames(Z) <- paste0("^", 1L:n - 1L)
 	Z
@@ -80,7 +80,7 @@ poly <- function(x, ..., degree = 1, coefs = NULL, raw = FALSE, simple = FALSE)
     if(degree < 1)
         stop("'degree' must be at least 1")
     if(raw) {
-        Z <- outer(x, 1L:degree, "^")
+        Z <- outer(x, 1L:degree, `^`)
         colnames(Z) <- 1L:degree
     } else {
 	if(is.null(coefs)) { # fitting
@@ -89,7 +89,7 @@ poly <- function(x, ..., degree = 1, coefs = NULL, raw = FALSE, simple = FALSE)
 		stop("'degree' must be less than number of unique points")
 	    xbar <- mean(x)
 	    x <- x - xbar
-	    X <- outer(x, 0L:degree, "^")
+	    X <- outer(x, 0L:degree, `^`)
 	    QR <- qr(X)
 	    if(QR$rank < degree)
 		stop("'degree' must be less than number of unique points")
@@ -131,15 +131,15 @@ predict.poly <- function(object, newdata, ...)
 
 makepredictcall.poly  <- function(var, call)
 {
-    if(as.character(call)[1L] != "poly") return(call)
-    call$coefs <- attr(var, "coefs")
+    if(as.character(call)[1L] == "poly" || (is.call(call) && identical(eval(call[[1L]]), poly)))
+	call$coefs <- attr(var, "coefs")
     call
 }
 
 polym <- function (..., degree = 1, coefs = NULL, raw = FALSE)
 {
     dots <- list(...)
-    nd <- length(if(is.null(coefs)) dots else coefs)  # number of variables
+    nd <- length(coefs %||% dots)  # number of variables
     if (nd == 0)
         stop("must supply one or more vectors")
     ## z:= combinations of (0:degree) of all variables
@@ -158,12 +158,12 @@ polym <- function (..., degree = 1, coefs = NULL, raw = FALSE)
 	n <- lengths(dots)
 	if (any(n != n[1L]))
 	    stop("arguments must have the same length")
-	res <- cbind(1, aPoly)[, 1L +z[, 1]]
+	res <- cbind(1, aPoly)[, 1L +z[, 1], drop=FALSE]
 	## attribute "coefs" = list of coefs from individual variables
 	if (!raw) coefs <- list(attr(aPoly, "coefs"))
 	for (i in 2:nd) {
 	    aPoly <- poly(dots[[i]], degree, raw = raw, simple = raw)
-	    res <- res * cbind(1, aPoly)[, 1L +z[, i]]
+	    res <- res * cbind(1, aPoly)[, 1L +z[, i], drop=FALSE]
 	    if (!raw) coefs <- c(coefs, list(attr(aPoly, "coefs")))
 	}
 	colnames(res) <- apply(z, 1L, function(x) paste(x, collapse = "."))
@@ -174,12 +174,11 @@ polym <- function (..., degree = 1, coefs = NULL, raw = FALSE)
     } else { ## use 'coefs' for prediction
 	newdata <- as.data.frame(dots) # new data
 	if (nd != ncol(newdata))
-	    stop("wrong number of columns in new data: ", deparse(substitute(...)))
-	res <- cbind(1, poly(newdata[[1]], degree=degree,
-			     coefs=coefs[[1]], simple=TRUE))[, 1L +z[, 1]]
-	if(nd > 1) for (i in 2:nd)
+	    stop("wrong number of columns in new data: ", deparse1(substitute(...)))
+	res <- 1
+	for (i in seq_len(nd))
             res <- res*cbind(1, poly(newdata[[i]], degree=degree,
-                                     coefs=coefs[[i]], simple=TRUE))[, 1L +z[, i]]
+				     coefs=coefs[[i]], simple=TRUE))[, 1L +z[, i], drop=FALSE]
 	colnames(res) <- apply(z, 1L, function(x) paste(x, collapse = "."))
         ## no 'coefs' and 'degree', nor "poly" class
 	res
