@@ -1,7 +1,7 @@
 #  File src/library/tools/R/Vignettes.R
 #  Part of the R package, https://www.R-project.org
 #
-#  Copyright (C) 1995-2016 The R Core Team
+#  Copyright (C) 1995-2022 The R Core Team
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -17,7 +17,7 @@
 #  https://www.R-project.org/Licenses/
 
 vignette_is_tex <- function(file, ...) {
-    (regexpr("[.]tex$", file, ignore.case = TRUE) != -1L)
+    endsWith(tolower(file), ".tex")
 }
 
 # Infers the vignette type (PDF or HTML) from the filename of the
@@ -33,13 +33,13 @@ vignette_type <- function(file) {
 }
 
 # Locates the vignette weave, tangle and texi2pdf product(s) based on the
-# vignette name.   All such products must have the name as their filename
+# vignette name.   All such products must have the same name as their filename
 # prefix (i.e. "^<name>").
 # For weave, final = TRUE will look for <name>.pdf and <name>.html, whereas
 # with final = FALSE it also looks for <name>.tex (if <name>.pdf is also
-# found, it will be returned).  For tangle, main = TRUE will look <name>.R,
+# found, it will be returned).  For tangle, main = TRUE will look for <name>.R,
 # whereas main = FALSE will look for <name><anything>*.R.
-# For texipdf, <name>.pdf is located.
+# For texi2pdf, <name>.pdf is located.
 find_vignette_product <-
     function(name, by = c("weave", "tangle", "texi2pdf"),
              final = FALSE, main = TRUE, dir = ".", engine, ...)
@@ -54,8 +54,7 @@ find_vignette_product <-
 
     exts <- c(exts, toupper(exts))
     pattern1 <- sprintf("^%s[.](%s)$", name, paste(exts, collapse = "|"))
-    output0 <- list.files(path = dir, all.files = FALSE, full.names = FALSE,
-                          no.. = TRUE)
+    output0 <- list.files(path = dir, no.. = TRUE)
     output0 <- output0[file_test("-f", file.path(dir, output0))]
     output <- grep(pattern1, output0, value = TRUE)
     # If main is FALSE, we want to find all other files with related
@@ -68,19 +67,24 @@ find_vignette_product <-
 	output2 <- grep(pattern2, output0, value = TRUE)
 	output <- c(output, setdiff(output2, output))
     }
-
+    fmt_file_sizes <- function(files)
+	paste(sprintf("%s (%g bytes)", sQuote(files), file.size(file.path(dir, files))),
+	      collapse=", ")
     if (by == "weave") {
         if (length(output) == 0L)
-            stop(gettextf("Failed to locate the %s output file (by engine %s) for vignette with name %s. The following files exist in directory %s: %s",
+            stop(gettextf("Failed to locate %s output file %s or %s for vignette with name %s and engine %s. The following files exist in working directory %s: %s",
                           sQuote(by),
+                          sQuote(paste0(name, ".pdf")), sQuote(paste0(name, ".html")),
+                          sQuote(name),
                           sQuote(sprintf("%s::%s", engine$package, engine$name)),
-                          sQuote(name), sQuote(dir),
-                          paste(sQuote(output0), collapse=", ")),
+                          sQuote(normalizePath(dir)),
+                          fmt_file_sizes(output0)),
                  domain = NA)
         if (length(output) > 2L || (final && length(output) > 1L))
             stop(gettextf("Located more than one %s output file (by engine %s) for vignette with name %s: %s", sQuote(by),
                           sQuote(sprintf("%s::%s", engine$package, engine$name)),
-                          sQuote(name), paste(sQuote(output), collapse=", ")),
+                          sQuote(name),
+                          fmt_file_sizes(output)),
                  domain  = NA)
 	# If weave produced a TeX and then a PDF without cleaning out
 	# the TeX, consider the newer one (PDF wins a tie) as the weave product
@@ -95,17 +99,20 @@ find_vignette_product <-
             stopifnot(length(output) <= 1L)
     } else if (by == "texi2pdf") {
         if (length(output) == 0L)
-            stop(gettextf("Failed to locate the %s output file (by engine %s) for vignette with name %s. The following files exist in directory %s: %s",
+            stop(gettextf("Failed to locate %s output file %s for vignette with name %s and engine %s. The following files exist in working directory %s: %s",
                           sQuote(by),
+                          sQuote(paste0(name, ".pdf")),
+                          sQuote(name),
                           sQuote(sprintf("%s::%s", engine$package, engine$name)),
-                          sQuote(name), sQuote(dir),
-                          paste(sQuote(output0), collapse=", ")),
+                          sQuote(normalizePath(dir)),
+                          fmt_file_sizes(output0)),
                  domain = NA)
         if (length(output) > 1L)
             stop(gettextf("Located more than one %s output file (by engine %s) for vignette with name %s: %s",
                           sQuote(by),
                           sQuote(sprintf("%s::%s", engine$package, engine$name)),
-                          sQuote(name), paste(sQuote(output), collapse=", ")),
+                          sQuote(name),
+                          fmt_file_sizes(output)),
                  domain = NA)
     }
 
@@ -158,12 +165,13 @@ function(package, dir, lib.loc = NULL,
     result <- list(tangle = list(), weave = list(),
                    source = list(), latex = list())
 
+    ## pkgVignettes has already done this
     loadVignetteBuilder(vigns$pkgdir)
 
     startdir <- getwd()
     for(i in seq_along(vigns$docs)) {
-        file <- vigns$docs[i]
-        file <- basename(file)
+        path <- vigns$docs[i]
+        file <- basename(path)
         name <- vigns$names[i]
     	engine <- vignetteEngine(vigns$engines[i])
 	enc <- vigns$encodings[i]
@@ -174,20 +182,20 @@ function(package, dir, lib.loc = NULL,
             message("  Running ", sQuote(file))
             .eval_with_capture({
                 result$tangle[[file]] <- tryCatch({
-                    engine$tangle(file, quiet = TRUE, encoding = enc)
+                    engine$tangle(path, quiet = TRUE, encoding = enc)
                     setwd(startdir) # in case a vignette changes the working dir
                     find_vignette_product(name, by = "tangle", main = FALSE, engine = engine)
-                }, error = function(e) e)
+                }, error = identity)
             })
         }
         if(weave) {
             setwd(startdir) # in case a vignette changes the working dir then errored out
             .eval_with_capture({
                 result$weave[[file]] <- tryCatch({
-                    engine$weave(file, quiet = TRUE, encoding = enc)
+                    engine$weave(path, quiet = TRUE, encoding = enc)
                     setwd(startdir)
                     find_vignette_product(name, by = "weave", engine = engine)
-                }, error = function(e) e)
+                }, error = identity)
             })
         }
         setwd(startdir) # in case a vignette changes the working dir then errored out
@@ -240,7 +248,7 @@ function(package, dir, lib.loc = NULL,
                 .eval_with_capture({
                     result$source[[file]] <- tryCatch({
                         source(file)
-                    }, error = function(e) e)
+                    }, error = identity)
                 })
                 setwd(startdir)
             }
@@ -248,7 +256,7 @@ function(package, dir, lib.loc = NULL,
     }
 
     if(weave && latex) {
-        if(!("Makefile" %in% list.files(vigns$dir))) {
+        if("Makefile" %notin% list.files(vigns$dir)) {
             ## <NOTE>
             ## This used to run texi2pdf on *all* vignettes, including
             ## the ones already known from the above to give trouble.
@@ -274,7 +282,7 @@ function(package, dir, lib.loc = NULL,
                     result$latex[[file]] <- tryCatch({
                        texi2pdf(file = output, clean = FALSE, quiet = TRUE)
                        find_vignette_product(name, by = "texi2pdf", engine = engine)
-                    }, error = function(e) e)
+                    }, error = identity)
                 })
             }
         }
@@ -289,7 +297,7 @@ function(package, dir, lib.loc = NULL,
           else
               NULL
         })
-        resultsT <- resultsT[!sapply(resultsT, FUN = is.null)]
+        resultsT <- resultsT[!vapply(resultsT, is.null, NA)]
         result[[name]] <- resultsT
     }
 
@@ -298,25 +306,24 @@ function(package, dir, lib.loc = NULL,
     result
 }
 
-print.checkVignettes <-
+format.checkVignettes <-
 function(x, ...)
 {
-    mycat <- function(y, title) {
-        if(length(y)){
-            cat("\n", title, "\n\n", sep = "")
-            for(k in seq_along(y)) {
-                cat("File", names(y)[k], ":\n")
-                cat(as.character(y[[k]]), "\n")
-            }
+    myfmt <- function(y, title) {
+        if(length(y)) {
+            paste(c(paste0("\n", title, "\n"),
+                    unlist(Map(c,
+                               paste0("File ", names(y), ":"),
+                               lapply(y, as.character)),
+                           use.names = FALSE)),
+                  collapse = "\n")
         }
     }
-
-    mycat(x$tangle, "*** Tangle Errors ***")
-    mycat(x$source, "*** Source Errors ***")
-    mycat(x$weave,  "*** Weave Errors ***")
-    mycat(x$latex,  "*** PDFLaTeX Errors ***")
-
-    invisible(x)
+    c(character(),
+      myfmt(x$tangle, "*** Tangle Errors ***"),
+      myfmt(x$source, "*** Source Errors ***"),
+      myfmt(x$weave,  "*** Weave Errors ***"),
+      myfmt(x$latex,  "*** PDFLaTeX Errors ***"))
 }
 
 ### get the engine from a file
@@ -378,6 +385,9 @@ function(package, dir, subdirs = NULL, lib.loc = NULL, output = FALSE,
 
     docs <- names <- engines <- patterns <- character()
     allFiles <- list.files(docdir, all.files = FALSE, full.names = TRUE)
+    exclude <- inRbuildignore(sub(paste0(dir, "/"), "", allFiles, fixed = TRUE), dir)
+    allFiles <- allFiles[!exclude]
+
     matchedPattern <- rep.int(FALSE, length(allFiles))
     msg <- character()
     if (length(allFiles) > 0L) {
@@ -420,14 +430,16 @@ function(package, dir, subdirs = NULL, lib.loc = NULL, output = FALSE,
     defaultEncoding <- .get_package_metadata(dir)["Encoding"]
     encodings <- vapply(docs, getVignetteEncoding, "", default = defaultEncoding)
 
-    z <- list(docs=docs, names=names, engines=engines, patterns=patterns, encodings = encodings,
-	      dir = docdir, pkgdir = dir, msg = msg)
+    z <- list(docs = docs, names = names, engines = engines,
+              patterns = patterns, encodings = encodings,
+              dir = docdir, pkgdir = dir, msg = msg)
 
     if (output) {
         outputs <- character(length(docs))
         for (i in seq_along(docs)) {
             file <- docs[i]
             name <- names[i]
+            engine <- vignetteEngine(engines[i])
             outputI <- find_vignette_product(name, by = "weave", dir = docdir, engine = engine)
             outputs[i] <- outputI
         }
@@ -439,9 +451,32 @@ function(package, dir, subdirs = NULL, lib.loc = NULL, output = FALSE,
         for (i in seq_along(docs)) {
             file <- docs[i]
             name <- names[i]
+            engine <- vignetteEngine(engines[i])
             sourcesI <- find_vignette_product(name, by = "tangle", main = FALSE, dir = docdir, engine = engine)
             sources[[file]] <- sourcesI
         }
+        ## If a package has vignettes 'foo.Rnw' and 'foo-xxx.Rnw' with
+        ## extracted sources 'foo.R' and 'foo-xxx.R', the above will
+        ## give both .R files as sources for 'foo.Rnw', as tangling
+        ## could split into several files and so matching file name
+        ## roots (without extensions) cannot look for exact matches
+        ## only.  However, if there were multiple matches but all
+        ## vignettes sources have an exact match, then we can drop the
+        ## non-exact matches from the multiple matches.
+        ## Ideally, we would teach R CMD build to process one vignette
+        ## source at a time and record the vignette products, and have
+        ## pkgVignettes() use the recorded info if available.
+        if(any(ind <- (lengths(sources) > 1L))) {
+            rootify <- function(s) sub("[.][^.]+$", "", basename(s))
+            dnm <- rootify(names(sources))
+            snm <- lapply(sources, rootify)
+            if(!any(is.na(match(dnm, unlist(snm))))) {
+                for(i in which(ind)) {
+                    sources[[i]] <-
+                        sources[[i]][is.na(match(snm[[i]], dnm[-i]))]
+                }
+            }
+        }
         z$sources <- sources
     }
 
@@ -454,14 +489,19 @@ function(package, dir, subdirs = NULL, lib.loc = NULL, output = FALSE,
 ###
 ### Run a weave and pdflatex on all vignettes of a package and try to
 ### remove all temporary files that were created.
+### Exported version, used in R CMD build/check
 buildVignettes <-
     function(package, dir, lib.loc = NULL, quiet = TRUE, clean = TRUE,
-             tangle = FALSE)
+             tangle = FALSE, ser_elibs = NULL)
 {
+    separate <- !is.null(ser_elibs)
+    if (separate) elibs <- readRDS(ser_elibs)
+    ## This has side effects, including loading vignette-buider pkgs
     vigns <- pkgVignettes(package = package, dir = dir, lib.loc = lib.loc,
                           check = TRUE)
-    if(is.null(vigns)) return(invisible())
-    if(length(vigns$msg))
+    if (is.null(vigns)) return(invisible())
+    if (length(vigns$docs) <= 1L) separate <- FALSE
+    if (length(vigns$msg))
         warning(paste(vigns$msg, collapse = "\n"), domain = NA)
 
     ## Check that duplicated vignette names do not exist, e.g.
@@ -470,16 +510,34 @@ buildVignettes <-
     if (any(dups)) {
         names <- unique(vigns$names[dups])
         docs <- sort(basename(vigns$docs[vigns$names %in% names]))
-        stop(gettextf("Detected vignette source files (%s) with shared names (%s) and therefore risking overwriting each others output files",
+        stop(gettextf("Detected vignette source files (%s) with shared names (%s) and therefore risking overwriting each other's output files",
                       paste(sQuote(docs), collapse = ", "),
                       paste(sQuote(names), collapse = ", ")),
              domain = NA)
     }
 
+    ## Check for duplicated titles (which look silly on CRAN pages)
+    titles <- character()
+    for (d in vigns$docs) {
+        this <- c(.get_vignette_metadata(readLines(d, warn = FALSE),
+                                         "IndexEntry"), "")[1L]
+        titles <- c(titles, this)
+    }
+    have_dup_titles <-
+        if (any(dup <- duplicated(titles))) {
+            dups <- unique(titles[dup])
+            message(ngettext(length(dups),
+                             "duplicated vignette title:",
+                             "duplicated vignette titles:"))
+            message(paste(.pretty_format(dups), collapse = "\n"))
+            message()
+            TRUE
+        } else FALSE
+
     ## unset SWEAVE_STYLEPATH_DEFAULT here to avoid problems
     Sys.unsetenv("SWEAVE_STYLEPATH_DEFAULT")
 
-    op <- options(warn = 1) # we run vignettes in this process
+    op <- options(warn = 1) # we run may run vignettes in this process
     wd <- getwd()
     if (is.null(wd))
         stop("current working directory cannot be ascertained")
@@ -495,55 +553,100 @@ buildVignettes <-
 
     ## Note, as from 2.13.0, only this case
     have.makefile <- "Makefile" %in% origfiles
-    WINDOWS <- .Platform$OS.type == "windows"
 
     file.create(".build.timestamp")
 
-    loadVignetteBuilder(vigns$pkgdir)
-    outputs <- NULL
+    ## pkgVignettes has already done this
+    ## loadVignetteBuilder(vigns$pkgdir)
+    outputs <- character()
     sourceList <- list()
     startdir <- getwd()
+    fails <- character()
     for(i in seq_along(vigns$docs)) {
+        thisOK <- TRUE
         file <- basename(vigns$docs[i])
-        name <- vigns$names[i]
-    	engine <- vignetteEngine(vigns$engines[i])
         enc <- vigns$encodings[i]
-        if (enc == "non-ASCII")
-            stop(gettextf("Vignette '%s' is non-ASCII but has no declared encoding",
-                 file), domain = NA, call. = FALSE)
-
-        output <- tryCatch({
-            ## FIXME: run this in a separate process
-            engine$weave(file, quiet = quiet, encoding = enc)
-            setwd(startdir)
-            find_vignette_product(name, by = "weave", engine = engine)
-        }, error = function(e) {
-            stop(gettextf("processing vignette '%s' failed with diagnostics:\n%s",
-                 file, conditionMessage(e)), domain = NA, call. = FALSE)
-        })
-
-        ## This can fail if run in a directory whose path contains spaces.
-        if(!have.makefile && vignette_is_tex(output)) {
-            texi2pdf(file = output, clean = FALSE, quiet = quiet)
-            output <- find_vignette_product(name, by = "texi2pdf", engine = engine)
+        if (enc == "non-ASCII") {
+            message(gettextf("Error: Vignette '%s' is non-ASCII but has no declared encoding",
+                             file))
+            fails <- c(fails, file)
+            next
         }
-        outputs <- c(outputs, output)
+        name <- vigns$names[i]
+        engine <- vignetteEngine(vigns$engines[i])
 
-        if (tangle) {  # This is set for all engines as of 3.0.2
+        if (separate) {  # --- run in separate process
+            tf2 <- gsub("\\", "/", tempfile(fileext = ".rds"), fixed = TRUE)
+            saveRDS(engine, tf2)
+            Rcmd <- sprintf('tools:::.buildOneVignette("%s", "%s", %s, %s, "%s", "%s", "%s")',
+                            file, vigns$pkgdir, quiet, have.makefile,
+                            name, enc, tf2)
+            tlim <- get_timeout(Sys.getenv("_R_CHECK_ONE_VIGNETTE_ELAPSED_TIMEOUT_",
+                                           Sys.getenv("_R_CHECK_ELAPSED_TIMEOUT_")))
+            tf <- tempfile()
+            status <- R_runR(Rcmd, "--vanilla --no-echo", elibs,
+                             stdout = tf, stderr = tf, timeout = tlim)
+            unlink(tf2)
+            ##print(status)
+            if (!status) {
+                this <- readLines(tf)
+                patt <- "^[+]-[+]"
+                l <- grepl(patt, this)
+                output <- gsub(patt, "", this[l])
+                outputs <- c(outputs, output)
+                cat(this[!l], sep = "\n")
+            } else {
+                fails <- c(fails, file)
+                cat(readLines(tf), sep = "\n")
+            }
+            unlink(tf)
+        } else {  # --- run in this process
+            message(gettextf("--- re-building %s using %s",
+                             sQuote(file), engine$name))
+            tryCatch({
+                engine$weave(file, quiet = quiet, encoding = enc)
+                setwd(startdir) # In case weave/vignette changed it
+                output <- find_vignette_product(name, by = "weave", engine = engine)
+                if (!have.makefile && vignette_is_tex(output)) {
+                    ## This can fail if run in a directory whose path contains spaces.
+                    texi2pdf(file = output, clean = FALSE, quiet = quiet)
+                    output <- find_vignette_product(name, by = "texi2pdf",
+                                                    engine = engine)
+                }
+                outputs <- c(outputs, output)
+            }, error = function(e) {
+                thisOK <<- FALSE
+                fails <<- c(fails, file)
+                message(gettextf("Error: processing vignette '%s' failed with diagnostics:\n%s",
+                                 file, conditionMessage(e)))
+            })
+        }         # end if (separate)
+
+        if (tangle && !separate) {  # This is set for all engines as of 3.0.2
+            ## It is unlikely that weave succeeds but tangle fails,
+            ## so we don't bother to report tangle failures specifically.
             output <- tryCatch({
-                ## FIXME: run this in a separate process
                 engine$tangle(file, quiet = quiet, encoding = enc)
-                setwd(startdir)
+                setwd(startdir) # In case tangle/vignette changed it
                 find_vignette_product(name, by = "tangle", main = FALSE, engine = engine)
             }, error = function(e) {
-                stop(gettextf("tangling vignette '%s' failed with diagnostics:\n%s",
-                     file, conditionMessage(e)), domain = NA, call. = FALSE)
+                thisOK <<- FALSE
+                fails <<- c(fails, file)
+                message(gettextf("Error: tangling vignette '%s' failed with diagnostics:\n%s",
+                     file, conditionMessage(e)))
             })
             sourceList[[file]] <- output
         }
-    }
+        if (!separate) {
+            if (thisOK)
+                message(gettextf("--- finished re-building %s\n", sQuote(file)))
+            else
+                message(gettextf("--- failed re-building %s\n", sQuote(file)))
+        }
+    } # end loop over vignettes
 
-    if(have.makefile) {
+    if (have.makefile) {
+        WINDOWS <- .Platform$OS.type == "windows"
         if (WINDOWS) {
             ## Some people have *assumed* that R_HOME uses / in Makefiles
             ## Spaces in paths might still cause trouble.
@@ -551,11 +654,11 @@ buildVignettes <-
             Sys.setenv(R_HOME = rhome)
         }
     	make <- Sys.getenv("MAKE", "make")
-        if(!nzchar(make)) make <- "make"
+        if (!nzchar(make)) make <- "make"
         yy <- system(make)
-        if(yy > 0) stop("running 'make' failed")
+        if (yy > 0) stop("running 'make' failed")
         ## See if Makefile has a clean: target, and if so run it.
-        if(clean &&
+        if (clean &&
 	   any(startsWith(readLines("Makefile", warn = FALSE), "clean:")))
             system(paste(make, "clean"))
     } else {
@@ -564,7 +667,7 @@ buildVignettes <-
         grDevices::graphics.off()
 
         keep <- c(outputs, unlist(sourceList))
-        if(clean) {
+        if (clean) {
             f <- setdiff(list.files(all.files = TRUE, no.. = TRUE), keep)
             newer <- file_test("-nt", f, ".build.timestamp")
             ## some packages, e.g. SOAR, create directories
@@ -576,16 +679,37 @@ buildVignettes <-
         }
     }
 
-    # Assert
-    stopifnot(length(outputs) == length(vigns$docs))
+    if (file.exists(".build.timestamp")) file.remove(".build.timestamp")
+    ## Might have been in origfiles ...
+
+    if (length(fails)) {
+        message(ngettext(length(fails),
+                         "SUMMARY: processing the following file failed:",
+                         "SUMMARY: processing the following files failed:"))
+        message(paste(.pretty_format(fails), collapse = "\n"))
+        message()
+    }
+
+
+    msg2 <- paste("Duplicate vignette titles.",
+                  "  Ensure that the %\\VignetteIndexEntry lines in the vignette sources",
+                  "  correspond to the vignette titles.",
+                  sep = "\n")
+
+    ## Assert
+    if (length(fails) || (length(outputs) != length(vigns$docs))) {
+        msg <- "Vignette re-building failed."
+        if (have_dup_titles) msg <- paste0(msg, "\nError: ", msg2)
+        stop(msg, domain = NA, call. = FALSE)
+    }
+
+    if (have_dup_titles)
+        stop(msg2, domain = NA, call. = FALSE)
 
     vigns$outputs <- outputs
     vigns$sources <- sourceList
 
-    if(file.exists(".build.timestamp")) file.remove(".build.timestamp")
-    ## Might have been in origfiles ...
-
-    invisible(vigns)
+    invisible(vigns) ## not documented on the help page.
 }
 
 ### * buildVignette
@@ -606,7 +730,7 @@ buildVignette <-
 
     if (!is.null(buildPkg))
 	for (pkg in buildPkg)
-	    loadNamespace(pkg)
+	    suppressPackageStartupMessages(loadNamespace(pkg))
 
     if (is.null(engine))
     # Infer vignette engine from vignette content
@@ -634,11 +758,9 @@ buildVignette <-
     olddir <- setwd(dir)
     if (!is.null(olddir)) on.exit(setwd(olddir))
 
-    ## # Record existing files
-    ## origfiles <- list.files(all.files = TRUE)
-    if (is.na(clean) || clean) {
-	file.create(".build.timestamp")
-    }
+    # Record existing files (not to be cleaned)
+    if (is.na(clean) || clean)
+        origfiles <- list.files(all.files = TRUE)
 
     tdir <- getwd()# if 'dir' was relative, resetting to tdir will work
     output <- NULL
@@ -671,24 +793,61 @@ buildVignette <-
 	clean <- TRUE
     }
     if (clean) {
-	f <- setdiff(list.files(all.files = TRUE, no.. = TRUE), keep)
-	newer <- file_test("-nt", f, ".build.timestamp")
+	f <- setdiff(list.files(all.files = TRUE, no.. = TRUE),
+                     c(keep, origfiles))
 	## some packages create directories
-	unlink(f[newer], recursive = TRUE)
-    }
-    ### huh?  2nd round of cleaning even if  clean is FALSE ??
-    ##     f <- setdiff(list.files(all.files = TRUE, no.. = TRUE), c(keep, origfiles))
-    ##     f <- f[file_test("-f", f)]
-    ##     file.remove(f)
-    ## #}
-
-    if((is.na(clean) || clean) && file.exists(".build.timestamp")) {
-        file.remove(".build.timestamp")
+	unlink(f, recursive = TRUE)
     }
 
     unique(keep)
 }
 
+### * .buildOneVignette
+
+## helper to be run in a separate process
+.buildOneVignette <-
+    function(file, pkgdir, quiet = TRUE, have.makefile = FALSE,
+             name, enc, ser_engine)
+{
+    op <- options(warn = 1)      # we run vignettes in this process
+    engine <- readRDS(ser_engine)
+    ## This is more than we need, but we cannot be sure that 'engine'
+    ## has references to all the namespaces required.  But packages
+    ## typically only have one engine, at most two.
+    loadVignetteBuilder(pkgdir)
+    OK <- TRUE
+    startdir <- getwd()
+    output <- character()
+
+    message(gettextf("--- re-building %s using %s",
+                     sQuote(file), engine$name))
+    tryCatch({
+        engine$weave(file, quiet = quiet, encoding = enc)
+        setwd(startdir)  # In case weave/vignette changed it
+        output <- find_vignette_product(name, by = "weave", engine = engine)
+        if(!have.makefile && vignette_is_tex(output)) {
+            texi2pdf(file = output, clean = FALSE, quiet = quiet)
+            output <- find_vignette_product(name, by = "texi2pdf",
+                                            engine = engine)
+        }
+    }, error = function(e) {
+        OK <<- FALSE
+        message(gettextf("Error: processing vignette '%s' failed with diagnostics:\n%s",
+                         file, conditionMessage(e)))
+    })
+
+    if (OK)
+        message(gettextf("--- finished re-building %s\n", sQuote(file)))
+    else {
+        message(gettextf("--- failed re-building %s\n", sQuote(file)))
+        q("no", status = 9L)
+    }
+
+    message("+-+", output)
+    invisible(output)
+}
+
+
 ### * getVignetteEncoding
 
 getVignetteEncoding <-  function(file, ...)
@@ -765,8 +924,7 @@ function(lines, tag)
     trimws(gsub(meta_RE, "\\1", meta))
 }
 
-vignetteInfo <-
-function(file)
+vignetteInfo <- function(file)
 {
     lines <- readLines(file, warn = FALSE)
 
@@ -929,27 +1087,10 @@ function(pkg, con, vignetteIndex = NULL)
                          iconv(urls, "", "UTF-8"), "</span></td></tr>"),
                   "</table>")
     }
-    html <- c(html, "</body></html>")
+    html <- c(html, "</div></body></html>")
     writeLines(html, con=con)
 }
 
-vignetteDepends <-
-    function(vignette, recursive = TRUE, reduce = TRUE,
-             local = TRUE, lib.loc = NULL)
-{
-    if (length(vignette) != 1L)
-        stop("argument 'vignette' must be of length 1")
-    if (!nzchar(vignette)) return(invisible()) # lets examples work.
-    if (!file.exists(vignette))
-        stop(gettextf("file '%s' not found", vignette), domain = NA)
-
-    vigDeps <- vignetteInfo(vignette)$depends
-
-    depMtrx <- getVigDepMtrx(vigDeps)
-    instPkgs <- utils::installed.packages(lib.loc=lib.loc)
-    getDepList(depMtrx, instPkgs, recursive, local, reduce)
-}
-
 getVigDepMtrx <-
 function(vigDeps)
 {
@@ -1069,7 +1210,7 @@ vignetteEngine <- local({
                 } else {
                     key <- engineKey(name)
                 }
-		loadNamespace(key[1])
+		suppressPackageStartupMessages(loadNamespace(key[1]))
                 name <- paste(key, collapse = "::")
                 result <- registry[[name]]
                 if (is.null(result))
@@ -1078,7 +1219,8 @@ vignetteEngine <- local({
             } else {
                 for (pkg in package) {
                     key <- engineKey(name, pkg)
-		    try(loadNamespace(key[1]), silent = TRUE)
+		    try(suppressPackageStartupMessages(loadNamespace(key[1])),
+                        silent = TRUE)
                     nameT <- paste(key, collapse = "::")
                     result <- registry[[nameT]]
                     if (!is.null(result))
@@ -1180,7 +1322,8 @@ function(pkgdir, mustwork = TRUE)
     pkgs <- unique(c(pkgs, "utils"))
 
     for (pkg in pkgs) {
-	res <- tryCatch(loadNamespace(pkg), error = function(e)e)
+	res <- tryCatch(suppressPackageStartupMessages(loadNamespace(pkg)),
+                        error = identity)
 	if (mustwork && inherits(res, "error"))
             stop(gettextf("vignette builder '%s' not found", pkg), domain = NA)
     }
@@ -1191,11 +1334,13 @@ function(pkgdir, mustwork = TRUE)
 
 getVignetteInfo <- function(package = NULL, lib.loc = NULL, all = TRUE)
 {
-    if (is.null(package)) {
-        package <- .packages(all.available = all, lib.loc)
-        ## allow for misnamed dirs
-        paths <- find.package(package, lib.loc, quiet = TRUE)
-    } else paths <- find.package(package, lib.loc)
+    paths <-
+        if (is.null(package)) {
+            package <- .packages(all.available = all, lib.loc)
+            ## allow for misnamed dirs
+            find.package(package, lib.loc, quiet = TRUE)
+        } else
+            find.package(package, lib.loc)
 
     ## Find the directories with a 'doc' subdirectory *possibly*
     ## containing vignettes.
@@ -1216,7 +1361,7 @@ getVignetteInfo <- function(package = NULL, lib.loc = NULL, all = TRUE)
             entries <- readRDS(INDEX)
         if (NROW(entries) > 0) {
             # FIXME:  this test is unnecessary?
-            R <- if (is.null(entries$R)) rep("", NROW(entries)) else entries$R
+            R <- if (is.null(entries$R)) rep.int("", NROW(entries)) else entries$R
             file <- basename(entries$File)
             pdf <- entries$PDF
             topic <- file_path_sans_ext(ifelse(R == "", ifelse(pdf == "", file, pdf), R))
