@@ -17,6 +17,87 @@
 #  https://www.R-project.org/Licenses/
 
 
+## Helper function used to declare mime-type for files served by
+## dynamic help, and for base64-encoded files embedded in example
+## output (see code2html.R).
+
+mime_type <- function(path, ext = NULL)
+{
+    stopifnot(length(path) == 1L)
+    if (missing(ext)) ext <- file_ext(path)
+    switch(ext,
+           "css" = "text/css",
+           "js" = "text/javascript",  # for katex etc
+           "sgml" = "text/sgml",    # in RGtk2
+           "xml" = "text/xml",      # in RCurl (RFC 7303 recommends "application/xml") 
+           "html" = "text/html",
+           "htm" = "text/html",
+           "xhtml" = "application/xhtml+xml",
+           "php" = "application/x-httpd-php",
+           "epub" = "application/epub+zip",
+           "csv" = "text/csv",
+           "json" = "application/json",
+           "jsonld"  = "application/ld+json",
+           "mjs" = "text/javascript",
+
+           ## common <img> types (see https://developer.mozilla.org/en-US/docs/Web/Media/Formats/Image_types)
+           "gif" = "image/gif",     # in R2HTML
+           "jpg" = "image/jpeg",
+           "jpeg" = "image/jpeg",
+           "png" = "image/png",
+           "svg" = "image/svg+xml",
+           "apng" = "image/apng",
+           "avif" = "image/avif",
+           "webp" = "image/webp",
+           "bmp" = "image/bmp",
+           "ico" = "image/x-icon",
+           "tiff" = "image/tiff",
+           "tif" = "image/tiff",
+
+           "pdf" = "application/pdf",
+           "eps" =,
+           "ps" = "application/postscript", # in GLMMGibbs, mclust
+
+           ## fonts
+           "eot" = "application/vnd.ms-fontobject",
+           "otf" = "font/otf",
+           "ttf" = "font/ttf",
+           "woff" = "font/woff",
+           "woff2" = "font/woff2",
+
+           ## media
+           "aac" = "audio/aac",
+           "avi" = "video/x-msvideo",
+           "cda" = "application/x-cdf",
+           "mid" = "audio/x-midi",
+           "midi" = "audio/x-midi",
+           "mp3" = "audio/mpeg",
+           "mp4" = "video/mp4",
+           "mpeg" = "video/mpeg",
+           "oga" = "audio/ogg",
+           "ogv" = "video/ogg",
+           "ogx" = "application/ogg",
+           "opus" = "audio/opus",
+           "3gp" = "video/3gpp",
+           "3g2" = "video/3gpp2",
+           "wav" = "audio/wav",
+           "weba" = "audio/webm",
+           "webm" = "video/webm",
+
+           ## archive / compression
+           "bz" = "application/x-bzip",
+           "bz2" = "application/x-bzip2",
+           "gz" = "application/gzip",
+           "rar" = "application/vnd.rar",
+           "zip" = "application/zip",
+           "7z" = "application/x-7z-compressed",
+           "tar" = "application/x-tar",
+
+           ## default
+           "text/plain")
+}
+
+
 ## This may be asked for
 ##  R.css, favicon.ico
 ##  searches with path = "/doc/html/Search"
@@ -27,6 +108,16 @@
 ##             or by file, /library/<pkg>/html/<file>.html
 httpd <- function(path, query, ...)
 {
+    logHelpRequests <-
+        config_val_to_logical(Sys.getenv("_R_HTTPD_LOG_MESSAGES_", "FALSE"))
+    if (logHelpRequests) {
+        message(sprintf("HTTPD-REQUEST %s%s", path,
+                        if (is.null(query)) ""
+                        else { # query is a named chr vector 
+                            paste(paste(names(query), query, sep = "="),
+                                  collapse = ",")
+                        }))
+    }
     linksToTopics <-
         config_val_to_logical(Sys.getenv("_R_HELP_LINKS_TO_TOPICS_", "TRUE"))
     .HTMLdirListing <- function(dir, base, up) {
@@ -168,11 +259,11 @@ httpd <- function(path, query, ...)
             c(HTMLheader("Help search concepts"),
               c("",
                 "<table>",
-                "<tr><th style=\"text-align: left\">Concept</th><th>Frequency</th><th>Packages</th><tr>",
+                "<tr><th style=\"text-align: left\">Concept</th><th>Frequency</th><th>Packages</th></tr>",
                 paste0("<tr><td>",
                        "<a href=\"/doc/html/Search?pattern=",
                        utils::URLencode(reQuote(s), reserved = TRUE),
-                       "&fields.concept=1&agrep=0\">",
+                       "&amp;fields.concept=1&amp;agrep=0\">",
                        shtmlify(substr(s, 1L, 80L)),
                        "</a>",
                        "</td><td style=\"text-align: right\">",
@@ -192,7 +283,7 @@ httpd <- function(path, query, ...)
             c(HTMLheader("Help search keywords"),
               c("",
                 "<table>",
-                "<tr><th style=\"text-align: left\">Keyword</th><th style=\"text-align: left\">Concept</th><th>Frequency</th><th>Packages</th><tr>",
+                "<tr><th style=\"text-align: left\">Keyword</th><th style=\"text-align: left\">Concept</th><th>Frequency</th><th>Packages</th></tr>",
                 paste0("<tr><td>",
                        "<a href=\"/doc/html/Search?category=",
                        keywords$Keyword,
@@ -228,24 +319,6 @@ httpd <- function(path, query, ...)
         list(file = file)
     }
 
-    mime_type <- function(path) {
-        ext <- strsplit(path, ".", fixed = TRUE)[[1L]]
-        if(n <- length(ext)) ext <- ext[n] else ""
-        switch(ext,
-               "css" = "text/css",
-               "gif" = "image/gif",     # in R2HTML
-               "jpg" = "image/jpeg",
-               "png" = "image/png",
-               "svg" = "image/svg+xml",
-               "html" = "text/html",
-               "pdf" = "application/pdf",
-               "eps" =,
-               "ps" = "application/postscript", # in GLMMGibbs, mclust
-               "sgml" = "text/sgml",    # in RGtk2
-               "xml" = "text/xml",      # in RCurl
-               "text/plain")
-    }
-
     charsetSetting <- function(pkg) {
     	encoding <- read.dcf(system.file("DESCRIPTION", package=pkg),
                              "Encoding")
@@ -260,16 +333,21 @@ httpd <- function(path, query, ...)
     mono <- function(text)
         paste0('<span class="samp">', text, "</span>")
 
-    error_page <- function(msg)
+    error_page <- function(msg) {
+        if (logHelpRequests) {
+            message(sprintf("HTTPD-ERROR %s %s", path, paste(msg, collapse = " ")))
+        }
         list(payload =
-             paste0(HTMLheader("httpd error"), msg, "\n</div></body></html>"))
-
+             paste(c(HTMLheader("httpd error"), msg, "\n</div></body></html>"), collapse = "\n"))
+    }
+        
     cssRegexp <- "^/library/([^/]*)/html/R.css$"
     if (grepl("R\\.css$", path) && !grepl(cssRegexp, path))
         return(list(file = file.path(R.home("doc"), "html", "R.css"),
                     "content-type" = "text/css"))
     else if(path == "/favicon.ico")
-        return(list(file = file.path(R.home("doc"), "html", "favicon.ico")))
+        return(list(file = file.path(R.home("doc"), "html", "favicon.ico"),
+                    "content-type" = "image/x-icon"))
     else if(path == "/NEWS")
          return(list(file = file.path(R.home("doc"), "html", "NEWS.html"),
                      "content-type" = "text/html"))
@@ -296,11 +374,12 @@ httpd <- function(path, query, ...)
     ## ----------------------- per-package documentation ---------------------
     ## seems we got ../..//<pkg> in the past
     fileRegexp <- "^/library/+([^/]*)/html/([^/]*)\\.html$"
-    topicRegexp <- "^/library/+([^/]*)/help/([^/]*)$"
+    topicRegexp <- "^/library/+([^/]*)/help/(.*)$"
     docRegexp <- "^/library/([^/]*)/doc(.*)"
     demoRegexp <- "^/library/([^/]*)/demo$"
     demosRegexp <- "^/library/([^/]*)/demo/([^/]*)$"
     DemoRegexp <- "^/library/([^/]*)/Demo/([^/]*)$"
+    ExampleRegexp <- "^/library/([^/]*)/Example/([^/]*)$"
     newsRegexp <- "^/library/([^/]*)/NEWS$"
     figureRegexp <- "^/library/([^/]*)/(help|html)/figures/([^/]*)$"
     sessionRegexp <- "^/session/"
@@ -311,8 +390,19 @@ httpd <- function(path, query, ...)
     	pkg <- sub(topicRegexp, "\\1", path)
     	if (pkg == "NULL") pkg <- NULL  # There were multiple hits in the console
     	topic <- sub(topicRegexp, "\\2", path)
-        ## if a package is specified, look there first, then everywhere
+        ## If a package is specified, look there first. If not found,
+        ## search in other packages. This is used to search for
+        ## off-package links where the target package is not specified
+        ## (they are nominally links to topics in the same package)
+
+        ## However, if pkg is specified but not installed, give an
+        ## error message.
     	if (!is.null(pkg)) { # () avoids deparse here
+            if (!nzchar(system.file(package = pkg))) {
+                msg <- gettextf("No package named %s could be found",
+                                mono(pkg))
+                return(error_page(msg))
+            }
     	    file <- utils::help(topic, package = (pkg), help_type = "text")
             ## Before searching other packages, check if topic.Rd is
             ## available as a file in the package.
@@ -331,7 +421,7 @@ httpd <- function(path, query, ...)
 	if (!length(file)) {
             msg <- gettextf("No help found for topic %s in any package.",
                             mono(topic))
-	    return(list(payload = error_page(msg)))
+	    return(error_page(msg))
 	} else if (length(file) == 1L) {
 	    path <- dirname(dirname(file))
 	    file <- paste0('../../', basename(path), '/html/',
@@ -367,7 +457,14 @@ httpd <- function(path, query, ...)
                                collapse = "\n")
 
             return(list(payload =
-                        paste0("<p>",
+                        paste0("<!DOCTYPE html>",
+                               "<html>",
+                               "<head>",
+                               "<title>R: help</title>",
+                               "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" />",
+                               "</head>",
+                               "<body>",
+                                "<p>",
                                ## for languages with multiple plurals ....
                                sprintf(ngettext(length(paths),
                                                 "Help on topic '%s' was found in the following package:",
@@ -375,6 +472,8 @@ httpd <- function(path, query, ...)
                                                 ), topic),
                                "</p><dl>\n",
                                packages, "</dl>",
+                               "</body>",
+                               "</html>",
                                collapse = "\n")
                         ))
         }
@@ -498,12 +597,19 @@ httpd <- function(path, query, ...)
     } else if (grepl(DemoRegexp, path)) {
     	pkg <- sub(DemoRegexp, "\\1", path)
     	demo <- sub(DemoRegexp, "\\2", path)
-    	demo(demo, package=pkg, character.only=TRUE, ask=FALSE)
-	return( list(payload = paste0("Demo '", pkg, "::", demo,
-				"' was run in the console.",
-				" To repeat, type 'demo(",
-				pkg, "::", demo,
-				")' in the console.")) )
+        if (logHelpRequests) {
+            message(sprintf("HTTPD-DEMO %s::%s", pkg, demo))
+        }
+        else return(demo2html(demo, pkg))
+    } else if (grepl(ExampleRegexp, path)) {
+    	pkg <- sub(ExampleRegexp, "\\1", path)
+    	topic <- sub(ExampleRegexp, "\\2", path)
+        if (logHelpRequests) {
+            message(sprintf("HTTPD-EXAMPLE %s::%s", pkg, topic))
+        }
+        else return(example2html(topic, pkg,
+                                 env = if (identical(query["local"], "FALSE")) .GlobalEnv
+                                       else NULL))
     } else if (grepl(newsRegexp, path)) {
     	pkg <- sub(newsRegexp, "\\1", path)
         if(identical(names(query), c("objects", "port")))
