@@ -12,15 +12,16 @@ function(title="R", logo=TRUE,
          headerTitle = paste("R:", title),
          outputEncoding = "UTF-8")
 {
-    result <- c('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">',
-        '<html xmlns="http://www.w3.org/1999/xhtml">',
-	paste0('<head><title>', headerTitle, '</title>'),
-	paste0('<meta http-equiv="Content-Type" content="text/html; charset=',
-	       mime_canonical_encoding(outputEncoding), '" />'),
-        '<meta name="viewport" content="width=device-width, initial-scale=1.0, user-scalable=yes" />',
-	paste0('<link rel="stylesheet" type="text/css" href="', css, '" />'),
-	'</head><body><div class="container">',
-	paste('<h1>', title))
+    result <-
+        c('<!DOCTYPE html>',
+          "<html>",
+          paste0('<head><title>', headerTitle, '</title>'),
+          paste0('<meta http-equiv="Content-Type" content="text/html; charset=',
+                 mime_canonical_encoding(outputEncoding), '" />'),
+          '<meta name="viewport" content="width=device-width, initial-scale=1.0, user-scalable=yes" />',
+          paste0('<link rel="stylesheet" type="text/css" href="', css, '" />'),
+          '</head><body><div class="container">',
+          paste('<h1>', title))
     if (logo)
     	result <- c(result,
                     paste0('<img class="toplogo" src="',
@@ -141,7 +142,7 @@ function(x, ...)
                                   cchunks))
                         } else {
                             c(vheaders[i],
-                              print_items(vchunk$Text))
+                              print_items(vchunk$HTML))
                         }
                     })
              ),
@@ -235,7 +236,7 @@ makeDemoTable <- function(demos, depth=2) {
 	    link <- c('<a href="', root, 'demo/', file, '">',
 			  pkg, "::", topic, '</a>')
 	    runlink <- c(' <a href="', root, 'Demo/', topic,
-	                 '">(Run demo in console)</a>')
+	                 '">(Run demo)</a>')
 	} else {
 	    link <- c(pkg, "::", topic)
 	    runlink <- ""
@@ -255,10 +256,13 @@ makeHelpTable <- function(help, depth=2) {
              '<col style="width:  2%;" />',
              '<col style="width: 74%;" />')
     pkg <- help[, "Package"]
-    root <- paste0(strrep("../", depth), "library/", pkg, "/html/")
+    ## Target could be ../library/pkg/help/topic or ../library/pkg/html/filename.html
+    ## We only have topic, so can only do the former. Topics may contain
+    ## special characters, so need to be encoded.
+    root <- paste0(strrep("../", depth), "library/", pkg, "/help/")
     topic <- help[, "Topic"]
     Title <- help[, "Title"]
-    links <- paste0('<a href="', root, topic, '.html">',
+    links <- paste0('<a href="', root, topic2url(topic), '">',
 		    ifelse(nchar(pkg), paste0(pkg, "::"), ""),
 		    topic, '</a>')
     lines <- paste0('<tr><td style="text-align: right; vertical-align: top;">', links,
@@ -376,8 +380,8 @@ function(x, header = TRUE, ...)
                   "<meta http-equiv=\"Content-Type\" content=\"text/html; charset=utf-8\" />",
                   '<meta name="viewport" content="width=device-width, initial-scale=1.0, user-scalable=yes" />',
                   "</head>")
-        header <- c("<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">",
-                  "<html xmlns=\"http://www.w3.org/1999/xhtml\">",
+        header <- c("<!DOCTYPE html>",
+                  "<html>",
                   header,
 		  '<body><div class="container">')
 	footer <- c("</div></body>", "</html>")
