@@ -1,7 +1,7 @@
 #  File src/library/utils/R/news.R
-#  Part of the R package, http://www.R-project.org
+#  Part of the R package, https://www.R-project.org
 #
-#  Copyright (C) 1995-2014 The R Core Team
+#  Copyright (C) 1995-2020 The R Core Team
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -14,23 +14,30 @@
 #  GNU General Public License for more details.
 #
 #  A copy of the GNU General Public License is available at
-#  http://www.r-project.org/Licenses/
+#  https://www.R-project.org/Licenses/
 
 
 news <-
 function(query, package = "R", lib.loc = NULL,
          format = NULL, reader = NULL, db = NULL)
 {
-    if(is.null(db)) {
+    if(new.db <- is.null(db)) {
+        ## we could allow vector 'package' here by rbind-ing,
+        ## preserving the classes.
         db <- if(package == "R")
-            tools:::.build_news_db_from_R_NEWS_Rd()
-        else
-            tools:::.build_news_db(package, lib.loc, format, reader)
+                  tools:::.build_news_db_from_R_NEWS_Rd()
+              else if (package == "R-3")
+                  tools:::.build_news_db_from_R_NEWS_Rd(Rfile = "NEWS.3.rds")
+              else if (package == "R-2")
+                  tools:::.build_news_db_from_R_NEWS_Rd(Rfile = "NEWS.2.rds")
+              else
+                  tools:::.build_news_db(package, lib.loc, format, reader)
     }
     if(is.null(db))
-        return(invisible())
+        return(NULL)
 
-    attr(db, "package") <- package
+    if(new.db)
+	attr(db, "package") <- package
 
     ## Is there a way to directly call/use subset.data.frame?
     ## E.g.,
@@ -45,9 +52,6 @@ function(query, package = "R", lib.loc = NULL,
     ## versions (e.g., R NEWS has "2.8.1 patched") or could have the
     ## version info missing (and package_version() does not like NAs).
 
-    has_bad_attr <-
-        !is.null(bad <- attr(db, "bad")) && (length(bad) == NROW(db))
-
     ## Manipulate fields for querying (but return the original ones).
     db1 <- db
     ## Canonicalize version entries which *start* with a valid numeric
@@ -64,24 +68,23 @@ function(query, package = "R", lib.loc = NULL,
 
     r <- eval(substitute(query), db1, parent.frame())
     ## Do something if this is not logical ...
-    if(is.null(r))
-        return(db)
-    else if(!is.logical(r) || length(r) != length(version))
-        stop("invalid query")
-    r <- r & !is.na(r)
-    if(has_bad_attr)
-        structure(db[r, ], bad = bad[r])
-    else
-        db[r, ]
+    if(!is.null(r)) {
+	if(!is.logical(r) || length(r) != length(version))
+	    stop("invalid query")
+	r <- r & !is.na(r)
+	db <- db[r, ]
+        ## This should no longer be necessary ...?
+	if(!all(r))
+	    attr(db, "subset") <- r
+    }
+
+    db
 }
 
 format.news_db <-
 function(x, ...)
 {
-    if(inherits(x, "news_db_from_Rd") ||
-       (!(is.null(bad <- attr(x, "bad")))
-        && (length(bad) == NROW(x))
-        && all(!bad))) {
+    if(tools:::.news_db_has_no_bad_entries(x)) {
 
         ## Format news in the preferred input format:
         ##   Changes in $VERSION [($DATE)]:
@@ -91,6 +94,12 @@ function(x, ...)
         ## Add support for DATE.
         ## </FIXME>
 
+        ## When formatting version and category headers, mimic the HTML
+        ## and legacy R layouts: center the former, and left-justify the
+        ## latter.  (Alternatively, we could e.g. left-justify the
+        ## former and also the latter with an extra indent of 2, but it
+        ## seems preferable to be consistent.)
+
         vchunks <- split(x, x$Version)
         ## Re-order according to decreasing version.
         ## R NEWS has invalid "versions" such as "R-devel" and
@@ -109,14 +118,16 @@ function(x, ...)
             ind <- vstrings != "R-devel"
             vstrings[ind] <- sprintf("version %s", vstrings[ind])
             vheaders <-
-                sprintf("Changes in %s%s:",
-                        vstrings,
-                        ifelse(is.na(dates), "",
-                               sprintf(" (%s)", dates)))
+                format(sprintf("Changes in %s%s",
+                               vstrings,
+                               ifelse(is.na(dates), "",
+                                      sprintf(" (%s)", dates))),
+                       justify = "centre", width = 72L)
+            ## No trailing colon when centering.
         } else vheaders <- character()
 
         format_items <- function(x)
-            paste0("    o   ", gsub("\n", "\n\t", x$Text))
+            paste0("    o   ", gsub("\n", "\n\t", x$Text, fixed=TRUE))
         format_vchunk <- function(vchunk) {
             if(all(!is.na(category <- vchunk$Category)
                    & nzchar(category))) {
@@ -124,7 +135,8 @@ function(x, ...)
                 cchunks <-
                     split(vchunk,
                           factor(category, levels = unique(category)))
-                Map(c, names(cchunks), lapply(cchunks, format_items),
+                cheaders <- names(cchunks)
+                Map(c, cheaders, lapply(cchunks, format_items),
                     USE.NAMES = FALSE)
             } else {
                 format_items(vchunk)
@@ -145,8 +157,49 @@ function(x, ...)
 }
 
 print.news_db <-
-function(x, ...)
+function(x, doBrowse = interactive(), browser = getOption("browser"), ...)
 {
-    writeLines(paste(unlist(format(x, ...)), collapse = "\n\n"))
+    port <- if(doBrowse && !identical("false", browser) &&
+               is.character(pkg <- attr(x, "package")) &&
+               tools:::.news_db_has_no_bad_entries(x))
+        tools::startDynamicHelp(NA) else 0L
+    if (port > 0L) {
+        tools:::.httpd_objects(port, x)
+        url <- if (pkg == "R") {
+            if(is.null(attr(x, "subset"))) {
+                ## Use the pre-built NEWS.html.
+                sprintf("http://127.0.0.1:%d/doc/html/NEWS.html",
+                        port)
+            } else
+                sprintf("http://127.0.0.1:%d/doc/html/NEWS.html?objects=1&port=%d",
+                        port, port)
+        } else
+            sprintf("http://127.0.0.1:%d/library/%s/NEWS?objects=1&port=%d",
+                    port, pkg, port)
+    	## if (!is.null(subset <- attr(x, "subset"))) {
+	##     # Subsets are typically ranges of dates or version numbers, so we run-length encode
+	##     # the subset vector.  We put TRUE in front so the values alternate TRUE, FALSE, ... .
+    	##     rle <- paste(rle(c(TRUE, subset))$lengths, collapse="_")
+    	##     url <- paste0(url, "?subset=", rle)
+        ## }
+    	browseURL(url)
+    } else ## simply show in console:
+	writeLines(paste(unlist(format(x, ...)), collapse = "\n\n"))
     invisible(x)
 }
+
+`[.news_db` <- function(x, i, j, drop) {
+    ## Ensure that 'bad' attribute is subscripted as necessary.
+    y <- NextMethod()
+    if(inherits(y, "news_db")
+       && !missing(i)
+       && !is.null(bad <- attr(x, "bad"))) {
+        attr(y, "bad") <- bad[i]
+    }
+    y
+}
+
+subset.news_db <-
+function(x, subset, ...) {
+    do.call(news, list(substitute(subset), db = x))
+}
