@@ -1,7 +1,7 @@
 #  File src/library/utils/R/packages.R
 #  Part of the R package, https://www.R-project.org
 #
-#  Copyright (C) 1995-2020 The R Core Team
+#  Copyright (C) 1995-2021 The R Core Team
 #
 #  This program is free software; you can redistribute it and/or modify
 #  it under the terms of the GNU General Public License as published by
@@ -23,6 +23,8 @@ function(contriburl = contrib.url(repos, type), method,
          ignore_repo_cache = FALSE, max_repo_cache_age,
          quiet = TRUE, ...)
 {
+    if (!is.character(type))
+        stop("invalid 'type'; must be a character string")
     requiredFields <-
         c(tools:::.get_standard_repository_db_fields(), "File")
     if (is.null(fields))
@@ -166,7 +168,7 @@ function(contriburl = contrib.url(repos, type), method,
             res0 <- cbind(res0[, fields, drop = FALSE], Repository = rp)
             res <- rbind(res, res0, deparse.level = 0L)
         }
-    }
+    } ## end  for(repos in *)
 
     if(!length(res)) return(res)
 
@@ -358,6 +360,8 @@ update.packages <- function(lib.loc = NULL, repos = getOption("repos"),
                             available = NULL, oldPkgs = NULL, ...,
                             checkBuilt = FALSE, type = getOption("pkgType"))
 {
+    if (!is.character(type))
+        stop("invalid 'type'; must be a character string")
     force(ask)  # just a check that it is valid before we start work
     text.select <- function(old)
     {
@@ -465,6 +469,8 @@ old.packages <- function(lib.loc = NULL, repos = getOption("repos"),
                          method, available = NULL, checkBuilt = FALSE,
                          ..., type = getOption("pkgType"))
 {
+    if (!is.character(type))
+        stop("invalid 'type'; must be a character string")
     if(is.null(lib.loc))
         lib.loc <- .libPaths()
     if(!missing(instPkgs)) {
@@ -515,6 +521,8 @@ new.packages <- function(lib.loc = NULL, repos = getOption("repos"),
                          method, available = NULL, ask = FALSE,
                          ..., type = getOption("pkgType"))
 {
+    if (!is.character(type))
+        stop("invalid 'type'; must be a character string")
     ask  # just a check that it is valid before we start work
     if(type == "both" && (!missing(contriburl) || !is.null(available))) {
         stop("specifying 'contriburl' or 'available' requires a single type, not type = \"both\"")
@@ -685,6 +693,18 @@ installed.packages <-
 
 remove.packages <- function(pkgs, lib)
 {
+    if(!length(pkgs)) return(invisible())
+    base <- vapply(pkgs, isBasePkg, FALSE)
+    if(any(base)) {
+        (if(all(base)) stop else warning)(
+            sprintf(ngettext(sum(base),
+                                 "package %s is a base package, and cannot be removed",
+                                 "packages %s are base packages, and cannot be removed"),
+                        paste(sQuote(pkgs[base]), collapse = ", ")),
+            domain = NA)
+        pkgs <- pkgs[!base]
+    }
+
     updateIndices <- function(lib) {
         ## This matches what install.packages() does
         if(lib == .Library && .Platform$OS.type == "unix") {
@@ -701,7 +721,6 @@ remove.packages <- function(pkgs, lib)
         }
     }
 
-    if(!length(pkgs)) return(invisible())
 
     if(missing(lib) || is.null(lib)) {
         lib <- .libPaths()[1L]
@@ -724,6 +743,8 @@ download.packages <- function(pkgs, destdir, available = NULL,
                               contriburl = contrib.url(repos, type),
                               method, type = getOption("pkgType"), ...)
 {
+    if (!is.character(type))
+        stop("invalid 'type'; must be a character string")
     nonlocalcran <- !all(startsWith(contriburl, "file:"))
     if(nonlocalcran && !dir.exists(destdir))
         stop("'destdir' is not a directory")
@@ -756,7 +777,8 @@ download.packages <- function(pkgs, destdir, available = NULL,
                          switch(type,
                                 "source" = ".tar.gz",
                                 "mac.binary" = ".tgz",
-                                "win.binary" = ".zip"))
+                                "win.binary" = ".zip",
+                                stop("invalid 'type'")))
             have_fn <- !is.na(File)
             fn[have_fn] <- File[have_fn]
             repos <- available[ok, "Repository"]
@@ -807,6 +829,8 @@ resolvePkgType <- function(type) {
 
 contrib.url <- function(repos, type = getOption("pkgType"))
 {
+    if (!is.character(type))
+        stop("invalid 'type'; must be a character string")
     type <- resolvePkgType(type)
     if(is.null(repos)) return(NULL)
     if("@CRAN@" %in% repos && interactive()) {
@@ -833,7 +857,8 @@ contrib.url <- function(repos, type = getOption("pkgType"))
     res <- switch(type,
 		"source" = paste(gsub("/$", "", repos), "src", "contrib", sep = "/"),
                 "mac.binary" = paste(gsub("/$", "", repos), "bin", mac.path, "contrib", ver, sep = "/"),
-                "win.binary" = paste(gsub("/$", "", repos), "bin", "windows", "contrib", ver, sep = "/")
+                "win.binary" = paste(gsub("/$", "", repos), "bin", "windows", "contrib", ver, sep = "/"),
+                stop("invalid 'type'")
                )
     res
 }
@@ -932,6 +957,8 @@ setRepositories <-
         stop("cannot set repositories non-interactively")
     a <- tools:::.get_repositories()
     pkgType <- getOption("pkgType")
+    if (!is.character(pkgType))
+        stop("invalid options(\"pkgType\"); must be a character string")
     if (pkgType == "both") pkgType <- "source" #.Platform$pkgType
     if (pkgType == "binary") pkgType <- .Platform$pkgType
     if(startsWith(pkgType, "mac.binary")) pkgType <- "mac.binary"
@@ -986,15 +1013,19 @@ compareVersion <- function(a, b)
 }
 
 ## ------------- private functions --------------------
-.clean_up_dependencies <- function(x, available = NULL)
-{
-    ## x is a character vector of Depends / Suggests / Imports entries
-    ## returns a character vector of all the package dependencies mentioned
-    x <- x[!is.na(x)]
-    if(!length(x)) return(x)
-    x <- unlist(strsplit(x, ","))
-    unique(sub("^[[:space:]]*([[:alnum:].]+).*$", "\\1" , x))
-}
+
+## .clean_up_dependencies <- function(x, available = NULL)
+## {
+##     ## x is a character vector of Depends / Suggests / Imports entries
+##     ## returns a character vector of all the package dependencies mentioned
+##     x <- x[!is.na(x)]
+##     if(!length(x)) return(x)
+##     x <- unlist(strsplit(x, ",", fixed = TRUE), use.names = FALSE)
+##     unique(sub("^[[:space:]]*([[:alnum:].]+).*$", "\\1" , x))
+## }
+
+.clean_up_dependencies <- function(x)
+    unique.default(tools:::.extract_dependency_package_names(x))
 
 .clean_up_dependencies2 <- function(x, installed, available)
 {
