@@ -1,6 +1,6 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
- *  Copyright (C) 2016--2017   The R Core Team
+ *  Copyright (C) 2016--2021   The R Core Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -27,6 +27,10 @@
 #include <Print.h> /* for R_print */
 #include <R_ext/Itermacros.h>
 
+#ifdef Win32
+#include <trioremap.h> /* for %lld */
+#endif
+
 
 /***
  *** ALTREP Concrete Class Implementations
@@ -136,9 +140,9 @@ Rboolean compact_intseq_Inspect(SEXP x, int pre, int deep, int pvec,
     }
 #endif
 
-    int n = LENGTH(x);
+    R_xlen_t n = XLENGTH(x); // int .. LENGTH(.) not ok, e.g. for -1e9:2e9
     int n1 = INTEGER_ELT(x, 0);
-    int n2 = inc == 1 ? n1 + n - 1 : n1 - n + 1;
+    int n2 = (int) ((inc == 1) ? n1 + n - 1 : n1 - n + 1);
     Rprintf(" %d : %d (%s)", n1, n2,
 	    COMPACT_SEQ_EXPANDED(x) == R_NilValue ? "compact" : "expanded");
     Rprintf("\n");
@@ -165,13 +169,13 @@ static void *compact_intseq_Dataptr(SEXP x, Rboolean writeable)
 
 	if (inc == 1) {
 	    /* compact sequences n1 : n2 with n1 <= n2 */
-	    for (int i = 0; i < n; i++)
-		data[i] = n1 + i;
+	    for (R_xlen_t i = 0; i < n; i++)
+		data[i] = (int) (n1 + i);
 	}
 	else if (inc == -1) {
 	    /* compact sequences n1 : n2 with n1 > n2 */
-	    for (int i = 0; i < n; i++)
-		data[i] = n1 - i;
+	    for (R_xlen_t i = 0; i < n; i++)
+		data[i] = (int) (n1 - i);
 	}
 	else
 	    error("compact sequences with increment %d not supported yet", inc);
@@ -385,7 +389,7 @@ Rboolean compact_realseq_Inspect(SEXP x, int pre, int deep, int pvec,
     R_xlen_t n = XLENGTH(x);
     R_xlen_t n1 = (R_xlen_t) REAL_ELT(x, 0);
     R_xlen_t n2 = inc == 1 ? n1 + n - 1 : n1 - n + 1;
-    Rprintf(" %ld : %ld (%s)", n1, n2,
+    Rprintf(" %lld : %lld (%s)", n1, n2,
 	    COMPACT_SEQ_EXPANDED(x) == R_NilValue ? "compact" : "expanded");
     Rprintf("\n");
     return TRUE;
@@ -919,6 +923,7 @@ SEXP attribute_hidden R_deferred_coerceToString(SEXP v, SEXP info)
    state object.
  */
 
+#ifndef Win32
 static SEXP make_mmap_state(SEXP file, size_t size, int type,
 			    Rboolean ptrOK, Rboolean wrtOK, Rboolean serOK)
 {
@@ -942,7 +947,8 @@ static SEXP make_mmap_state(SEXP file, size_t size, int type,
     UNPROTECT(2);
     return state;
 }
-			    
+#endif
+
 #define MMAP_STATE_FILE(x) CAR(x)
 #define MMAP_STATE_SIZE(x) ((size_t) REAL_ELT(CADR(x), 0))
 #define MMAP_STATE_LENGTH(x) ((size_t) REAL_ELT(CADR(x), 1))
@@ -968,6 +974,7 @@ static R_altrep_class_t mmap_real_class;
    pointer for use by the finalizer.
 */
 
+#ifndef Win32
 static void register_mmap_eptr(SEXP eptr);
 static SEXP make_mmap(void *p, SEXP file, size_t size, int type,
 		      Rboolean ptrOK, Rboolean wrtOK, Rboolean serOK)
@@ -995,6 +1002,7 @@ static SEXP make_mmap(void *p, SEXP file, size_t size, int type,
     UNPROTECT(2); /* state, eptr */
     return ans;
 }
+#endif
 
 #define MMAP_EPTR(x) R_altrep_data1(x)
 #define MMAP_STATE(x) R_altrep_data2(x)
@@ -1021,6 +1029,8 @@ static R_INLINE void *MMAP_ADDR(SEXP x)
    to run a finalizer after unloading would result in an illegal
    instruction. */
 
+
+#ifndef Win32
 static SEXP mmap_list = NULL;
 
 #define MAXCOUNT 10
@@ -1054,6 +1064,7 @@ static void register_mmap_eptr(SEXP eptr)
     /* store the weak reference in the external pointer for do_munmap_file */
     R_SetExternalPtrTag(eptr, CAR(CDR(mmap_list)));
 }
+#endif
 
 #ifdef SIMPLEMMAP
 static void finalize_mmap_objects()
@@ -1108,8 +1119,8 @@ static SEXP mmap_Unserialize(SEXP class, SEXP state)
     return val;
 }
 
-Rboolean mmap_Inspect(SEXP x, int pre, int deep, int pvec,
-		      void (*inspect_subtree)(SEXP, int, int, int))
+static Rboolean mmap_Inspect(SEXP x, int pre, int deep, int pvec,
+			     void (*inspect_subtree)(SEXP, int, int, int))
 {
     Rboolean ptrOK = MMAP_PTROK(x);
     Rboolean wrtOK = MMAP_WRTOK(x);
@@ -1250,15 +1261,17 @@ static void InitMmapRealClass(DllInfo *dll)
  */
 
 #ifdef Win32
+/* unused
 static void mmap_finalize(SEXP eptr)
 {
-    error("mmop objects not supported on Windows yet");
+    error("mmap objects not supported on Windows yet");
 }
+*/
 
 static SEXP mmap_file(SEXP file, int type, Rboolean ptrOK, Rboolean wrtOK,
 		      Rboolean serOK, Rboolean warn)
 {
-    error("mmop objects not supported on Windows yet");
+    error("mmap objects not supported on Windows yet");
 }
 #else
 /* derived from the example in
@@ -1278,7 +1291,6 @@ static void mmap_finalize(SEXP eptr)
     DEBUG_PRINT("finalizing ... ");
     void *p = R_ExternalPtrAddr(eptr);
     size_t size = MMAP_STATE_SIZE(MMAP_EPTR_STATE(eptr));
-    R_SetExternalPtrAddr(eptr, NULL);
 
     if (p != NULL) {
 	munmap(p, size); /* don't check for errors */
@@ -1446,6 +1458,15 @@ static R_INLINE SEXP WRAPPER_WRAPPED_RW(SEXP x)
 
 static SEXP wrapper_Serialized_state(SEXP x)
 {
+    /* If the wrapped value is not an ALTREP and there is no useful
+       metadata then return NULL to allow this to be serialized as a
+       standard object. This avoids serializing potentially large
+       attributes on the wrapped value (PR18142). */
+    if (! ALTREP(WRAPPER_WRAPPED(x)) &&
+	WRAPPER_SORTED(x) == UNKNOWN_SORTEDNESS &&
+	! WRAPPER_NO_NA(x))
+	return NULL;
+
     return CONS(WRAPPER_WRAPPED(x), WRAPPER_METADATA(x));
 }
 
@@ -1482,8 +1503,8 @@ static SEXP wrapper_Duplicate(SEXP x, Rboolean deep)
     return ans;
 }
 
-Rboolean wrapper_Inspect(SEXP x, int pre, int deep, int pvec,
-			 void (*inspect_subtree)(SEXP, int, int, int))
+static Rboolean wrapper_Inspect(SEXP x, int pre, int deep, int pvec,
+				void (*inspect_subtree)(SEXP, int, int, int))
 {
     Rboolean srt = WRAPPER_SORTED(x);
     Rboolean no_na = WRAPPER_NO_NA(x);
@@ -1948,7 +1969,7 @@ SEXP attribute_hidden do_wrap_meta(SEXP call, SEXP op, SEXP args, SEXP env)
     return wrap_meta(x, srt, no_na);
 }
 
-SEXP R_tryWrap(SEXP x)
+SEXP /*attribute_hidden*/ R_tryWrap(SEXP x)
 {
     return wrap_meta(x, UNKNOWN_SORTEDNESS, FALSE);
 }
@@ -1961,7 +1982,7 @@ SEXP attribute_hidden do_tryWrap(SEXP call, SEXP op, SEXP args, SEXP env)
 }
 
 /* When a wrapper has no useful meta-data, is no longer referenced
-   anywhere, and its data is unly accessible from the wrapper, then
+   anywhere, and its data is only accessible from the wrapper, then
    the wrapper can be unwrapped to its wrapped data, and its
    attributes can be transferred to the data.
 
@@ -1972,7 +1993,7 @@ SEXP attribute_hidden do_tryWrap(SEXP call, SEXP op, SEXP args, SEXP env)
    operation. It could be used in other places, but extreme caution is
    needed to make sure there is no possibliity that the wrapper object
    will be referenced from C code after it is cleared. */
-SEXP R_tryUnwrap(SEXP x)
+SEXP attribute_hidden R_tryUnwrap(SEXP x)
 {
     if (! MAYBE_SHARED(x) && is_wrapper(x) &&
 	WRAPPER_SORTED(x) == UNKNOWN_SORTEDNESS && ! WRAPPER_NO_NA(x)) {
