@@ -2,7 +2,7 @@
  *  R : A Computer Language for Statistical Data Analysis
  *  Copyright (C) 1995, 1996  Robert Gentleman and Ross Ihaka
  *  Copyright (C) 2002-2017  The R Foundation
- *  Copyright (C) 1999-2018  The R Core Team.
+ *  Copyright (C) 1999-2022  The R Core Team.
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -29,7 +29,7 @@
 #define R_USE_SIGNALS 1
 #include <Defn.h>
 #include <Internal.h>
-#include <R_ext/RS.h> /* for Calloc, Realloc and for S4 object bit */
+#include <R_ext/RS.h> /* for R_Calloc, R_Realloc and for S4 object bit */
 
 static SEXP GetObject(RCNTXT *cptr)
 {
@@ -234,6 +234,7 @@ SEXP R_LookupMethod(SEXP method, SEXP rho, SEXP callrho, SEXP defrho)
     SEXP val, top = R_NilValue;	/* -Wall */
     static SEXP s_S3MethodsTable = NULL;
     static int lookup_baseenv_after_globalenv = -1;
+    static int lookup_report_search_path_uses = -1;
     char *lookup;
     PROTECT_INDEX validx;
 
@@ -255,6 +256,12 @@ SEXP R_LookupMethod(SEXP method, SEXP rho, SEXP callrho, SEXP defrho)
     if(lookup_baseenv_after_globalenv == -1) {
 	lookup = getenv("_R_S3_METHOD_LOOKUP_BASEENV_AFTER_GLOBALENV_");
 	lookup_baseenv_after_globalenv = 
+	    ((lookup != NULL) && StringFalse(lookup)) ? 0 : 1;
+    }
+
+    if(lookup_report_search_path_uses == -1) {
+	lookup = getenv("_R_S3_METHOD_LOOKUP_REPORT_SEARCH_PATH_USES_");
+	lookup_report_search_path_uses = 
 	    ((lookup != NULL) && StringTrue(lookup)) ? 1 : 0;
     }
 
@@ -296,6 +303,23 @@ SEXP R_LookupMethod(SEXP method, SEXP rho, SEXP callrho, SEXP defrho)
 	REPROTECT(val = findFunWithBaseEnvAfterGlobalEnv(method, top),
 	          validx);
     }
+    else if(lookup_report_search_path_uses) {
+	if(top != R_GlobalEnv) 
+	    REPROTECT(val = findFunInEnvRange(method, ENCLOS(top),
+	                                      R_GlobalEnv), validx);
+	if(val == R_UnboundValue) {
+	    REPROTECT(val = findFunInEnvRange(method, ENCLOS(R_GlobalEnv),
+	                                      R_EmptyEnv), validx);
+	    if((val != R_UnboundValue) && 
+	       (CLOENV(val) != R_BaseNamespace) &&
+	       (CLOENV(val) != R_BaseEnv)) {
+		/* Note that we do not really know where on the search
+		   path we found the method. */
+		REprintf("S3 method lookup found '%s' on search path \n",
+			 CHAR(PRINTNAME(method)));
+	    }
+	}
+    }
     else
 	REPROTECT(val = findFunInEnvRange(method, ENCLOS(top), R_EmptyEnv),
 	          validx);
@@ -398,11 +422,14 @@ SEXP dispatchMethod(SEXP op, SEXP sxp, SEXP dotClass, RCNTXT *cptr, SEXP method,
 	}
     }
 
-    if( (RDEBUG(op) && R_current_debug_state()) || RSTEP(op) ) {
+    /* Debug a method when debugging the generic. When called via UseMethod or
+       NextMethod, RSTEP(op) will always be zero because the bit is cleared by
+       applyClosure. We thus approximate and enter the debugger also when
+       RDEBUG(rho) is set. */
+    if ((RDEBUG(op) && R_current_debug_state()) || RSTEP(op) || RDEBUG(rho))
 	SET_RSTEP(sxp, 1);
-    }
 
-    SEXP newcall =  PROTECT(duplicate(cptr->call));
+    SEXP newcall =  PROTECT(shallow_duplicate(cptr->call));
     SETCAR(newcall, method);
     R_GlobalContext->callflag = CTXT_GENERIC;
     SEXP matchedarg = PROTECT(cptr->promargs); /* ? is this PROTECT needed ? */
@@ -480,11 +507,14 @@ SEXP attribute_hidden NORET do_usemethod(SEXP call, SEXP op, SEXP args, SEXP env
     RCNTXT *cptr;
     static SEXP do_usemethod_formals = NULL;
 
+    static int lookup_use_topenv_as_defenv = -1;
+    char *lookup;
+
     if (do_usemethod_formals == NULL)
 	do_usemethod_formals = allocFormalsList2(install("generic"),
 						 install("object"));
 
-    PROTECT(argList = matchArgs(do_usemethod_formals, args, call));
+    PROTECT(argList = matchArgs_NR(do_usemethod_formals, args, call));
     if (CAR(argList) == R_MissingArg)
 	errorcall(call, _("there must be a 'generic' argument"));
     else
@@ -492,6 +522,11 @@ SEXP attribute_hidden NORET do_usemethod(SEXP call, SEXP op, SEXP args, SEXP env
     if(!isString(generic) || LENGTH(generic) != 1)
 	errorcall(call, _("'generic' argument must be a character string"));
 
+    if(lookup_use_topenv_as_defenv == -1) {
+	lookup = getenv("_R_S3_METHOD_LOOKUP_USE_TOPENV_AS_DEFENV_");
+	lookup_use_topenv_as_defenv = 
+	    ((lookup != NULL) && StringFalse(lookup)) ? 0 : 1;
+    }
 
     /* get environments needed for dispatching.
        callenv = environment from which the generic was called
@@ -512,10 +547,15 @@ SEXP attribute_hidden NORET do_usemethod(SEXP call, SEXP op, SEXP args, SEXP env
 	The generic need not be a closure (Henrik Bengtsson writes
 	UseMethod("$"), although only functions are documented.)
     */
-    val = findVar1(installTrChar(STRING_ELT(generic, 0)),
-		   ENCLOS(env), FUNSXP, TRUE); /* That has evaluated promises */
-    if(TYPEOF(val) == CLOSXP) defenv = CLOENV(val);
-    else defenv = R_BaseNamespace;
+    if(lookup_use_topenv_as_defenv) {
+	defenv = topenv(R_NilValue, env);
+    } else {
+	val = findVar1(installTrChar(STRING_ELT(generic, 0)),
+		       ENCLOS(env), FUNSXP, TRUE); /* That has evaluated
+						    * promises */
+	if(TYPEOF(val) == CLOSXP) defenv = CLOENV(val);
+	else defenv = R_BaseNamespace;
+    }
 
     if (CADR(argList) != R_MissingArg)
 	PROTECT(obj = eval(CADR(argList), env));
@@ -570,7 +610,7 @@ static SEXP fixcall(SEXP call, SEXP args)
 		if( !found ) {
 			SETCDR(s, allocList(1));
 			SET_TAG(CDR(s), TAG(t));
-			SETCAR(CDR(s), duplicate(CAR(t)));
+			SETCAR(CDR(s), lazy_duplicate(CAR(t)));
 		}
 	}
     }
@@ -654,7 +694,7 @@ SEXP attribute_hidden do_nextmethod(SEXP call, SEXP op, SEXP args, SEXP env)
     if (cptr == NULL)
 	error(_("'NextMethod' called from outside a function"));
 
-    PROTECT(newcall = duplicate(cptr->call));
+    PROTECT(newcall = shallow_duplicate(cptr->call));
 
     /* eg get("print.ts")(1) or do.call() */
     if (TYPEOF(CAR(cptr->call)) != SYMSXP)
@@ -896,7 +936,7 @@ SEXP attribute_hidden do_unclass(SEXP call, SEXP op, SEXP args, SEXP env)
 	    break;
 	}
 	if (MAYBE_REFERENCED(CAR(args)))
-	    SETCAR(args, shallow_duplicate(CAR(args)));
+	    SETCAR(args, R_shallow_duplicate_attr(CAR(args)));
 	setAttrib(CAR(args), R_ClassSymbol, R_NilValue);
     }
     return CAR(args);
@@ -1206,7 +1246,7 @@ static SEXP dispatchNonGeneric(SEXP name, SEXP env, SEXP fdef)
 	cptr = cptr->nextcontext;
     }
 
-    PROTECT(e = duplicate(R_syscall(0, cptr)));
+    PROTECT(e = shallow_duplicate(R_syscall(0, cptr)));
     SETCAR(e, fun);
     /* evaluate a call the non-generic with the same arguments and from
        the same environment as the call to the generic version */
@@ -1222,7 +1262,7 @@ SEXP attribute_hidden do_standardGeneric(SEXP call, SEXP op, SEXP args, SEXP env
 {
     SEXP arg, value, fdef; R_stdGen_ptr_t ptr = R_get_standardGeneric_ptr();
 
-    checkArity(op, args);
+    checkArity(op, args); /* set to -1 */
     check1arg(args, call, "f");
 
     if(!ptr) {
@@ -1232,7 +1272,6 @@ SEXP attribute_hidden do_standardGeneric(SEXP call, SEXP op, SEXP args, SEXP env
 	ptr = R_get_standardGeneric_ptr();
     }
 
-    checkArity(op, args); /* set to -1 */
     arg = CAR(args);
     if(!isValidStringF(arg))
 	errorcall(call,
@@ -1360,11 +1399,11 @@ SEXP do_set_prim_method(SEXP op, const char *code_string, SEXP fundef,
 	if(prim_methods) {
 	    int i;
 
-	    prim_methods  = Realloc(prim_methods,  n, prim_methods_t);
-	    prim_generics = Realloc(prim_generics, n, SEXP);
-	    prim_mlist	  = Realloc(prim_mlist,	   n, SEXP);
+	    prim_methods  = R_Realloc(prim_methods,  n, prim_methods_t);
+	    prim_generics = R_Realloc(prim_generics, n, SEXP);
+	    prim_mlist	  = R_Realloc(prim_mlist,	   n, SEXP);
 
-	    /* Realloc does not clear the added memory, hence: */
+	    /* R_Realloc does not clear the added memory, hence: */
 	    for (i = maxMethodsOffset ; i < n ; i++) {
 		prim_methods[i]	 = NO_METHODS;
 		prim_generics[i] = NULL;
@@ -1372,9 +1411,9 @@ SEXP do_set_prim_method(SEXP op, const char *code_string, SEXP fundef,
 	    }
 	}
 	else {
-	    prim_methods  = Calloc(n, prim_methods_t);
-	    prim_generics = Calloc(n, SEXP);
-	    prim_mlist	  = Calloc(n, SEXP);
+	    prim_methods  = R_Calloc(n, prim_methods_t);
+	    prim_generics = R_Calloc(n, SEXP);
+	    prim_mlist	  = R_Calloc(n, SEXP);
 	}
 	maxMethodsOffset = n;
     }
@@ -1437,42 +1476,26 @@ argument to standardGeneric.
 */
 static SEXP get_this_generic(SEXP args)
 {
-    const void *vmax = vmaxget();
-    SEXP value = R_NilValue; static SEXP gen_name;
-    int i, n;
+    static SEXP gen_name = NULL;
     RCNTXT *cptr;
-    const char *fname;
+    SEXP fname;
 
     /* a second argument to the call, if any, is taken as the function */
     if(CDR(args) != R_NilValue)
 	return CAR(CDR(args));
-    /* else use sys.function (this is fairly expensive-- would be good
-     * to force a second argument if possible) */
-    PROTECT(args);
     if(!gen_name)
 	gen_name = install("generic");
-    cptr = R_GlobalContext;
-    fname = translateChar(asChar(CAR(args)));
-    n = framedepth(cptr);
-    /* check for a matching "generic" slot */
-    for(i=0;  i<n; i++) {
-	SEXP rval = R_sysfunction(i, cptr);
-	if(isObject(rval)) {
-	    PROTECT(rval);
-	    SEXP generic = getAttrib(rval, gen_name);
-	    if(TYPEOF(generic) == STRSXP &&
-	       !strcmp(translateChar(asChar(generic)), fname)) {
-	      value = rval;
-	      UNPROTECT(1); /* rval */
-	      break;
-	    }
-	    UNPROTECT(1); /* rval */
-	}
-    }
-    UNPROTECT(1);
-    vmaxset(vmax);
+    fname = STRING_ELT(CAR(args), 0); /* type and length checked by caller */
 
-    return value;
+    /* check for a matching "generic" slot */
+    for(cptr = R_GlobalContext; cptr != NULL; cptr = cptr->nextcontext)
+	if((cptr->callflag & CTXT_FUNCTION) && isObject(cptr->callfun)) {
+	    SEXP generic = getAttrib(cptr->callfun, gen_name);
+	    if(isValidString(generic) && Seql(fname, STRING_ELT(generic, 0)))
+		/* not duplicating/marking immutable, used read-only */
+		return cptr->callfun;
+	}
+    return R_NilValue;
 }
 
 /* Could there be methods for this op?	Checks
@@ -1571,7 +1594,12 @@ R_possible_dispatch(SEXP call, SEXP op, SEXP args, SEXP rho,
 		UNPROTECT(2);
 		return value;
 	    } else {
+		/* INC/DEC of REFCNT needed for non-tracking args */
+		for (SEXP a = args; a != R_NilValue; a = CDR(a))
+		    INCREMENT_REFCNT(CAR(a));
 		value = applyClosure(call, value, args, rho, suppliedvars);
+		for (SEXP a = args; a != R_NilValue; a = CDR(a))
+		    DECREMENT_REFCNT(CAR(a));
                 UNPROTECT(1);
                 return value;
             }
@@ -1591,8 +1619,14 @@ R_possible_dispatch(SEXP call, SEXP op, SEXP args, SEXP rho,
 	    SET_PRVALUE(CAR(b), CAR(a));
 	value = applyClosure(call, fundef, s, rho, R_NilValue);
 	UNPROTECT(1);
-    } else
+    } else {
+	/* INC/DEC of REFCNT needed for non-tracking args */
+	for (SEXP a = args; a != R_NilValue; a = CDR(a))
+	    INCREMENT_REFCNT(CAR(a));
 	value = applyClosure(call, fundef, args, rho, R_NilValue);
+	for (SEXP a = args; a != R_NilValue; a = CDR(a))
+	    DECREMENT_REFCNT(CAR(a));
+    }
     prim_methods[offset] = current;
     if(value == deferred_default_object)
 	return NULL;
@@ -1691,7 +1725,7 @@ SEXP R_do_new_object(SEXP class_def)
 	TYPEOF(value) == EXTPTRSXP;
     if((TYPEOF(value) == S4SXP || getAttrib(e, R_PackageSymbol) != R_NilValue) &&
        !xDataType)
-    { /* Anything but an object from a base "class" (numeric, matrix,..) */
+    {
 	setAttrib(value, R_ClassSymbol, e);
 	SET_S4_OBJECT(value);
     }
