@@ -24,6 +24,7 @@
  *  vectorIndex()     -- used for "[[" and "[[<-" with a vector arg
 
  *  mat2indsub()      -- for "mat[i]"     "    "            "
+ *  strmat2intmat()   -- for "mat[<ch>]"  "    "            "
 
  *  makeSubscript()   -- for "[" and "[<-" in ./subset.c and ./subassign.c,
  *			 and "[[<-" with a scalar in ./subassign.c
@@ -44,6 +45,33 @@
 /* We might get a call with R_NilValue from subassignment code */
 #define ECALL(call, yy)     if(call == R_NilValue) error(yy);    else errorcall(call, yy);
 #define ECALL3(call, yy, A) if(call == R_NilValue) error(yy, A); else errorcall(call, yy, A);
+#define ECALL4(call, yy, A, B) if(call == R_NilValue) error(yy, A, B); else errorcall(call, yy, A, B);
+
+static void NORET ECALL_OutOfBounds(SEXP x, int subscript,
+				    R_xlen_t index, SEXP call)
+{
+    if (call == R_NilValue)
+	call = R_CurrentExpression; /* default behaves like error() */
+    SEXP sindex = ScalarReal((double) index);
+    PROTECT(sindex);
+    SEXP cond = R_makeOutOfBoundsError(x, subscript, sindex, call, NULL);
+    PROTECT(cond);
+    R_signalErrorCondition(cond, call);
+    UNPROTECT(2); /* sindex, cond; not reached */
+}
+
+static void NORET ECALL_OutOfBoundsCHAR(SEXP x, int subscript,
+					SEXP sindex, SEXP call)
+{
+    if (call == R_NilValue)
+	call = R_CurrentExpression; /* default behaves like error() */
+    sindex = ScalarString(sindex);
+    PROTECT(sindex);
+    SEXP cond = R_makeOutOfBoundsError(x, subscript, sindex, call, NULL);
+    PROTECT(cond);
+    R_signalErrorCondition(cond, call);
+    UNPROTECT(2); /* sindex, cond; not reached */
+}
 
 /* This allows for the unusual case where x is of length 2,
    and x[[-m]] selects one element for m = 1, 2.
@@ -95,11 +123,11 @@ OneIndex(SEXP x, SEXP s, R_xlen_t nx, int partial, SEXP *newname,
 	double dblind = REAL_ELT(s, pos);
 	if(!ISNAN(dblind)) {
 	    /* see comment above integerOneIndex */
-	    if (dblind > 0) indx = (R_xlen_t)(dblind - 1);
-	    else if (dblind == 0 || nx < 2) {
+	    if (dblind >= 1) indx = (R_xlen_t)(dblind - 1);
+	    else if (dblind > -1 || nx < 2) {
 		ECALL3(call, _("attempt to select less than one element in %s"), "OneIndex <real>");
 	    } else if (nx == 2 && dblind > -3)
-		indx = (R_xlen_t)(2 + dblind);
+		indx = 2 + (R_xlen_t) dblind;
 	    else {
 		ECALL3(call, _("attempt to select more than one element in %s"), "OneIndex <real>");
 	    }
@@ -212,19 +240,19 @@ get1index(SEXP s, SEXP names, R_xlen_t len, int pok, int pos, SEXP call)
 	double dblind = REAL_ELT(s, pos);
 	if(!ISNAN(dblind)) {
 	    /* see comment above integerOneIndex */
-	    if (dblind > 0) {
+	    if (dblind >= 1) {
 		if(R_FINITE(dblind)) indx = (R_xlen_t)(dblind - 1);
-	    } else if (dblind == 0 || len < 2) {
+	    } else if (dblind > -1 || len < 2) {
 		ECALL3(call,
-		       _((dblind < 0) ? "invalid negative subscript in %s"
-			 : "attempt to select less than one element in %s"),
+		       (dblind <= -1) ? _("invalid negative subscript in %s")
+		       : _("attempt to select less than one element in %s"),
 		       "get1index <real>");
 	    } else if (len == 2 && dblind > -3) // dblind = -2 or -1 {why exception ?}
-		indx = (R_xlen_t)(2 + dblind);
+		indx = 2 + (R_xlen_t) dblind;
 	    else {
 		ECALL3(call,
-		       _((dblind < 0) ? "invalid negative subscript in %s"
-			 : "attempt to select more than one element in %s"),
+		       (dblind <= -1) ? _("invalid negative subscript in %s")
+		       : _("attempt to select more than one element in %s"),
 		       "get1index <real>");
 	    }
 	}
@@ -366,9 +394,14 @@ vectorIndex(SEXP x, SEXP thesub, int start, int stop, int pok, SEXP call,
    A zero/NA anywhere in a row will cause a zero/NA in the same
    position in the result.
 */
+/*
+  Called from `[`   VectorSubset()    ./subset.c	and
+              `[<-` VectorSubassign() ./subassign.c
 
-
-SEXP attribute_hidden mat2indsub(SEXP dims, SEXP s, SEXP call)
+  only if(isMatrix(s) && isArray(x) && ncols(s) == "length(dim(x))"),
+  where dims = getAttrib(x, R_DimSymbol);
+*/
+SEXP attribute_hidden mat2indsub(SEXP dims, SEXP s, SEXP call, SEXP x)
 {
     int nrs = nrows(s);
     R_xlen_t NR = nrs;
@@ -396,12 +429,13 @@ SEXP attribute_hidden mat2indsub(SEXP dims, SEXP s, SEXP call)
 		for (int j = 0; j < ndim; j++) {
 		    double k = ps[i + j * NR];
 		    if(ISNAN(k)) {rv[i] = NA_REAL; break;}
+		    k = trunc(k); // = "as.integer(k)" for large k
 		    if(k < 0) {
 			ECALL(call, _("negative values are not allowed in a matrix subscript"));
 		    }
 		    if(k == 0.) {rv[i] = 0.; break;}
 		    if (k > pdims[j]) {
-			ECALL(call, _("subscript out of bounds"));
+			ECALL_OutOfBounds(x, j, (R_xlen_t)k, call);
 		    }
 		    rv[i] += (k - 1.) * tdim;
 		    tdim *= pdims[j];
@@ -420,7 +454,7 @@ SEXP attribute_hidden mat2indsub(SEXP dims, SEXP s, SEXP call)
 		    }
 		    if(k == 0) {rv[i] = 0.; break;}
 		    if (k > pdims[j]) {
-			ECALL(call, _("subscript out of bounds"));
+			ECALL_OutOfBounds(x, j, k, call);
 		    }
 		    rv[i] += (double) ((k - 1) * tdim);
 		    tdim *= pdims[j];
@@ -445,7 +479,7 @@ SEXP attribute_hidden mat2indsub(SEXP dims, SEXP s, SEXP call)
 		}
 		if(k == 0) {iv[i] = 0; break;}
 		if (k > pdims[j]) {
-		    ECALL(call, _("subscript out of bounds"));
+		    ECALL_OutOfBounds(x, j, k, call);
 		}
 		iv[i] += (k - 1) * tdim;
 		tdim *= pdims[j];
@@ -463,35 +497,54 @@ array and i is a character matrix with n columns, this code converts i
 to an integer matrix by matching against the dimnames of x. NA values
 in any row of i propagate to the result.  Unmatched entries result in
 a subscript out of bounds error.  */
+/*
+  Called from `[`   VectorSubset()    ./subset.c	and
+              `[<-` VectorSubassign() ./subassign.c
 
-SEXP attribute_hidden strmat2intmat(SEXP s, SEXP dnamelist, SEXP call)
+  only if(isMatrix(s) && isArray(x) && ncols(s) == "length(dim(x))"),
+  where dnamelist = PROTECT(GetArrayDimnames(x))
+ */
+SEXP attribute_hidden strmat2intmat(SEXP s, SEXP dnamelist, SEXP call, SEXP x)
 {
     /* XXX: assumes all args are protected */
     SEXP dim = getAttrib(s, R_DimSymbol);
-    int nr = INTEGER(dim)[0]; // = nrow(s)
+    int nr = INTEGER(dim)[0], // = nrow(s)
+	nc = INTEGER(dim)[1]; // = ncol(s)
     if(isNull(dnamelist)) { // && nr > 0)
+#if 1
 	ECALL(call, _("no 'dimnames' attribute for array")); // as in  arraySubscript()
+#else
+	ECALL4(call,
+          _("%d x %d character matrix subscript illegal for NULL dimnames"),
+	       nr, nc);
+#endif
     }
-    int i, j, v;
-    R_xlen_t idx, NR = nr;
-    SEXP dnames, snames, si, sicol, s_elt;
-    PROTECT(snames = allocVector(STRSXP, nr));
-    PROTECT(si = allocVector(INTSXP, xlength(s)));
+#if 0 /* unneeded as only called  if (length(dnamelist) == nc) : */
+    if(length(dnamelist) != nc) {
+	ECALL4(call,
+          _("subscripting by character matrix of %d columns does not match length(dimnames(x)) == %d"),
+	       nc, length(dnamelist));
+    }
+#endif
+    SEXP snames = PROTECT(allocVector(STRSXP, nr));
+    // result si := an integer matrix of same dimension as 's'
+    SEXP si = PROTECT(allocVector(INTSXP, xlength(s)));
     dimgets(si, dim);
     int *psi = INTEGER(si);
     memset(psi, 0, XLENGTH(si) * sizeof(int));
-    for (i = 0; i < length(dnamelist); i++) {
-	dnames = VECTOR_ELT(dnamelist, i);
-	for (j = 0; j < nr; j++)
-	    SET_STRING_ELT(snames, j, STRING_ELT(s, j + (i * NR)));
-	PROTECT(sicol = match(dnames, snames, 0));
-	for (j = 0; j < nr; j++) {
-	    v = INTEGER_ELT(sicol, j);
-	    idx = j + (i * NR);
-	    s_elt = STRING_ELT(s, idx);
+    for (int i = 0; i < nc; i++) {
+	R_xlen_t iNR = i * (R_xlen_t) nr;
+	for (int j = 0; j < nr; j++)
+	    SET_STRING_ELT(snames, j, STRING_ELT(s, j + iNR));
+	SEXP dnames = VECTOR_ELT(dnamelist, i);
+	SEXP sicol = PROTECT(match(dnames, snames, 0));
+	for (int j = 0; j < nr; j++) {
+	    int v = INTEGER_ELT(sicol, j);
+	    R_xlen_t idx = j + iNR;
+	    SEXP s_elt = STRING_ELT(s, idx);
 	    if (s_elt == NA_STRING) v = NA_INTEGER;
-	    if (!CHAR(s_elt)[0]) v = 0; /* disallow "" match */
-	    if (v == 0) errorcall(call, _("subscript out of bounds"));
+	    else if (!CHAR(s_elt)[0]) v = 0; /* disallow "" match */
+	    if (v == 0) ECALL_OutOfBoundsCHAR(x, i, s_elt, call);
 	    psi[idx] = v;
 	}
 	UNPROTECT(1);
@@ -681,7 +734,8 @@ static SEXP positiveSubscript(SEXP s, R_xlen_t ns, R_xlen_t nx)
 }
 
 static SEXP
-integerSubscript(SEXP s, R_xlen_t ns, R_xlen_t nx, R_xlen_t *stretch, SEXP call)
+integerSubscript(SEXP s, R_xlen_t ns, R_xlen_t nx, R_xlen_t *stretch,
+		 SEXP call, SEXP x)
 {
     R_xlen_t i;
     int ii, neg, max, canstretch;
@@ -705,7 +759,7 @@ integerSubscript(SEXP s, R_xlen_t ns, R_xlen_t nx, R_xlen_t *stretch, SEXP call)
     if (max > nx) {
 	if(canstretch) *stretch = max;
 	else {
-	    ECALL(call, _("subscript out of bounds"));
+	    ECALL_OutOfBounds(x, -1, max, call);
 	}
     }
     if (neg) {
@@ -719,25 +773,22 @@ integerSubscript(SEXP s, R_xlen_t ns, R_xlen_t nx, R_xlen_t *stretch, SEXP call)
 }
 
 static SEXP
-realSubscript(SEXP s, R_xlen_t ns, R_xlen_t nx, R_xlen_t *stretch, SEXP call)
+realSubscript(SEXP s, R_xlen_t ns, R_xlen_t nx, R_xlen_t *stretch,
+	      SEXP call, SEXP x)
 {
-    R_xlen_t i;
-    int canstretch;
-    double ii, min, max;
-    Rboolean isna = FALSE;
-    canstretch = *stretch > 0;
+    int canstretch = *stretch > 0;
     *stretch = 0;
-    min = 0;
-    max = 0;
+    double min = 0, max = 0;
     const double *ps = REAL_RO(s);
-    for (i = 0; i < ns; i++) {
-	ii = ps[i];
+    Rboolean isna = FALSE;
+    for (R_xlen_t i = 0; i < ns; i++) {
+	double ii = ps[i];
 	if (R_FINITE(ii)) {
 	    if (ii < min) min = ii;
 	    if (ii > max) max = ii;
 	} else isna = TRUE;
     }
-    if (max > nx) {
+    if (max >= nx+1.) {
 #ifndef LONG_VECTOR_SUPPORT
 	if (max > INT_MAX) {
 	    ECALL(call, _("subscript too large for 32-bit R"));
@@ -745,22 +796,19 @@ realSubscript(SEXP s, R_xlen_t ns, R_xlen_t nx, R_xlen_t *stretch, SEXP call)
 #endif
 	if(canstretch) *stretch = (R_xlen_t) max;
 	else {
-	    ECALL(call, _("subscript out of bounds"));
+	    ECALL_OutOfBounds(x, -1, (R_xlen_t) max, call);
 	}
     }
-    if (min < 0) {
-	if (max == 0 && !isna) {
-	    SEXP indx;
-	    R_xlen_t stretch = 0;
-	    double dx;
-	    R_xlen_t i, ix;
-	    PROTECT(indx = allocVector(LGLSXP, nx));
+    if (min <= -1) {
+	if (max < 1 && !isna) {
+	    R_xlen_t i, stretch = 0;
+	    SEXP indx = PROTECT(allocVector(LGLSXP, nx));
 	    int *pindx = LOGICAL(indx);
 	    for (i = 0; i < nx; i++) pindx[i] = 1;
 	    for (i = 0; i < ns; i++) {
-		dx = ps[i];
-		if (R_FINITE(dx) && dx != 0  && -dx <= nx) {
-		    ix = (R_xlen_t)(-dx - 1);
+		double dx = ps[i];
+		if (R_FINITE(dx) && dx <= -1  && -dx < nx+1.) {
+		    R_xlen_t ix = (R_xlen_t)(-dx - 1);
 		    pindx[ix] = 0;
 		}
 	    }
@@ -829,7 +877,7 @@ realSubscript(SEXP s, R_xlen_t ns, R_xlen_t nx, R_xlen_t *stretch, SEXP call)
 
 static SEXP
 stringSubscript(SEXP s, R_xlen_t ns, R_xlen_t nx, SEXP names,
-		R_xlen_t *stretch, SEXP call)
+		R_xlen_t *stretch, SEXP call, SEXP x, int dim)
 {
     SEXP indx, indexnames = R_NilValue;
     R_xlen_t i, j, nnames, extra, sub;
@@ -901,7 +949,7 @@ stringSubscript(SEXP s, R_xlen_t ns, R_xlen_t nx, SEXP names,
 	}
 	if (sub == 0) {
 	    if (!canstretch) {
-		ECALL(call, _("subscript out of bounds"));
+		ECALL_OutOfBoundsCHAR(x, dim, STRING_ELT(s, i), call);
 	    }
 	    extra += 1;
 	    sub = extra;
@@ -942,11 +990,11 @@ int_arraySubscript(int dim, SEXP s, SEXP dims, SEXP x, SEXP call)
     case LGLSXP:
 	return logicalSubscript(s, ns, nd, &stretch, call);
     case INTSXP:
-	return integerSubscript(s, ns, nd, &stretch, call);
+	return integerSubscript(s, ns, nd, &stretch, call, x);
     case REALSXP:
 	/* We don't yet allow subscripts > R_SHORT_LEN_MAX */
 	PROTECT(tmp = coerceVector(s, INTSXP));
-	tmp = integerSubscript(tmp, ns, nd, &stretch, call);
+	tmp = integerSubscript(tmp, ns, nd, &stretch, call, x);
 	UNPROTECT(1);
 	return tmp;
     case STRSXP:
@@ -955,7 +1003,7 @@ int_arraySubscript(int dim, SEXP s, SEXP dims, SEXP x, SEXP call)
 	    ECALL(call, _("no 'dimnames' attribute for array"));
 	}
 	dnames = VECTOR_ELT(dnames, dim);
-	return stringSubscript(s, ns, nd, dnames, &stretch, call);
+	return stringSubscript(s, ns, nd, dnames, &stretch, call, x, dim);
     case SYMSXP:
 	if (s == R_MissingArg)
 	    return nullSubscript(nd);
@@ -1024,16 +1072,16 @@ makeSubscript(SEXP x, SEXP s, R_xlen_t *stretch, SEXP call)
 	ans = logicalSubscript(s, ns, nx, stretch, call);
 	break;
     case INTSXP:
-	ans = integerSubscript(s, ns, nx, stretch, call);
+	ans = integerSubscript(s, ns, nx, stretch, call, x);
 	break;
     case REALSXP:
-	ans = realSubscript(s, ns, nx, stretch, call);
+	ans = realSubscript(s, ns, nx, stretch, call, x);
 	break;
     case STRSXP:
     {
 	SEXP names = PROTECT(getAttrib(x, R_NamesSymbol));
 	/* *stretch = 0; */
-	ans = stringSubscript(s, ns, nx, names, stretch, call);
+	ans = stringSubscript(s, ns, nx, names, stretch, call, x, -1);
 	UNPROTECT(1); /* names */
 	break;
     }
