@@ -1,7 +1,7 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
+ *  Copyright (C) 1997--2022  The R Core Team
  *  Copyright (C) 1995, 1996  Robert Gentleman and Ross Ihaka
- *  Copyright (C) 1997--2020  The R Core Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -281,10 +281,10 @@ void InitTypeTables(void) {
 
 SEXP type2str_nowarn(SEXPTYPE t) /* returns a CHARSXP */
 {
-    if (t < MAX_NUM_SEXPTYPE) { /* FIXME: branch not really needed */
+    // if (t >= 0 && t < MAX_NUM_SEXPTYPE) { /* branch not really needed */
 	SEXP res = Type2Table[t].rcharName;
 	if (res != NULL) return res;
-    }
+    // }
     return R_NilValue;
 }
 
@@ -302,10 +302,10 @@ SEXP type2str(SEXPTYPE t) /* returns a CHARSXP */
 
 SEXP type2rstr(SEXPTYPE t) /* returns a STRSXP */
 {
-    if (t < MAX_NUM_SEXPTYPE) { /* FIXME: branch not really needed */
+    // if (t < MAX_NUM_SEXPTYPE) {
 	SEXP res = Type2Table[t].rstrName;
 	if (res != NULL) return res;
-    }
+    // }
     error(_("type %d is unimplemented in '%s'"), t,
 	  "type2ImmutableScalarString");
     return R_NilValue; /* for -Wall */
@@ -313,10 +313,10 @@ SEXP type2rstr(SEXPTYPE t) /* returns a STRSXP */
 
 const char *type2char(SEXPTYPE t) /* returns a char* */
 {
-    if (t < MAX_NUM_SEXPTYPE) { /* FIXME: branch not really needed */
+    // if (t >=0 && t < MAX_NUM_SEXPTYPE) { /* branch not really needed */
 	const char * res = Type2Table[t].cstrName;
 	if (res != NULL) return res;
-    }
+    // }
     warning(_("type %d is unimplemented in '%s'"), t, "type2char");
     static char buf[50];
     snprintf(buf, 50, "unknown type #%d", t);
@@ -326,12 +326,10 @@ const char *type2char(SEXPTYPE t) /* returns a char* */
 #ifdef UNUSED
 SEXP NORET type2symbol(SEXPTYPE t)
 {
-    if (t >= 0 && t < MAX_NUM_SEXPTYPE) { /* FIXME: branch not really needed */
+    // if (t >= 0 && t < MAX_NUM_SEXPTYPE) { /* branch not really needed */
 	SEXP res = Type2Table[t].rsymName;
-	if (res != NULL) {
-	    return res;
-	}
-    }
+	if (res != NULL) return res;
+    // }
     error(_("type %d is unimplemented in '%s'"), t, "type2symbol");
 }
 #endif
@@ -418,7 +416,8 @@ Rboolean isBlankString(const char *s)
     if(mbcslocale) {
 	wchar_t wc; size_t used; mbstate_t mb_st;
 	mbs_init(&mb_st);
-	while( (used = Mbrtowc(&wc, s, MB_CUR_MAX, &mb_st)) ) {
+	// This does not allow for surrogate pairs, but all blanks are in BMP
+	while( (used = Mbrtowc(&wc, s, R_MB_CUR_MAX, &mb_st)) ) {
 	    if(!iswspace((wint_t) wc)) return FALSE;
 	    s += used;
 	}
@@ -475,10 +474,11 @@ SEXP attribute_hidden EnsureString(SEXP s)
     return s;
 }
 
-/* FIXME: ngettext reguires unsigned long, but %u would seem appropriate */
+// NB: have  checkArity(a,b) :=  Rf_checkArityCall(a,b,call)
 void Rf_checkArityCall(SEXP op, SEXP args, SEXP call)
 {
     if (PRIMARITY(op) >= 0 && PRIMARITY(op) != length(args)) {
+	/* FIXME: ngettext reguires unsigned long, but %u would seem appropriate */
 	if (PRIMINTERNAL(op))
 	    error(ngettext("%d argument passed to .Internal(%s) which requires %d",
 		     "%d arguments passed to .Internal(%s) which requires %d",
@@ -496,10 +496,9 @@ void Rf_checkArityCall(SEXP op, SEXP args, SEXP call)
 void attribute_hidden Rf_check1arg(SEXP arg, SEXP call, const char *formal)
 {
     SEXP tag = TAG(arg);
-    const char *supplied;
-    size_t ns;
     if (tag == R_NilValue) return;
-    supplied = CHAR(PRINTNAME(tag)); ns = strlen(supplied);
+    const char *supplied = CHAR(PRINTNAME(tag));
+    size_t ns = strlen(supplied);
     if (ns > strlen(formal) || strncmp(supplied, formal, ns))
 	errorcall(call, _("supplied argument name '%s' does not match '%s'"),
 		  supplied, formal);
@@ -520,6 +519,25 @@ SEXP nthcdr(SEXP s, int n)
     return R_NilValue;/* for -Wall */
 }
 
+/* Destructively removes R_NilValue ('NULL') elements from a pairlist. */
+SEXP R_listCompact(SEXP s, Rboolean keep_initial) {
+    if(!keep_initial)
+    // skip initial NULL values
+	while (s != R_NilValue && CAR(s) == R_NilValue)
+	    s = CDR(s);
+
+    SEXP val = s;
+    SEXP prev = s;
+    while (s != R_NilValue) {
+	s = CDR(s);
+	if (CAR(s) == R_NilValue) // skip it
+	    SETCDR(prev, CDR(s));
+	else
+	    prev = s;
+    }
+    return val;
+}
+
 
 /* This is a primitive (with no arguments) */
 SEXP attribute_hidden do_nargs(SEXP call, SEXP op, SEXP args, SEXP rho)
@@ -893,8 +911,9 @@ SEXP attribute_hidden do_dirname(SEXP call, SEXP op, SEXP args, SEXP rho)
 		wcscpy (buf, pp);
 		R_wfixslash(buf);
 		/* remove trailing file separator(s) */
-		while ( *(p = buf + wcslen(buf) - 1) == L'/'  && p > buf
-			&& (p > buf+2 || *(p-1) != L':')) *p = L'\0';
+		p = buf + wcslen(buf) - 1;
+		while (p > buf && *p == L'/'
+		       && (p > buf+2 || *(p-1) != L':')) *p-- = L'\0';
 		p = wcsrchr(buf, L'/');
 		if(p == NULL) wcscpy(buf, L".");
 		else {
@@ -1129,7 +1148,7 @@ SEXP attribute_hidden do_encodeString(SEXP call, SEXP op, SEXP args, SEXP rho)
        of print dispatch with WinUTF8out being already set to TRUE). */
     if (WinUTF8out) {
 	begincontext(&cntxt, CTXT_CCODE, R_NilValue, R_BaseEnv, R_BaseEnv,
-	             R_NilValue, R_NilValue);
+		     R_NilValue, R_NilValue);
 	cntxt.cend = &encode_cleanup;
 	havecontext = TRUE;
 	WinUTF8out = FALSE;
@@ -1181,6 +1200,9 @@ SEXP attribute_hidden do_encoding(SEXP call, SEXP op, SEXP args, SEXP rho)
     return ans;
 }
 
+#define IS_NATIVE(tmp) \
+    (! IS_LATIN1(tmp) && ! IS_UTF8(tmp) && ! IS_BYTES(tmp))
+
 SEXP attribute_hidden do_setencoding(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     SEXP x, enc, tmp;
@@ -1208,15 +1230,17 @@ SEXP attribute_hidden do_setencoding(SEXP call, SEXP op, SEXP args, SEXP rho)
 	tmp = STRING_ELT(x, i);
 	if(tmp == NA_STRING) continue;
 	if (! ((ienc == CE_LATIN1 && IS_LATIN1(tmp)) ||
-	       (ienc == CE_UTF8 && IS_UTF8(tmp)) ||
-	       (ienc == CE_BYTES && IS_BYTES(tmp)) ||
-	       (ienc == CE_NATIVE && ! IS_LATIN1(tmp) && ! IS_UTF8(tmp))))
+	       (ienc == CE_UTF8   && IS_UTF8(tmp))   ||
+	       (ienc == CE_BYTES  && IS_BYTES(tmp))  ||
+	       (ienc == CE_NATIVE && IS_NATIVE(tmp))))
 	    SET_STRING_ELT(x, i, mkCharLenCE(CHAR(tmp), LENGTH(tmp), ienc));
     }
     UNPROTECT(1);
     return x;
 }
 
+/* `*s` should point to a string derived from `ref` after `ref` has been
+   translated to native encoding.  See `?Encoding` */
 SEXP attribute_hidden markKnown(const char *s, SEXP ref)
 {
     int ienc = 0;
@@ -1249,8 +1273,9 @@ int attribute_hidden utf8clen(char c)
     return 1 + utf8_table4[c & 0x3f];
 }
 
+/* These are misnamed: they convert a single char */
 static R_wchar_t
-utf16toucs(wchar_t high, wchar_t low) 
+utf16toucs(wchar_t high, wchar_t low)
 {
     return 0x10000 + ((int) (high & 0x3FF) << 10 ) + (int) (low & 0x3FF);
 }
@@ -1268,7 +1293,7 @@ utf8toucs32(wchar_t high, const char *s)
     return utf16toucs(high, utf8toutf16low(s));
 }
 
-/* These return the result in wchar_t.  If wchar_t is 16 bit (e.g. UTF-16LE on Windows
+/* These return the result in wchar_t.  If wchar_t is 16 bit (e.g. UTF-16LE on Windows)
    only the high surrogate is returned; call utf8toutf16low next. */
 size_t attribute_hidden
 utf8toucs(wchar_t *wc, const char *s)
@@ -1281,9 +1306,11 @@ utf8toucs(wchar_t *wc, const char *s)
     if (byte == 0) {
 	*w = (wchar_t) 0;
 	return 0;
-    } else if (byte < 0xC0) {
+    } else if (byte < 0x80) {
 	*w = (wchar_t) byte;
 	return 1;
+    } else if (byte < 0xC0) {
+	return (size_t)-1;
     } else if (byte < 0xE0) {
 	if(strlen(s) < 2) return (size_t)-2;
 	if ((s[1] & 0xC0) == 0x80) {
@@ -1302,7 +1329,7 @@ utf8toucs(wchar_t *wc, const char *s)
 	    if(byte == 0xFFFE || byte == 0xFFFF) return (size_t)-1;
 	    return 3;
 	} else return (size_t)-1;
-    
+
     } else if (byte < 0xf8) {
 	if(strlen(s) < 4) return (size_t)-2;
 	if (((s[1] & 0xC0) == 0x80) && ((s[2] & 0xC0) == 0x80) && ((s[3] & 0xC0) == 0x80)) {
@@ -1339,6 +1366,8 @@ utf8toucs(wchar_t *wc, const char *s)
     }
 }
 
+/* despite its name this translates to UTF-16 if there are (invalid)
+ * UTF-8 codings for surrogates in the input */
 size_t
 utf8towcs(wchar_t *wc, const char *s, size_t n)
 {
@@ -1355,9 +1384,9 @@ utf8towcs(wchar_t *wc, const char *s, size_t n)
 	    res ++;
 	    if (res >= n) break;
 	    if (IS_HIGH_SURROGATE(*p)) {
-	    	*(++p) = utf8toutf16low(t);
-	    	res ++;
-	    	if (res >= n) break;
+		*(++p) = utf8toutf16low(t);
+		res ++;
+		if (res >= n) break;
 	    }
 	}
     else
@@ -1367,7 +1396,38 @@ utf8towcs(wchar_t *wc, const char *s, size_t n)
 	    if (m == 0) break;
 	    res ++;
 	    if (IS_HIGH_SURROGATE(local))
-		res++;
+		res ++;
+	}
+    return (size_t) res;
+}
+
+size_t
+utf8towcs4(R_wchar_t *wc, const char *s, size_t n)
+{
+    ssize_t m, res = 0;
+    const char *t;
+    R_wchar_t *p;
+
+    if(wc)
+	for(p = wc, t = s; ; p++, t += m) {
+	    wchar_t local = L'\0';
+	    m  = (ssize_t) utf8toucs(&local, t);
+	    /* Needed to write all of R_wchar_t even on Windows where it is bigger
+	       than wchar_t */
+	    *p = (R_wchar_t) local;
+	    if (m < 0) error(_("invalid input '%s' in 'utf8towcs32'"), s);
+	    if (m == 0) break;
+	    if (IS_HIGH_SURROGATE(*p)) *p = utf8toucs32(*p, s);
+	    res ++;
+	    if (res >= n) break;
+	}
+    else
+	for(t = s; ; t += m) {
+	    wchar_t local;
+	    m  = (ssize_t) utf8toucs(&local, t);
+	    if (m < 0) error(_("invalid input '%s' in 'utf8towcs32'"), s);
+	    if (m == 0) break;
+	    res ++;
 	}
     return (size_t) res;
 }
@@ -1377,7 +1437,11 @@ static const unsigned int utf8_table1[] =
   { 0x7f, 0x7ff, 0xffff, 0x1fffff, 0x3ffffff, 0x7fffffff};
 static const unsigned int utf8_table2[] = { 0, 0xc0, 0xe0, 0xf0, 0xf8, 0xfc};
 
-/* s is NULL, or it contains at least n bytes.  Just write a a terminator if it's not big enough. */
+/* s is NULL, or it contains at least n bytes.  Just write a
+   terminator if it's not big enough.
+
+   Strangely named: converts from UCS-4 to UTF-8.
+*/
 
 static size_t Rwcrtomb32(char *s, R_wchar_t cvalue, size_t n)
 {
@@ -1389,35 +1453,61 @@ static size_t Rwcrtomb32(char *s, R_wchar_t cvalue, size_t n)
 	if (cvalue <= utf8_table1[i]) break;
     if (i >= n - 1) return 0;  /* need space for terminal null */
     if (s) {
-    	s += i;
-    	for (j = i; j > 0; j--) {
+	s += i;
+	for (j = i; j > 0; j--) {
 	    *s-- = (char) (0x80 | (cvalue & 0x3f));
 	    cvalue >>= 6;
-        }
-    	*s = (char) (utf8_table2[i] | cvalue);
+	}
+	*s = (char) (utf8_table2[i] | cvalue);
     }
     return i + 1;
 }
 
-/* on input, wc is a string encoded in UTF-16 or UCS-2 or UCS-4.
-   s can be a buffer of size n>=0 chars, or NULL.  If n=0 or s=NULL, nothing is written. 
-   The return value is the number of chars including the terminating null.  If the
-   buffer is not big enough, the result is truncated but still null-terminated */
+/* On input, wc is a wide string encoded in UTF-16 or UCS-2 or UCS-4.
+
+   s can be a buffer of size n >= 0 chars, or NULL.  If n = 0 or s =
+   NULL, nothing is written.
+
+   The return value is the number of chars including the terminating
+   null.  If the buffer is not big enough, the result is truncated but
+   still null-terminated
+*/
 attribute_hidden // but used in windlgs
 size_t wcstoutf8(char *s, const wchar_t *wc, size_t n)
 {
-    size_t m, res=0;
+    size_t m, res = 0;
     char *t;
     const wchar_t *p;
     if (!n) return 0;
     for(p = wc, t = s; ; p++) {
-    	if (IS_SURROGATE_PAIR(*p, *(p+1))) {
-    	    R_wchar_t cvalue =  ((*p & 0x3FF) << 10) + (*(p+1) & 0x3FF) + 0x010000;
+	if (IS_SURROGATE_PAIR(*p, *(p+1))) {
+	    R_wchar_t cvalue =  ((*p & 0x3FF) << 10) + (*(p+1) & 0x3FF) + 0x010000;
 	    m = Rwcrtomb32(t, cvalue, n - res);
 	    p++;
-    	} else 
-    	    m = Rwcrtomb32(t, (R_wchar_t)(*p), n - res);
-    	if (!m) break;
+	} else {
+	    if (IS_HIGH_SURROGATE(*p) || IS_LOW_SURROGATE(*p))
+		warning("unpaired surrogate Unicode point %x", *p);
+	    m = Rwcrtomb32(t, (R_wchar_t)(*p), n - res);
+	}
+	if (!m) break;
+	res += m;
+	if (t)
+	    t += m;
+    }
+    return res + 1;
+}
+
+/* convert from R_wchar_t * (UCS-4) */
+attribute_hidden
+size_t wcs4toutf8(char *s, const R_wchar_t *wc, size_t n)
+{
+    size_t m, res=0;
+    char *t;
+    const R_wchar_t *p;
+    if (!n) return 0;
+    for(p = wc, t = s; ; p++) {
+	m = Rwcrtomb32(t, (*p), n - res);
+	if (!m) break;
 	res += m;
 	if (t)
 	    t += m;
@@ -1461,18 +1551,34 @@ size_t Mbrtowc(wchar_t *wc, const char *s, size_t n, mbstate_t *ps)
 }
 
 /* Truncate a string in place (in native encoding) so that it only contains
-   valid multi-byte characters. Has no effect in non-mbcs locales. */
+   valid multi-byte characters. Has no effect in non-mbcs locales.
+
+   This function may be invoked by the error handler via
+   REvprintf->Rvsnprintf_mbcs.  Do not change it unless you are SURE that
+   your changes are compatible with the error handling mechanism.
+
+   REvprintf is also used in R_Suicide on Unix.
+   */
 attribute_hidden
 char* mbcsTruncateToValid(char *s)
 {
-    if (!mbcslocale)
+    if (!mbcslocale || *s == '\0')
 	return s;
 
     mbstate_t mb_st;
-    size_t slen = strlen(s);
+    size_t slen = strlen(s); /* at least 1 */
     size_t goodlen = 0;
 
     mbs_init(&mb_st);
+
+    if (utf8locale) {
+	/* UTF-8 is self-synchronizing so we can look back from the end
+	   for the first non-continuation byte */
+	goodlen = slen - 1; /* at least 0 */
+	/* for char == signed char we assume 2's complement representation */
+	while (goodlen && ((s[goodlen] & '\xC0') == '\x80'))
+	    --goodlen;
+    }
     while(goodlen < slen) {
 	size_t res;
 	res = mbrtowc(NULL, s + goodlen, slen - goodlen, &mb_st);
@@ -1485,8 +1591,8 @@ char* mbcsTruncateToValid(char *s)
 	goodlen += res;
     }
     return s;
-} 
-    
+}
+
 attribute_hidden
 Rboolean mbcsValid(const char *str)
 {
@@ -1544,7 +1650,7 @@ char *Rf_strchr(const char *s, int c)
 
     if(!mbcslocale || utf8locale) return strchr(s, c);
     mbs_init(&mb_st);
-    while( (used = Mbrtowc(NULL, p, MB_CUR_MAX, &mb_st)) ) {
+    while( (used = Mbrtowc(NULL, p, R_MB_CUR_MAX, &mb_st)) ) {
 	if(*p == c) return p;
 	p += used;
     }
@@ -1559,7 +1665,7 @@ char *Rf_strrchr(const char *s, int c)
 
     if(!mbcslocale || utf8locale) return strrchr(s, c);
     mbs_init(&mb_st);
-    while( (used = Mbrtowc(NULL, p, MB_CUR_MAX, &mb_st)) ) {
+    while( (used = Mbrtowc(NULL, p, R_MB_CUR_MAX, &mb_st)) ) {
 	if(*p == c) plast = p;
 	p += used;
     }
@@ -1574,7 +1680,7 @@ void R_fixslash(char *s)
     if(mbcslocale) {
 	mbstate_t mb_st; int used;
 	mbs_init(&mb_st);
-	while((used = Mbrtowc(NULL, p, MB_CUR_MAX, &mb_st))) {
+	while((used = Mbrtowc(NULL, p, R_MB_CUR_MAX, &mb_st))) {
 	    if(*p == '\\') *p = '/';
 	    p += used;
 	}
@@ -1610,7 +1716,7 @@ void R_fixbackslash(char *s)
     if(mbcslocale) {
 	mbstate_t mb_st; int used;
 	mbs_init(&mb_st);
-	while((used = Mbrtowc(NULL, p, MB_CUR_MAX, &mb_st))) {
+	while((used = Mbrtowc(NULL, p, R_MB_CUR_MAX, &mb_st))) {
 	    if(*p == '/') *p = '\\';
 	    p += used;
 	}
@@ -1633,6 +1739,7 @@ void NORET F77_SYMBOL(rexitc)(char *msg, int *nchar)
     }
     strncpy(buf, msg, (size_t) nc);
     buf[nc] = '\0';
+    mbcsTruncateToValid(buf);
     error("%s", buf);
 }
 
@@ -1650,6 +1757,7 @@ void F77_SYMBOL(rwarnc)(char *msg, int *nchar)
     }
     strncpy(buf, msg, (size_t) nc);
     buf[nc] = '\0';
+    mbcsTruncateToValid(buf);
     warning("%s", buf);
 }
 
@@ -1680,7 +1788,7 @@ char *acopy_string(const char *in)
 http://unicode.org/Public/MAPPINGS/VENDORS/ADOBE/symbol.txt
 */
 
-/* Conversion table that DOES use Private Usage Area 
+/* Conversion table that DOES use Private Usage Area
  * (should work better with specialised "symbol" fonts)
  */
 static int s2u[224] = {
@@ -1716,13 +1824,13 @@ static int s2u[224] = {
 
 /* Conversion table that does NOT use Private Usage Area (0xF8*)
  * (should work better with fonts that have good Unicode coverage)
- * 
+ *
  * NOTE that ...
  *   23D0 VERTICAL LINE EXTENTION is used for VERTICAL ARROW EXTENDER
  *   23AF HORIZONTAL LINE EXTENSION is used for HORIZONTAL ARROW EXTENDER
  * ... neither of which may be very good AND ...
  *   23AF HORIZONTAL LINE EXTENSION is also used for RADICAL EXTENDER
- * ... and that is unlikely to be right for BOTH this use AND 
+ * ... and that is unlikely to be right for BOTH this use AND
  * HORIZONTAL ARROW EXTENDER (if either)
  */
 static int s2unicode[224] = {
@@ -1757,7 +1865,7 @@ static int s2unicode[224] = {
 };
 
 void *Rf_AdobeSymbol2utf8(char *work, const char *c0, size_t nwork,
-                          Rboolean usePUA)
+			  Rboolean usePUA)
 {
     const unsigned char *c = (unsigned char *) c0;
     unsigned char *t = (unsigned char *) work;
@@ -1765,11 +1873,11 @@ void *Rf_AdobeSymbol2utf8(char *work, const char *c0, size_t nwork,
 	if (*c < 32) *t++ = ' ';
 	else {
 	    unsigned int u;
-            if (usePUA) {
-                u = (unsigned int) s2u[*c - 32];
-            } else {
-                u = (unsigned int) s2unicode[*c - 32];
-            }
+	    if (usePUA) {
+		u = (unsigned int) s2u[*c - 32];
+	    } else {
+		u = (unsigned int) s2unicode[*c - 32];
+	    }
 	    if (u < 128) *t++ = (unsigned char) u;
 	    else if (u < 0x800) {
 		*t++ = (unsigned char) (0xc0 | (u >> 6));
@@ -1787,7 +1895,7 @@ void *Rf_AdobeSymbol2utf8(char *work, const char *c0, size_t nwork,
     return (char*) work;
 }
 
-/* Convert UTF8 symbol back to single-byte symbol 
+/* Convert UTF8 symbol back to single-byte symbol
  * ASSUME fontface == 5 and 'str' is UTF8, i.e., we are dealing with
  * a UTF8 string that has been through Rf_AdobeSymbol2utf8(usePUA=TRUE)
  * (or through Rf_AdobeSymbol2ucs2() then Rf_ucstoutf8())
@@ -1801,22 +1909,22 @@ int Rf_utf8toAdobeSymbol(char *out, const char *in) {
     for ( ; *p; p += utf8clen(*p)) nc++;
     symbolint = (int *) R_alloc(nc, sizeof(int));
     for (i = 0, j = 0; i < nc; i++, j++) {
-        /* Convert UTF8 to int */
+	/* Convert UTF8 to int */
 	used = mbrtoint(&tmp, s);
-        if (used < 0) 
-            error(_("invalid UTF-8 string"));
+	if (used < 0)
+	    error(_("invalid UTF-8 string"));
 	symbolint[j] = tmp;
-        found = 0;
-        /* Convert int to CE_SYMBOL char */
-        for (k = 0; k < 224; k++) {
-            if (symbolint[j] == s2u[k]) {
-                out[j] = k + 32;
-                found = 1;
-            }
-            if (found) break;
-        }
-        if (!found) 
-            error(_("Conversion failed"));
+	found = 0;
+	/* Convert int to CE_SYMBOL char */
+	for (k = 0; k < 224; k++) {
+	    if (symbolint[j] == s2u[k]) {
+		out[j] = (char)(k + 32);
+		found = 1;
+	    }
+	    if (found) break;
+	}
+	if (!found)
+	    error(_("Conversion failed"));
 	s += used;
     }
     out[nc] = '\0';
@@ -1827,47 +1935,47 @@ const char* Rf_utf8Toutf8NoPUA(const char *in)
 {
     int i, j, used, tmp;
     /* At least enough because assumes each incoming char only one byte */
-    int nChar = 3*strlen(in) + 1;
+    int nChar = 3*(int)strlen(in) + 1;
     char *result = R_alloc(nChar, sizeof(char));
     const char *s = in;
     char *p = result;
     for (i = 0; i < nChar; i++) {
-        /* Convert UTF8 char to int */
+	/* Convert UTF8 char to int */
 	used = mbrtoint(&tmp, s);
-        /* Only re-encode if necessary 
-         * This is more efficient AND protects against input that is 
-         * NOT from Rf_AdobeSymbol2utf8(), e.g., plotmath on Windows
-         * (which is from reEnc(CE_LATIN1, CE_UTF8))
-         */
-        if (tmp > 0xF600) {
-            char inChar[4], symbolChar[2], utf8Char[4];
-            char *q;
-            for (j = 0; j < used; j++) {
-                inChar[j] = *s++;
-            }
-            inChar[used] = '\0';
-            Rf_utf8toAdobeSymbol(symbolChar, inChar);
-            Rf_AdobeSymbol2utf8(utf8Char, symbolChar, 4, FALSE);
-            q = utf8Char;
-            while (*q) {
-                *p++ = *q++;
-            }
-        } else {
-            for (j = 0; j < used; j++) {
-                *p++ = *s++;
-            }
-        }
+	/* Only re-encode if necessary
+	 * This is more efficient AND protects against input that is
+	 * NOT from Rf_AdobeSymbol2utf8(), e.g., plotmath on Windows
+	 * (which is from reEnc(CE_LATIN1, CE_UTF8))
+	 */
+	if (tmp > 0xF600) {
+	    char inChar[4], symbolChar[2], utf8Char[4];
+	    char *q;
+	    for (j = 0; j < used; j++) {
+		inChar[j] = *s++;
+	    }
+	    inChar[used] = '\0';
+	    Rf_utf8toAdobeSymbol(symbolChar, inChar);
+	    Rf_AdobeSymbol2utf8(utf8Char, symbolChar, 4, FALSE);
+	    q = utf8Char;
+	    while (*q) {
+		*p++ = *q++;
+	    }
+	} else {
+	    for (j = 0; j < used; j++) {
+		*p++ = *s++;
+	    }
+	}
     }
     *p = '\0';
     return result;
-}                         
+}
 
 const char* Rf_utf8ToLatin1AdobeSymbol2utf8(const char *in, Rboolean usePUA)
 {
   const char *latinStr;
   char *utf8str;
   latinStr = reEnc(in, CE_UTF8, CE_LATIN1, 2);
-  int nc = 3*strlen(latinStr) + 1;
+  int nc = 3*(int)strlen(latinStr) + 1;
   utf8str = R_alloc(nc, sizeof(char));
   Rf_AdobeSymbol2utf8(utf8str, latinStr, nc, usePUA);
   return utf8str;
@@ -1882,8 +1990,8 @@ int attribute_hidden Rf_AdobeSymbol2ucs2(int n)
 double R_strtod5(const char *str, char **endptr, char dec,
 		 Rboolean NA, int exact)
 {
-    LDOUBLE ans = 0.0, p10 = 10.0, fac = 1.0;
-    int n, expn = 0, sign = 1, ndigits = 0, exph = -1;
+    LDOUBLE ans = 0.0;
+    int sign = 1;
     const char *p = str;
 
     /* optional whitespace */
@@ -1917,7 +2025,10 @@ double R_strtod5(const char *str, char **endptr, char dec,
 	goto done;
     }
 
-    if(strlen(p) > 2 && p[0] == '0' && (p[1] == 'x' || p[1] == 'X')) {
+    int n, expn = 0;
+    if(strlen(p) > 2 && p[0] == '0' && (p[1] == 'x' || p[1] == 'X')) { // Hexadecimal "0x....."
+	int exph = -1;
+
 	/* This will overflow to Inf if appropriate */
 	for(p += 2; p; p++) {
 	    if('0' <= *p && *p <= '9') ans = 16*ans + (*p -'0');
@@ -1953,11 +2064,12 @@ double R_strtod5(const char *str, char **endptr, char dec,
 #define MAX_EXPONENT_PREFIX 9999
 	    for (n = 0; *p >= '0' && *p <= '9'; p++) n = (n < MAX_EXPONENT_PREFIX) ? n * 10 + (*p - '0') : n;
 	    if (ans != 0.0) { /* PR#15976:  allow big exponents on 0 */
+		LDOUBLE fac = 1.0;
 		expn += expsign * n;
 		if(exph > 0) {
-		    if (expn - exph < -122) {	/* PR#17199:  fac may overflow below if expn - exph is too small.  
-		                                   2^-122 is a bit bigger than 1E-37, so should be fine on all systems */
-		    	for (n = exph, fac = 1.0; n; n >>= 1, p2 *= p2)
+		    if (expn - exph < -122) {	/* PR#17199:  fac may overflow below if expn - exph is too small.
+						   2^-122 is a bit bigger than 1E-37, so should be fine on all systems */
+			for (n = exph, fac = 1.0; n; n >>= 1, p2 *= p2)
 			    if (n & 1) fac *= p2;
 			ans /= fac;
 			p2 = 2.0;
@@ -1976,8 +2088,9 @@ double R_strtod5(const char *str, char **endptr, char dec,
 	    }
 	}
 	goto done;
-    }
+    } // end {hexadecimal case}
 
+    int ndigits = 0;
     for ( ; *p >= '0' && *p <= '9'; p++, ndigits++) ans = 10*ans + (*p - '0');
     if (*p == dec)
 	for (p++; *p >= '0' && *p <= '9'; p++, ndigits++, expn--)
@@ -2005,6 +2118,7 @@ double R_strtod5(const char *str, char **endptr, char dec,
 	for (n = 0; n < ndigits; n++) ans /= 10.0;
 	expn += ndigits;
     }
+    LDOUBLE p10 = 10., fac = 1.0;
     if (expn < -307) { /* use underflow, not overflow */
 	for (n = -expn, fac = 1.0; n; n >>= 1, p10 *= p10)
 	    if (n & 1) fac /= p10;
@@ -2222,6 +2336,12 @@ static const char *getLocale(void)
     const char *p = getenv("R_ICU_LOCALE");
     if (p && p[0]) return p;
 
+    // FIXME: ideally, we would use a locale name corresponding to the current
+    //        C runtime locale, so as reported by setlocale(); since 4.2 this
+    //        means a UCRT locale name. As this is only approximated, we don't
+    //        use ICU by default on Windows yet for collation, even though
+    //        already having UTF-8 as the native encoding.
+
     // This call is >= Vista/Server 2008
     // ICU should accept almost all of these, e.g. en-US and uz-Latn-UZ
     PGSDLN pGSDLN = (PGSDLN)
@@ -2346,7 +2466,7 @@ int Scollate(SEXP a, SEXP b)
 	if (!envl || !envl[0])
 	    envl = getenv("LC_COLLATE");
 	int useC = envl && !strcmp(envl, "C");
-	    
+
 #ifndef Win32
 	if (!useC && strcmp("C", getLocale()) ) {
 #else
@@ -2583,10 +2703,12 @@ SEXP attribute_hidden do_findinterval(SEXP call, SEXP op, SEXP args, SEXP rho)
 # undef ERROR
 #endif
 #include <R_ext/Applic.h>
+/* .Internal(pretty(min(x), max(x), n, min.n, shrink.sml,
+ *                  c(high.u.bias, u5.bias), eps.correct))
+ */
 SEXP attribute_hidden do_pretty(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
     checkArity(op, args);
-    SEXP ans, nm, hi;
     double l = asReal(CAR(args)); args = CDR(args);
     if (!R_FINITE(l)) error(_("invalid '%s' argument"), "l");
     double u = asReal(CAR(args)); args = CDR(args);
@@ -2599,26 +2721,42 @@ SEXP attribute_hidden do_pretty(SEXP call, SEXP op, SEXP args, SEXP rho)
     double shrink = asReal(CAR(args)); args = CDR(args);
     if (!R_FINITE(shrink) || shrink <= 0.)
 	error(_("invalid '%s' argument"), "shrink.sml");
-    PROTECT(hi = coerceVector(CAR(args), REALSXP)); args = CDR(args);
-    double z;
-    if (!R_FINITE(z = REAL(hi)[0]) || z < 0.)
+    SEXP hi = PROTECT(coerceVector(CAR(args), REALSXP)); args = CDR(args); // (h, h5, f.min)
+    double *z = REAL(hi);
+    if (!R_FINITE(z[0]) || z[0] < 0.)
 	error(_("invalid '%s' argument"), "high.u.bias");
-    if (!R_FINITE(z = REAL(hi)[1]) || z < 0.)
+    if (!R_FINITE(z[1]) || z[1] < 0.)
 	error(_("invalid '%s' argument"), "u5.bias");
-    int eps = asInteger(CAR(args)); /* eps.correct */
+    if (!R_FINITE(z[2]) || z[2] <= 0.)
+	error(_("invalid '%s' argument"), "f.min");
+    int eps = asInteger(CAR(args)); args = CDR(args); /* eps.correct */
     if (eps == NA_INTEGER || eps < 0 || eps > 2)
 	error(_("'eps.correct' must be 0, 1, or 2"));
-    R_pretty(&l, &u, &n, min_n, shrink, REAL(hi), eps, 1);
-    //------ (returns 'unit' which we do not need)
-    PROTECT(ans = allocVector(VECSXP, 3));
+    int return_bounds = asLogical(CAR(args)); args = CDR(args); /* bounds */
+    if (return_bounds == NA_LOGICAL)
+	error(_("'bounds' must be TRUE or FALSE"));
+    double unit;
+    if(return_bounds)
+	       R_pretty(&l, &u, &n, min_n, shrink, REAL(hi), eps, 1);
+    else // unit  and (ns,nu)
+	unit = R_pretty(&l, &u, &n, min_n, shrink, REAL(hi), eps, 0);
+    int l_ans = return_bounds ? 3 : 4;
+    SEXP ans = PROTECT(allocVector(VECSXP, l_ans)),
+	nm = allocVector(STRSXP, l_ans);
+    setAttrib(ans, R_NamesSymbol, nm);
     SET_VECTOR_ELT(ans, 0, ScalarReal(l));
     SET_VECTOR_ELT(ans, 1, ScalarReal(u));
     SET_VECTOR_ELT(ans, 2, ScalarInteger(n));
-    nm = allocVector(STRSXP, 3);
-    setAttrib(ans, R_NamesSymbol, nm);
-    SET_STRING_ELT(nm, 0, mkChar("l"));
-    SET_STRING_ELT(nm, 1, mkChar("u"));
     SET_STRING_ELT(nm, 2, mkChar("n"));
+    if(return_bounds) {
+	SET_STRING_ELT(nm, 0, mkChar("l"));
+	SET_STRING_ELT(nm, 1, mkChar("u"));
+    } else {
+	SET_STRING_ELT(nm, 0, mkChar("ns"));
+	SET_STRING_ELT(nm, 1, mkChar("nu"));
+	SET_STRING_ELT(nm, 3, mkChar("unit"));
+	SET_VECTOR_ELT(ans,3, ScalarReal(unit));
+    }
     UNPROTECT(2);
     return ans;
 }
@@ -2707,13 +2845,13 @@ SEXP attribute_hidden do_formatC(SEXP call, SEXP op, SEXP args, SEXP rho)
  *		e.g., "0" pads leading zeros; "-" does left adjustment
  *		the other possible flags are  "+", " ", and "#".
  *	  New (Feb.98): if flag has more than one character, all are passed..
- *   
- *  Gabe Becker (2019-05-21): Added str_signif_sexp which wraps 
- *  original DATAPTR based str_signif to support ALTREPs. 
- *     
- *     Any future calls to str_signif on SEXP data should be via 
+ *
+ *  Gabe Becker (2019-05-21): Added str_signif_sexp which wraps
+ *  original DATAPTR based str_signif to support ALTREPs.
+ *
+ *     Any future calls to str_signif on SEXP data should be via
  *     str_signif_sexp to ensure ALTREP support.
- *     
+ *
  */
 
 /* <UTF8> char here is either ASCII or handled as a whole */
@@ -2846,7 +2984,7 @@ void str_signif(void *x, R_xlen_t n, const char *type, int width, int digits,
 }
 
 
-/* wrap original DATAPTR based str_signif in ITERATE_BY_REGION calls to 
+/* wrap original DATAPTR based str_signif in ITERATE_BY_REGION calls to
    support ALTREPs
 
    We still accept type because it is part of the defined API and only defaults
@@ -2873,3 +3011,16 @@ void str_signif_sexp(SEXP x, const char *type, int width, int digits,
 	error("unsupported type ");
     }
 }
+
+/* added in R 4.1.0.
+   This checks if it succeeds.
+   FIXME: is this worth inlining?
+ */
+char *Rstrdup(const char *s)
+{
+    size_t nb = strlen(s) + 1;
+    void *cpy = malloc(nb);
+    if (cpy == NULL) error("allocation error in Rstrdup");
+    memcpy (cpy, s, nb);
+    return (char *) cpy;
+}
