@@ -1,6 +1,6 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
- *  Copyright (C) 2015-2020 The R Core Team
+ *  Copyright (C) 2015-2022 The R Core Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -55,12 +55,15 @@
 extern void Rsleep(double timeint);
 #endif
 
+static int current_timeout = 0;
+
 # if (LIBCURL_VERSION_MAJOR == 7 && LIBCURL_VERSION_MINOR < 28)
 
 // curl/curl.h includes <sys/select.h> and headers it requires.
 
 #define curl_multi_wait R_curl_multi_wait
 
+
 static CURLMcode
 R_curl_multi_wait(CURLM *multi_handle,
 		  /* IGNORED */ void *unused,
@@ -215,7 +218,11 @@ static int curlMultiCheckerrs(CURLM *mhnd)
 			url, type, status, strerr);
 	    } else {
 		strerr = curl_easy_strerror(msg->data.result);
-		warning(_("URL '%s': status was '%s'"), url, strerr);
+		if (streql(strerr, "Timeout was reached"))
+		    warning(_("URL '%s': Timeout of %d seconds was reached"),
+			    url, current_timeout);
+		else
+		    warning(_("URL '%s': status was '%s'"), url, strerr);
 	    }
 	    retval++;
 	}
@@ -266,7 +273,8 @@ static void curlCommon(CURL *hnd, int redirect, int verify)
     }
 #endif
     int timeout0 = asInteger(GetOption1(install("timeout")));
-    long timeout = timeout0 = NA_INTEGER ? 0 : 1000L * timeout0;
+    long timeout = (timeout0 == NA_INTEGER) ? 0 : (1000L * timeout0);
+    current_timeout = (timeout0 == NA_INTEGER) ? 0 : timeout0;
     curl_easy_setopt(hnd, CURLOPT_CONNECTTIMEOUT_MS, timeout);
     curl_easy_setopt(hnd, CURLOPT_TIMEOUT_MS, timeout);
     if (redirect) {
@@ -323,6 +331,14 @@ in_do_curlGetHeaders(SEXP call, SEXP op, SEXP args, SEXP rho)
     int verify = asLogical(CADDR(args));
     if (verify == NA_LOGICAL)
 	error(_("invalid %s argument"), "verify");
+    int timeout = asInteger(CADDDR(args));
+    if (timeout == NA_INTEGER)
+	error(_("invalid %s argument"), "timeout");
+    SEXP sTLS = CAD4R(args);
+    const char *TLS = "";
+    if (isString(sTLS) && LENGTH(sTLS) == 1 && STRING_ELT(sTLS, 0) != NA_STRING)
+	TLS = translateChar(STRING_ELT(sTLS, 0));
+    else error(_("invalid %s argument"), "TLS");
 
     CURL *hnd = curl_easy_init();
     curl_easy_setopt(hnd, CURLOPT_URL, url);
@@ -334,6 +350,27 @@ in_do_curlGetHeaders(SEXP call, SEXP op, SEXP args, SEXP rho)
        for some ftp header info (Content-Length and Accept-ranges). */
     curl_easy_setopt(hnd, CURLOPT_WRITEFUNCTION, &rcvBody);
     curlCommon(hnd, redirect, verify);
+    if (timeout > 0) {
+	curl_easy_setopt(hnd, CURLOPT_TIMEOUT, timeout);
+	current_timeout = timeout;
+    }
+    if (!streql(TLS, "")) {
+	// 7.34.0 was released 2013-12-17
+#if LIBCURL_VERSION_MAJOR > 7 || (LIBCURL_VERSION_MAJOR == 7 && LIBCURL_VERSION_MINOR >= 34)
+	long TLS_ver = CURL_SSLVERSION_TLSv1_0;
+	if (streql(TLS, "1.0")) TLS_ver = CURL_SSLVERSION_TLSv1_0; 
+	else if (streql(TLS, "1.1")) TLS_ver = CURL_SSLVERSION_TLSv1_1; 
+	else if (streql(TLS, "1.2")) TLS_ver = CURL_SSLVERSION_TLSv1_2;
+# if LIBCURL_VERSION_MAJOR > 7 ||  (LIBCURL_VERSION_MAJOR == 7 && LIBCURL_VERSION_MINOR >= 52)
+	else if (streql(TLS, "1.3")) TLS_ver = CURL_SSLVERSION_TLSv1_3;
+# endif
+	else error(_("invalid %s argument"), "TLS");
+	curl_easy_setopt(hnd, CURLOPT_SSLVERSION, TLS_ver);
+# else
+	error("TLS argument is unsupported in this libcurl version %d.%d",
+	      LIBCURL_VERSION_MAJOR, LIBCURL_VERSION_MINOR);
+#endif
+    }
 
     char errbuf[CURL_ERROR_SIZE];
     curl_easy_setopt(hnd, CURLOPT_ERRORBUFFER, errbuf);
@@ -375,7 +412,6 @@ static void putdashes(int *pold, int new)
 }
 
 # ifdef Win32
-// ------- Windows progress bar -----------
 #include <ga.h>
 
 /* We could share this window with internet.c, then re-positioning
@@ -395,28 +431,37 @@ static void doneprogressbar(void *data)
     winprogressbar *pbar = data;
     hide(pbar->wprog);
 }
+# endif // Win32
 
 static
 int progress(void *clientp, double dltotal, double dlnow,
 	     double ultotal, double ulnow)
 {
+    CURL *hnd = (CURL *) clientp;
+    long status;
+    curl_easy_getinfo(hnd, CURLINFO_RESPONSE_CODE, &status);
+
+# ifdef Win32
     static int factor = 1;
+# endif
+
     // we only use downloads.  dltotal may be zero.
-    if (dltotal > 0.) {
+    if ((status < 300) && (dltotal > 0.)) {
 	if (total == 0.) {
 	    total = dltotal;
 	    char *type = NULL;
-	    CURL *hnd = (CURL *) clientp;
 	    curl_easy_getinfo(hnd, CURLINFO_CONTENT_TYPE, &type);
+	    REprintf("Content type '%s'", type ? type : "unknown");
 	    if (total > 1024.0*1024.0)
 		// might be longer than long, and is on 64-bit windows
 		REprintf(" length %0.0f bytes (%0.1f MB)\n",
 			 total, total/1024.0/1024.0);
 	    else if (total > 10240)
-		REprintf("Content length %d bytes (%d KB)\n",
+		REprintf(" length %d bytes (%d KB)\n",
 			 (int)total, (int)(total/1024));
 	    else
-		REprintf("Content length %d bytes\n", (int)total);
+		REprintf(" length %d bytes\n", (int)total);
+# ifdef Win32
 	    R_FlushConsole();
 	    if(R_Interactive) {
 		if (total > 1e9) factor = total/1e6; else factor = 1;
@@ -439,42 +484,15 @@ int progress(void *clientp, double dltotal, double dlnow,
     }
     R_ProcessEvents();
     return 0;
-}
-
 # else
-// ------- Unix-alike progress bar -----------
 
-static
-int progress(void *clientp, double dltotal, double dlnow,
-	     double ultotal, double ulnow)
-{
-    CURL *hnd = (CURL *) clientp;
-    long status;
-    curl_easy_getinfo(hnd, CURLINFO_RESPONSE_CODE, &status);
-
-    // we only use downloads.  dltotal may be zero.
-    if ((status < 300) && (dltotal > 0.)) {
-	if (total == 0.) {
-	    total = dltotal;
-	    char *type = NULL;
-	    curl_easy_getinfo(hnd, CURLINFO_CONTENT_TYPE, &type);
-	    REprintf("Content type '%s'", type ? type : "unknown");
-	    if (total > 1024.0*1024.0)
-		// might be longer than long, and is on 64-bit windows
-		REprintf(" length %0.0f bytes (%0.1f MB)\n",
-			 total, total/1024.0/1024.0);
-	    else if (total > 10240)
-		REprintf(" length %d bytes (%d KB)\n",
-			 (int)total, (int)(total/1024));
-	    else
-		REprintf(" length %d bytes\n", (int)total);
 	    if (R_Consolefile) fflush(R_Consolefile);
 	}
 	putdashes(&ndashes, (int)(50*dlnow/total));
     }
     return 0;
+# endif
 }
-# endif // Win32
 #endif // HAVE_LIBCURL
 
 /* download(url, destfile, quiet, mode, headers, cacheOK) */
@@ -528,7 +546,7 @@ in_do_curlDownload(SEXP call, SEXP op, SEXP args, SEXP rho)
 
     /* This comes mainly from curl --libcurl on the call used by
        download.file(method = "curl").
-       Also http://curl.haxx.se/libcurl/c/multi-single.html.
+       Also https://curl.haxx.se/libcurl/c/multi-single.html.
     */
 
     if (!cacheOK) {
