@@ -1,7 +1,7 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
  *  Copyright (C) 1995, 1996  Robert Gentleman and Ross Ihaka
- *  Copyright (C) 1997--2020  The R Core Team
+ *  Copyright (C) 1997--2022  The R Core Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -48,7 +48,7 @@
 #include "Startup.h"
 #include <R_ext/Riconv.h>
 #include <R_ext/Print.h> // for REprintf
-#include <R_ext/RS.h> // for Calloc
+#include <R_ext/RS.h> // for R_Calloc
 
 #define __SYSTEM__
 /* includes <sys/select.h> and <sys/time.h> */
@@ -226,7 +226,7 @@ addInputHandler(InputHandler *handlers, int fd, InputHandlerProc handler,
 {
     InputHandler *input, *tmp;
 //    input = (InputHandler*) calloc(1, sizeof(InputHandler));
-    input = Calloc(1, InputHandler);
+    input = R_Calloc(1, InputHandler);
 
     input->activity = activity;
     input->fileDescriptor = fd;
@@ -267,7 +267,7 @@ removeInputHandler(InputHandler **handlers, InputHandler *it)
 
     if(*handlers == it) {
 	*handlers = (*handlers)->next;
-	Free(it); // use Free to match allocation with Calloc
+	R_Free(it); // use R_Free to match allocation with R_Calloc
 	return(1);
     }
 
@@ -276,7 +276,7 @@ removeInputHandler(InputHandler **handlers, InputHandler *it)
     while(tmp) {
 	if(tmp->next == it) {
 	    tmp->next = it->next;
-	    Free(it); // use Free to match allocation with Calloc
+	    R_Free(it); // use R_Free to match allocation with R_Calloc
 	    return(1);
 	}
 	tmp = tmp->next;
@@ -498,9 +498,13 @@ attribute_hidden
 char *R_ExpandFileName_readline(const char *s, char *buff)
 {
     char *s2 = tilde_expand_word(s);
+    size_t len = strlen(s2);
 
     strncpy(buff, s2, PATH_MAX);
-    if(strlen(s2) >= PATH_MAX) buff[PATH_MAX-1] = '\0';
+    if(len >= PATH_MAX) {
+	buff[PATH_MAX-1] = '\0';
+	warning(_("expanded path length %d would be too long for\n%s\n"), len, s);
+    }
     free(s2);
     return buff;
 }
