@@ -1,7 +1,7 @@
 /*
  *  R : A Computer Language for Statistical Data Analysis
  *  Copyright (C) 1995, 1996  Robert Gentleman and Ross Ihaka
- *  Copyright (C) 1997--2017  The R Core Team
+ *  Copyright (C) 1997--2022  The R Core Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -107,44 +107,81 @@ FILE *R_OpenInitFile(void)
 char *R_ExpandFileName_readline(const char *s, char *buff);  /* sys-std.c */
 #endif
 
-static char newFileName[PATH_MAX];
-static int HaveHOME=-1;
-static char UserHOME[PATH_MAX];
+#if defined(HAVE_PWD_H)
+# include <pwd.h>
+#endif
 
-/* Only interpret inputs of the form ~ and ~/... */
 static const char *R_ExpandFileName_unix(const char *s, char *buff)
 {
-    char *p;
-
     if(s[0] != '~') return s;
-    if(strlen(s) > 1 && s[1] != '/') return s;
-    if(HaveHOME < 0) {
-	p = getenv("HOME");
-	if(p && *p && (strlen(p) < PATH_MAX)) {
-	    strcpy(UserHOME, p);
-	    HaveHOME = 1;
-	} else
-	    HaveHOME = 0;
+
+    const char *user, *temp, *s2, *home;
+    char buff2[PATH_MAX];
+    struct passwd *pass;
+
+    temp = strchr(s + 1, '/');
+    if (temp == NULL) { // ~name
+	user = s + 1;
+    } else { // ~name/path
+	/* extract text befoe first slash */
+	size_t len = (size_t)(temp - s + 1);
+	(void) strncpy(buff2, s + 1, len - 2);
+	buff2[len - 2] = '\0';
+	user = buff2;
+	s2 = s + len;
     }
-    if(HaveHOME > 0 && (strlen(UserHOME) + strlen(s+1) < PATH_MAX)) {
-	strcpy(buff, UserHOME);
-	strcat(buff, s+1);
-	return buff;
-    } else return s;
+    if (user[0] == 0) {
+	// follow readline (and not libedit) in preferring HOME
+	home = getenv("HOME");
+#if defined(HAVE_GETPWUID) && defined(HAVE_GETUID)
+	if(home == NULL || streql(home, "")) {
+	   pass = getpwuid(getuid());
+           if(pass == NULL) return s;
+	   home = pass->pw_dir;
+	}
+#endif
+	if(home == NULL) return s;
+    } else {
+#ifdef HAVE_GETPWNAM
+	pass = getpwnam(user);
+	if(pass == NULL) return s;
+	home = pass->pw_dir;
+#else
+        return s;
+#endif
+    }
+
+    if (temp == NULL) { // ~name
+	strcpy(buff, home);
+    } else { // ~name/path
+	// ask snprintf to compute the length, as GCC 12 complains otherwise.
+	size_t len = snprintf(NULL, 0, "%s/%s", home, s2);
+	// buff is passed from R_ExpandFileName, uses static array of
+	// size PATH_MAX.
+	if (len >= PATH_MAX) {
+	    warning(_("expanded path length %d would be too long for\n%s\n"),
+		       len, s);
+	    return s;
+	}
+	(void)snprintf(buff, len + 1,  "%s/%s", home, s2);
+    }
+
+    return buff;
 }
 
-/* tilde_expand (in libreadline) mallocs storage for its return value.
+/* tilde_expand_word (in libreadline) mallocs storage for its return value.
    The R entry point does not require that storage to be freed, so we
    copy the value to a static buffer, to void a memory leak in R<=1.6.0.
 
    This is not thread-safe, but as R_ExpandFileName is a public entry
-   point (in R-exts.texi) it will need to deprecated and replaced by a
+   point (in R-exts.texi) it would need to deprecated and replaced by a
    version which takes a buffer as an argument.
 
    BDR 10/2002
 */
 
 extern Rboolean UsingReadline;
+static char newFileName[PATH_MAX];
 
 const char *R_ExpandFileName(const char *s)
 {
@@ -643,6 +680,28 @@ static void warn_status(const char *cmd, int res)
 	warning(_("running command '%s' had status %d"), cmd, res);
 }
 
+static void NORET cmdError(const char *cmd, const char *format, ...)
+{
+    SEXP call = R_CurrentExpression;
+    int nextra = errno ? 3 : 1;
+
+    va_list(ap);
+    va_start(ap, format);
+    SEXP cond = R_vmakeErrorCondition(call, "cmdError", NULL,
+				      nextra, format, ap);
+    va_end(ap);
+
+    PROTECT(cond);
+    R_setConditionField(cond, 2, "cmd", mkString(cmd));
+    if (errno) {
+	R_setConditionField(cond, 3, "errno", ScalarInteger(errno));
+	R_setConditionField(cond, 4, "error", mkString(strerror(errno)));
+    }
+
+    R_signalErrorCondition(cond, call);
+    UNPROTECT(1); /* cond; not reached */
+}
+
 #define INTERN_BUFSIZE 8096
 SEXP attribute_hidden do_system(SEXP call, SEXP op, SEXP args, SEXP rho)
 {
@@ -687,7 +746,8 @@ SEXP attribute_hidden do_system(SEXP call, SEXP op, SEXP args, SEXP rho)
 	    *buf = NULL;
 	size_t buf_len = 0;
 #else
-	    buf[INTERN_BUFSIZE];
+	    buf[INTERN_BUFSIZE + MB_LEN_MAX],
+	    buf2[MB_LEN_MAX + 1]; /* overflow bytes from MBCS truncation */
 #endif
 	int i, j, res;
 	SEXP tchar, rval;
@@ -699,23 +759,42 @@ SEXP attribute_hidden do_system(SEXP call, SEXP op, SEXP args, SEXP rho)
 	else
 	    fp = R_popen_timeout(cmd, x, timeout);
 	if(!fp)
-	    error(_("cannot popen '%s', probable reason '%s'"),
-		  cmd, strerror(errno));
+	    cmdError(cmd, _("cannot popen '%s', probable reason '%s'"),
+		     cmd, strerror(errno));
 #ifdef HAVE_GETLINE
         size_t read;
         for(i = 0; (read = getline(&buf, &buf_len, fp)) != (size_t)-1; i++) {
 	    if (buf[read - 1] == '\n')
 #else
-	for (i = 0; fgets(buf, INTERN_BUFSIZE, fp); i++) {
-	    size_t read = strlen(buf);
-	    if(read >= INTERN_BUFSIZE - 1)
+	size_t read = 0;
+	size_t truncb = 0; /* number of overflow bytes from MBCS truncation */
+	buf[0] = '\0';
+	for (i = 0; fgets(buf + truncb, INTERN_BUFSIZE, fp) || truncb; i++) {
+	    buf2[0] = '\0';
+	    read = strlen(buf);
+	    if (read >= INTERN_BUFSIZE - 1 + truncb) {
 		warning(_("line %d may be truncated in call to system(, intern = TRUE)"), i + 1);
+		/* save trailing bytes in buf2 in case they are from incomplete
+		   multi-byte character, and hence will be removed from buf by
+		   mbcsTruncateToValid */
+		memcpy(buf2, buf + read - MB_LEN_MAX, MB_LEN_MAX);
+		buf2[MB_LEN_MAX] = '\0';
+		mbcsTruncateToValid(buf);
+		truncb = read - strlen(buf);
+	    }
+	    else truncb = 0;
+
 	    if (read > 0 && buf[read-1] == '\n')
 #endif
 		buf[read - 1] = '\0'; /* chop final CR */
 	    tchar = mkChar(buf);
 	    UNPROTECT(1);
 	    PROTECT(tlist = CONS(tchar, tlist));
+#ifndef HAVE_GETLINE
+	    if (truncb > 0 && truncb <= MB_LEN_MAX) 
+		/* recover overflow bytes and prepend them for next line */
+		strcpy(buf, buf2 + (MB_LEN_MAX - truncb));
+#endif
 	}
 #ifdef HAVE_GETLINE
         if (buf != NULL)
@@ -740,9 +819,10 @@ SEXP attribute_hidden do_system(SEXP call, SEXP op, SEXP args, SEXP rho)
 #endif
 	if ((res & 0xff)  == 127) {/* 127, aka -1 */
 	    if (errno)
-		error(_("error in running command: '%s'"), strerror(errno));
+		cmdError(cmd, _("error in running command: '%s'"),
+			 strerror(errno));
 	    else
-		error(_("error in running command"));
+		cmdError(cmd, _("error in running command"));
 	}
 
 	if (timeout && tost.timedout) {
@@ -881,27 +961,8 @@ void R_ProcessEvents(void)
     if (ptr_R_ProcessEvents) ptr_R_ProcessEvents();
 #endif
     R_PolledEvents();
-    if (cpuLimit > 0.0 || elapsedLimit > 0.0) {
-	double cpu, data[5];
-	R_getProcTime(data);
-	cpu = data[0] + data[1] + data[3] + data[4];
-	if (elapsedLimit > 0.0 && data[2] > elapsedLimit) {
-	    cpuLimit = elapsedLimit = -1;
-	    if (elapsedLimit2 > 0.0 && data[2] > elapsedLimit2) {
-		elapsedLimit2 = -1.0;
-		error(_("reached session elapsed time limit"));
-	    } else
-		error(_("reached elapsed time limit"));
-	}
-	if (cpuLimit > 0.0 && cpu > cpuLimit) {
-	    cpuLimit = elapsedLimit = -1;
-	    if (cpuLimit2 > 0.0 && cpu > cpuLimit2) {
-		cpuLimit2 = -1.0;
-		error(_("reached session CPU time limit"));
-	    } else
-		error(_("reached CPU time limit"));
-	}
-    }
+    if (cpuLimit > 0.0 || elapsedLimit > 0.0)
+	R_CheckTimeLimits();
 }
 
 
@@ -936,6 +997,22 @@ void fpu_setup(Rboolean start)
     _MM_SET_FLUSH_ZERO_MODE(_MM_FLUSH_ZERO_OFF);
     _MM_SET_DENORMALS_ZERO_MODE(_MM_DENORMALS_ZERO_OFF);
 #endif
+
+#if defined(__ARM_ARCH) && defined(__ARM_32BIT_STATE) && defined(__ARM_FP)
+    uint32_t fpscr;
+
+    asm volatile("vmrs %0, fpscr" : "=r"(fpscr));
+    /* clear/disable DN (default NaN) and FZ (flush to zero) bits */
+    fpscr = fpscr & 0xfcffffff;
+    asm volatile("vmsr fpscr, %0" : : "r"(fpscr));
+#elif defined(__ARM_ARCH) && defined(__ARM_64BIT_STATE) && defined(__ARM_FP)
+    uint64_t fpcr;
+
+    asm volatile("mrs %0, fpcr" : "=r"(fpcr));
+    /* clear/disable DN (default NaN) and FZ (flush to zero) bits */
+    fpcr = fpcr & 0xfffffffffcffffff;
+    asm volatile("msr fpcr, %0" : : "r"(fpcr));
+#endif
     } else {
 #ifdef __FreeBSD__
     fpsetmask(~0);
