@@ -1,4 +1,4 @@
-## Regression tests for R 3.[0-3].*
+## Regression tests for R 3.[0-3].*  _plus_ later modifications of these
 
 pdf("reg-tests-1c.pdf", encoding = "ISOLatin1.enc")
 .pt <- proc.time()
@@ -360,19 +360,19 @@ save(one, file = tempfile(), envir = my_env)
 
 ## Conversion to numeric in boundary case
 ch <- "0x1.ffa0000000001p-1"
-rr <- type.convert(ch, numerals = "allow.loss")
-rX <- type.convert(ch, numerals = "no.loss")
+rr <- type.convert(ch, numerals = "allow.loss", as.is=FALSE)
+rX <- type.convert(ch, numerals = "no.loss",    as.is=FALSE)
 stopifnot(is.numeric(rr), identical(rr, rX),
           all.equal(rr, 0.999267578125),
-	  all.equal(type.convert(ch,	      numerals = "warn"),
-		    type.convert("0x1.ffap-1",numerals = "warn"), tol = 5e-15))
+	  all.equal(type.convert(ch,	      numerals = "warn", as.is=FALSE),
+		    type.convert("0x1.ffap-1",numerals = "warn", as.is=FALSE), tol = 5e-15))
 ## type.convert(ch) was not numeric in R 3.1.0
 ##
 ch <- "1234567890123456789"
-rr <- type.convert(ch, numerals = "allow.loss")
-rX <- type.convert(ch, numerals = "no.loss")
-rx <- type.convert(ch, numerals = "no.loss", as.is = TRUE)
-tools::assertWarning(r. <- type.convert(ch, numerals = "warn.loss"))
+rr <- type.convert(ch, numerals = "allow.loss", as.is=FALSE)
+rX <- type.convert(ch, numerals = "no.loss",    as.is=FALSE)
+rx <- type.convert(ch, numerals = "no.loss",    as.is= TRUE)
+tools::assertWarning(r. <- type.convert(ch, numerals = "warn.loss", as.is=FALSE))
 stopifnot(is.numeric(rr), identical(rr, r.), all.equal(rr, 1.234567890e18),
 	  is.factor(rX),  identical(rx, ch))
 
@@ -638,9 +638,8 @@ stopifnot(identical(check2(one, , three), c(FALSE, TRUE, FALSE)))
 
 ### envRefClass check moved to methods package
 
-
-## commented out for 4.0.0, as not very important.
-## ## takes too long with JIT enabled:
+### This is an error in R 4.2.0
+## takes too long with JIT enabled:
 ## .jit.lev <- compiler::enableJIT(0)
 ## Sys.getenv("_R_CHECK_LENGTH_1_CONDITION_") -> oldV
 ## Sys.setenv("_R_CHECK_LENGTH_1_CONDITION_" = "false") # only *warn*
@@ -1072,14 +1071,14 @@ stopifnot(identical(names(myD), names(format(head(myD)))),
 		    data.frame(X2.1 = 2:1, X1.2 = 1:2)))
 ## format.data.frame() did not show "stringsAsFactors" in R <= 3.2.2
 ## Follow up: the new as.data.frame.list() must be careful with 'AsIs' columns:
-desc <- structure( c("a", NA, "z"), .Names = c("A", NA, "Z"))
+desc <- structure( c("a", NA, "z"), names = c("A", NA, "Z"))
 tools::assertError( data.frame(desc = desc, stringsAsFactors = FALSE) )
 ## however
 dd <- data.frame(desc = structure(desc, class="AsIs"),
                  row.names = c("A","M","Z"), stringsAsFactors = FALSE)
 ## is "legal" (because "AsIs" can be 'almost anything')
 dd ## <- did not format nor print correctly in R-devel early Nov.2015
-fdesc <- structure(c("a", "NA", "z"), .Names=names(desc), class="AsIs")
+fdesc <- structure(c("a", "NA", "z"), names=names(desc), class="AsIs")
 stopifnot(identical(format(dd),
                     data.frame(desc = fdesc, row.names = c("A", "M", "Z"))),
           identical(capture.output(dd),
@@ -1564,23 +1563,33 @@ stopifnot(all.equal(Fn(t), t/5))
 ## In R <= 3.2.3,  NaN values resulted in something like (n-1)/n.
 
 
-## tar() default (i.e. "no files") behaviour:
+## PR#16716: tar() default (i.e. "no files") behaviour and regular 'files'
 doit <- function(...) {
     dir.create(td <- tempfile("tar-experi"))
     setwd(td)
     dfil <- "base_Desc"
     file.copy(system.file("DESCRIPTION"), dfil)
-    ## tar w/o specified files
+    ## tar w/o specified files (empty in R < 3.3.0)
     tar("ex.tar", ... ) # all files, i.e. 'dfil'
     unlink(dfil)
     stopifnot(grepl(dfil, untar("ex.tar", list = TRUE)))
     untar("ex.tar")
     myF2 <- c(dfil, "ex.tar")
     stopifnot(identical(list.files(), myF2))
+    ## tar with specified regular files (empty in R <= 4.0.2)
+    tar("ex2.tar", files = myF2, ...)
     unlink(myF2)
+    stopifnot(identical(untar("ex2.tar", list = TRUE), myF2))
+    untar("ex2.tar")
+    stopifnot(identical(list.files(), c(myF2, "ex2.tar")))
 }
-doit() # produced an empty tar file in R < 3.3.0, PR#16716
+doit()
+if(nzchar(Sys.getenv("tar"))) doit(tar = "internal")
 if(nzchar(Sys.which("tar"))) doit(tar = "tar")
+## internal tar silently ignored unused 'files' in R <= 4.0.2
+tools::assertWarning(verbose=TRUE,
+    tar(tempfile(fileext=".tar"), files = tempfile(), tar = "internal")
+)
 
 
 ## format.POSIXlt() of Jan.1 if  1941 or '42 is involved:
@@ -1595,6 +1604,8 @@ stopifnot(identical(w8, 141:142),# exactly 1941:1942 had CEST on Jan.1
 ## for R-devel Jan.2016 to Mar.14 -- *AND* for R 3.2.4 -- the above gave
 ## integer(0)  and  c(41:42, 99:100, ..., 389:390)  respectively
 
+## the above gives 1:142 and 1:42 respectively on Solaris 10 when not using
+## --with-internal-tzcode; R-Admin recommends --with-internal-tzcode.
 
 ## tsp<- did not remove mts class
 z <- ts(cbind(1:5,1:5))
